﻿
#include "battleship_creator.h"

static void Help()
{
	printf("\nbattleship solver/creator\n");
	printf("  To slove problems: battleship_creator.exe s (file name)\n");
	printf("  To create problem: battleship_creator.exe c[1-4]\n");
	printf("    where 1-4 specifies board size. (1:8x8, 2:10x10, 3:12x12, 4:15x15)\n");
}

static void Solve(const char *filename)
{
	FILE *file = fopen(filename, "rb");
	if(!file){
		printf("failed to open: %s\n", filename);
		return;
	}
	
	char line[2048];
	while(fgets(line, sizeof(line), file)){
		StageInfo info;
		if(!info.ParseData(line)){
			break;
		}
		SolveProblem(info);
	}
	fclose(file);
}

static FILE * GetFileName(const char *ext1, int ext2)
{
	SYSTEMTIME tm;
	GetLocalTime(&tm);
	
	int count = 0;
	char fileName[256];
	for( ; ; ){
		sprintf(fileName, "%04d%02d%02d_%02d(%s%d).dat", tm.wYear, 
			tm.wMonth, tm.wDay, ++count, ext1, ext2);
		
		if(GetFileAttributes(fileName) == INVALID_FILE_ATTRIBUTES){
			FILE *ret = fopen(fileName, "wb");
			if(!ret){
				printf("failed to open: %s\n", fileName);
			}
			return ret;
		}
	}
}

static void Save(FILE *file, StageInfo &info, int *difficulty)
{
#ifdef USE_BASE64_FORMAT

	byte data[100];
	Base64Encoder encoder(data);
	
	encoder.WriteBits(0, 2);
	encoder.WriteBits(info.width, 4);
	
	for(int y = 0; y < info.height; y++){
		for(int x = 0; x < info.width; x++){
			char s = info.problem[y * info.width + x];
			
			if(s == 0){
				encoder.WriteBits(0, 1);
			}else if(s == 1){
				encoder.WriteBits(2, 2);
			}else if(s == 2){
				encoder.WriteBits(6, 3);
			}else{
				encoder.WriteBits(7, 3);
			}
		}
	}
	encoder.Flush();
	
	fprintf(file, "%d/%s\n", info.level, data);

#else

	fprintf(file, "%d/%d/%d/", info.level, info.width, info.height);
	
	for(int i = 0; i < info.width + info.height; i++){
		fprintf(file, "%X", info.hint[i]);
	}
	fprintf(file, "/");
	
	for(int i = 0; i < MAX_BOAT_SIZE; i++){
		fprintf(file, "%X", info.hist[i]);
	}
	fprintf(file, "/");
	
	for(int y = 0; y < info.height; y++){
		for(int x = 0; x < info.width; x++){
			char p = info.problem[y * info.width + x];
			if(p == UND){
				fprintf(file, ".");
			}else if(p == SEA){
				fprintf(file, "0");
			}else{
				fprintf(file, "%c", '1' + p - BSB);
			}
		}
		fprintf(file, "/");
	}
	fprintf(file, "\n");

#endif

#if 0
	printf("%d,%d,%d,%d,%d,%d\n", difficulty[0], difficulty[1], 
		difficulty[2], difficulty[3], difficulty[4], difficulty[5]);
#endif
}

static void Create(int size, int count)
{
	static const char SIZE[] = {
		8, 10, 12, 15, 
	};
	
	FILE *file = GetFileName("size", size);
	if(!file){
		return;
	}
	
	Rand rand;
	StageInfo *info = new StageInfo[count];
	int *difficulty = new int[count * MAX_COUNT];
	
	printf("creating problems...\n");
	printf("+----------------------------+\n");
	int diff = (count + 29) / 30;
	
#pragma omp parallel for
	for(int i = 0; i < count; i++){
		CreateProblem(rand, SIZE[size - 1], SIZE[size - 1], info[i], 
			&difficulty[i * MAX_COUNT]);
		if((i % diff) == diff - 1){
			printf("o");
		}
	}
	
	printf("\n");
	for(int i = 0; i < count; i++){
		Save(file, info[i], &difficulty[i * MAX_COUNT]);
	}
	
	delete[] info;
	delete[] difficulty;
	fclose(file);
	
	printf("finished.\n");
}

int main(int argc, char *argv[])
{
#ifdef _DEBUG
	_CrtSetDbgFlag(_CRTDBG_CHECK_ALWAYS_DF | _CRTDBG_LEAK_CHECK_DF | 
		_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG));
//	_CrtSetBreakAlloc(80);
#endif

	if(argc == 3 && (argv[1][0] == 's' || argv[1][0] == 'S')){
		Solve(argv[2]);
	}else if(argc == 2 && (argv[1][0] == 'c' || argv[1][0] == 'C')){
		int size = argv[1][1] - '0';
		if(size < 1 || size > 4){
			Help();
		}else{
			Create(size, 30);
		}
	}else{
		Help();
	}
	
	return 0;
}
