
#include "battleship_creator.h"

static const char HISTS[5][5] = {
	{ 3, 2, 1, 0, 0, }, // 6
	{ 4, 3, 2, 1, 0, }, // 8
	{ 4, 3, 2, 1, 0, }, // 10
	{ 4, 3, 2, 1, 1, }, // 12
	{ 5, 4, 3, 2, 1, }, // 15
};

static const int BOAT_RATIO = 8;
static const int SEA_RATIO = 4;
static const int REMOVE_LOOP = 12;
static const int PRESET_RATIO = 4;

void Problem::Initialize(Rand &rand_, char width_, char height_)
{
	rand = &rand_;
	
	width = width_;
	height = height_;
	
	answer = new Stage(width, height, stageHist, stageHint);
	stage = new Stage(width, height, stageHist, stageHint);
	
	int h;
	if(width_ == 6){
		h = 0;
	}else if(width == 8){
		h = 1;
	}else if(width == 10){
		h = 2;
	}else if(width == 12){
		h = 3;
	}else{
		h = 4;
	}
	
	for(int i = 0; i < MAX_BOAT_SIZE; i++){
		stageHist[i] = HISTS[h][i];
	}
}

void Problem::Reset()
{
	delete answer;
	answer = new Stage(width, height, stageHist, stageHint);
	
	delete stage;
	stage = new Stage(width, height, stageHist, stageHint);
}

int Problem::SearchBoatSpace(int size, bool vertical, int *und)
{
	int ex;
	int ey;
	int diff;
	
	if(vertical){
		ex = width;
		ey = height - size + 1;
		diff = answer->width;
	}else{
		ex = width - size + 1;
		ey = height;
		diff = 1;
	}
	
	int u = 0;
	
	for(int y = 0; y < ey; y++){
		for(int x = 0; x < ex; x++){
			int pos = (y + 1) * answer->width + x + 1;
			
			bool res = true;
			for(int i = 0; i < size; i++){
				if(answer->status[pos + diff * i] != UND){
					res = false;
					break;
				}
			}
			if(res){
				und[u++] = pos;
			}
		}
	}
	
	return u;
}

bool Problem::SetBoat(int size, int *und)
{
	bool vertical = (rand->Next(2) == 0);
	int u = SearchBoatSpace(size, vertical, und);
	
	if(u == 0){
		return false;
	}
	
	int d0;
	int d1;
	if(vertical){
		d0 = answer->width;
		d1 = 1;
	}else{
		d0 = 1;
		d1 = answer->width;
	}
	
	int pos = und[rand->Next(u)];
	
	answer->status[pos - d0] = SEA;
	answer->status[pos - d0 - d1] = SEA;
	answer->status[pos - d0 + d1] = SEA;
	
	for(int i = 0; i < size; i++){
		answer->status[pos] = BSA;
		answer->status[pos - d1] = SEA;
		answer->status[pos + d1] = SEA;
		pos += d0;
	}
	
	answer->status[pos] = SEA;
	answer->status[pos - d1] = SEA;
	answer->status[pos + d1] = SEA;
	
	return true;
}

bool Problem::SetBoatList()
{
	int und[MAX_BOARD_SIZE * MAX_BOARD_SIZE];
	int hist[MAX_BOAT_SIZE];
	
	int sum = 0;
	
	for(int i = 0; i < MAX_BOAT_SIZE; i++){
		hist[i] = stageHist[i];
	}
	for(int i = 1; i < MAX_BOAT_SIZE; i++){
		sum += hist[i];
	}
	
	while(sum != 0){
		int r = rand->Next(9);
		int size = 0;
		
		if(r < 1){
			size = 2;
		}else if(r < 5){
			size = 3;
		}else if(r < 7){
			size = 4;
		}else{
			size = 5;
		}
		
		if(hist[size - 1] > 0){
			if(!SetBoat(size, und)){
				return false;
			}
			
			hist[size - 1]--;
			sum--;
		}
	}
	while(hist[0] > 0){
		if(!SetBoat(1, und)){
			return false;
		}
		hist[0]--;
	}
	
	DumpStage(*answer);
	
	return true;
}

int Problem::CalcHintLine(int pos, int diff, int size)
{
	int boat = 0;
	
	for(int i = 0; i < size; i++){
		if(answer->status[pos] == UND){
			answer->status[pos] = SEA;
		}else if(answer->status[pos] == BSA){
			boat++;
		}
		pos += diff;
	}
	
	return boat;
}

bool Problem::CalcHint()
{
	int zero = 0;
	for(int x = 0; x < width; x++){
		stageHint[x] = (char)CalcHintLine(answer->width + x + 1, answer->width, 
			height);
		if(stageHint[x] == 0){
			zero++;
		}
	}
	if(zero >= width / 5){
		return false;
	}
	
	zero = 0;
	for(int y = 0; y < height; y++){
		stageHint[width + y] = 
			(char)CalcHintLine(answer->width * (y + 1) + 1, 1, width);
		if(stageHint[width + y] == 0){
			zero++;
		}
	}
	if(zero >= height / 5){
		return false;
	}
	
	return true;
}

bool Problem::SearchPreset(int boatRatio, int seaRatio)
{
	static const char LUT[16] = {
		PER, PER, PER, BSB, 
		PER, PER, PER, BSD, 
		
		PER, PER, PER, BSU, 
		BSB, BSR, BSL, BSC, 
	};
	
	int size = width * height;
	int boatCount = (boatRatio * size + size / 2) / 100;
	int seaCount = (seaRatio * size + size / 2) / 100;
	
	for( ; ; ){
		int x = rand->Next(width);
		int y = rand->Next(height);
		int pos = stage->width * (y + 1) + x + 1;
		
		int e = 0;
		if(stage->status[pos] == UND){
			if((x > 0 && stage->status[pos - 1] != UND) || 
				(x < width - 1 && stage->status[pos + 1] != UND) || 
				(y > 0 && stage->status[pos - stage->width] != UND) || 
				(y < height - 1 && stage->status[pos + stage->width] != UND)){
				e = 15;
			}
			if((x > 1 && stage->status[pos - 2] != UND) || 
				(x < width - 2 && stage->status[pos + 2] != UND) || 
				(y > 1 && stage->status[pos - stage->width * 2] != UND) || 
				(y < height - 2 && stage->status[pos + stage->width * 2] != UND)){
				e += 5;
			}
			if(x > 0){
			 	if((y > 0 && stage->status[pos - 1 - stage->width] != UND) || 
			 		(y < height - 1 && stage->status[pos - 1 + stage->width] != UND)){
					e += 3;
				}
			}
			if(x < width - 1){
			 	if((y > 0 && stage->status[pos + 1 - stage->width] != UND) || 
			 		(y < height - 1 && stage->status[pos + 1 + stage->width] != UND)){
					e += 3;
				}
			}
			if(e == 0 || rand->Next(e) == 0){
				if(answer->status[pos] == SEA){
					if(seaCount > 0){
						stage->status[pos] = SEA;
						seaCount--;
					}
				}else{
					if(boatCount > 0){
						stage->status[pos] = BSA;
						boatCount--;
					}
				}
				if(seaCount == 0 && boatCount == 0){
					break;
				}
			}
		}
	}
	
	for(int y = 1; y < answer->height - 1; y++){
		for(int x = 1; x < answer->width - 1; x++){
			int pos = y * answer->width + x;
			
			if(stage->status[pos] == BSA){
				byte nb = 0;
				
				if(answer->status[pos - 1] == SEA){
					nb |= 1;
				}
				if(answer->status[pos + 1] == SEA){
					nb |= 2;
				}
				if(answer->status[pos - answer->width] == SEA){
					nb |= 4;
				}
				if(answer->status[pos + answer->width] == SEA){
					nb |= 8;
				}
				
				if(LUT[nb] == PER){
					return false;
				}
				stage->status[pos] = LUT[nb];
			}
		}
	}
	
	DumpStage(*stage);
	
	Solve solve;
	Stage solveStage(*stage);
	if(!solve.Process(solveStage, false)){
		return false;
	}
	
	return true;
}

bool Problem::RemovePreset(int loop, int presetRatio)
{
	{
		Solve solve;
		Stage solveStage(*stage);
		if(!solve.Process(solveStage, loop, false)){
			return false;
		}
	}
	
	int pre[MAX_BOARD_SIZE * MAX_BOARD_SIZE];
	int sum = 0;
	
	for(int y = 1; y < stage->height - 1; y++){
		for(int x = 1; x < stage->width - 1; x++){
			int pos = y * stage->width + x;
			
			if(stage->status[pos] != UND){
				pre[sum++] = pos;
			}
		}
	}
	
	for(int i = sum - 1; i != 0; i--){
		int j = rand->Next(i + 1);
		int temp = pre[i];
		pre[i] = pre[j];
		pre[j] = temp;
	}
	
	for(int i = 0; i < sum; i++){
		int pos = pre[i];
		
		if(stage->status[pos] != UND){
			char prev = stage->status[pos];
			stage->status[pos] = UND;
			
			Solve solve;
			Stage solveStage(*stage);
			if(!solve.Process(solveStage, loop, false)){
				stage->status[pos] = prev;
			}
		}
	}
	
	DumpStage(*stage);
	
	sum = 0;
	for(int y = 1; y < stage->height - 1; y++){
		for(int x = 1; x < stage->width - 1; x++){
			int pos = y * stage->width + x;
			
			if(stage->status[pos] != UND){
				sum++;
			}
		}
	}
	
	int size = width * height;
	int presetCount = (presetRatio * size + size / 2) / 100;
	return (sum <= presetCount);
}

bool Problem::GetProblem(StageInfo &info, int *count)
{
#ifdef USE_BASE64_FORMAT

	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int pos = (y + 1) * stage->width + (x + 1);
			int s = stage->status[pos];
			int a = answer->status[pos];
			
			if(s == UND){
				info.problem[y * width + x] = (a == SEA) ? 0 : 1;
			}else{
				info.problem[y * width + x] = (a == SEA) ? 2 : 3;
			}
		}
	}

#else

	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int pos = (y + 1) * stage->width + (x + 1);
			
			info.problem[y * width + x] = stage->status[pos];
		}
	}

#endif

	for(int i = 0; i < MAX_BOAT_SIZE; i++){
		info.hist[i] = stageHist[i];
	}
	for(int i = 0; i < MAX_BOARD_SIZE * 2; i++){
		info.hint[i] = stageHint[i];
	}
	
	Solve solve;
	Stage solveStage(*stage);
	if(solve.Process(solveStage, false)){
		info.level = solve.GetDifficulty();
		
		memcpy(count, solve.GetCount(), sizeof(int) * 6);
		
		return true;
	}else{
		return false;
	}
}

void Problem::Create(StageInfo &info, int *count)
{
	for( ; ; ){
		Reset();
		
		if(!SetBoatList()){
			continue;
		}
		if(!CalcHint()){
			continue;
		}
		if(!SearchPreset(BOAT_RATIO, SEA_RATIO)){
			continue;
		}
		if(!RemovePreset(REMOVE_LOOP, PRESET_RATIO)){
			continue;
		}
		
		if(GetProblem(info, count)){
			break;
		}
	}
}

void CreateProblem(Rand &rand, char width, char height, StageInfo &info, 
	int *count)
{
	info.width = width;
	info.height = height;
	info.problem = new char[width * height];
	
	Problem problem;
	problem.Initialize(rand, width, height);
	problem.Create(info, count);
}
