
#include "battleship_creator.h"

bool Solve::ProcessLoop1(Stage &stage)
{
	for( ; ; ){
		bool error = false;
		bool updated = false;
		
		stage.UpdateHint(&updated, &error);
		++count[1];
		
		DumpStage(stage);
		
		if(error){
			return false;
		}else if(!updated){
			return true;
		}
	}
}

bool Solve::ProcessLoop2(Stage &stage)
{
	for( ; ; ){
		if(!ProcessLoop1(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		bool error = false;
		bool updated = false;
		
		stage.UpdateCell(&updated, &error);
		++count[2];
		
		DumpStage(stage);
		
		if(error){
			return false;
		}else if(!updated){
			return true;
		}
	}
}

bool Solve::ProcessLoop3(Stage &stage)
{
	for( ; ; ){
		if(!ProcessLoop2(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		bool error = false;
		bool updated = false;
		
		stage.UpdateNum(&updated, &error);
		++count[3];
		
		DumpStage(stage);
		
		if(error){
			return false;
		}else if(!updated){
			return true;
		}
	}
}

bool Solve::ProcessLoop4(Stage &stage)
{
	int pos = 0;
	
	for( ; ; ){
		if(!ProcessLoop3(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		if(!TryOnce(stage, &pos)){
			return true;
		}
		++count[4];
		
		DumpStage(stage);
	}
}

bool Solve::ProcessLoop4(Stage &stage, int loop)
{
	int pos = 0;
	
	for(int i = 0; i < loop; i++){
		if(!ProcessLoop3(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		if(!TryOnce(stage, &pos)){
			return true;
		}
		++count[4];
		
		DumpStage(stage);
	}
	
	return true;
}

bool Solve::Process(Stage &stage, bool dump)
{
	DumpStage(stage);
	
	count[0] = stage.width;
	
	bool error = false;
	stage.Prepare(&error);
	
	DumpStage(stage);
	
	if(error){
		return false;
	}
	
	bool result = ProcessLoop4(stage);
	
	if(dump){
		DumpResult(stage);
	}
	
	if(!result || !stage.IsSolved() || stage.HasError()){
		return false;
	}
	
	return true;
}

bool Solve::Process(Stage &stage, int loop, bool dump)
{
	DumpStage(stage);
	
	count[0] = stage.width;
	
	bool error = false;
	stage.Prepare(&error);
	
	DumpStage(stage);
	
	if(error){
		return false;
	}
	
	bool result = ProcessLoop4(stage, loop);
	
	if(dump){
		DumpResult(stage);
	}
	
	if(!result || !stage.IsSolved() || stage.HasError()){
		return false;
	}
	
	return true;
}

bool Solve::TryOnce(Stage &stage, int *pos)
{
	Stage substage(stage);
	bool error = false;
	bool updated = false;
	int start = substage.GetNextTryPos(*pos + 1);
	int p = start;
	
	if(p == 0){
		*pos = 0;
		return false;
	}
	
	for( ; ; ){
		substage.SetBoat(p, &updated, &error);
		
		if(!ProcessLoop3(substage) || substage.HasError()){
			stage.SetSea(p, &updated, &error);
			*pos = p;
			return true;
		}
		
		substage.Reset(stage);
		p = substage.GetNextTryPos(p + 1);
		if(p == start){
			break;
		}
	}
	
	*pos = p;
	return false;
}

char Solve::GetDifficulty()
{
	static const double coef[] = {
		-132.86575, 16.54848, 2.73919, -6.48588, 32.98161, -96.44162, 0, 
		-0.13303, 0.5238, -2.25781, 6.71006, 0, -0.1106, 0.17597, -0.0342, 0, 
		0.02908, 1.40026, 0, -2.43131, 0, 0, 
	};
	static const double thre[] = {
		60, 120, 180, 300, 600,
	};
	
	int c = 0;
	double d = coef[c++];
	for(int i = 0; i < MAX_COUNT; i++){
		d += (double)count[i] * coef[c++];
	}
	for(int i = 0; i < MAX_COUNT - 1; i++){
		for(int j = i + 1; j < MAX_COUNT; j++){
			d += (double)count[i] * (double)count[j] * coef[c++];
		}
	}
	
	for(char i = 0; i < 5; i++){
		if(d < thre[i]){
			return i + 1;
		}
	}
	return 5;
}

bool SolveProblem(StageInfo &info)
{
	Stage stage(info);
	Solve solve;
	
	bool result = solve.Process(stage, true);
	
	if(result){
		printf("finished.\n");
		return true;
	}else{
		printf("failed.\n");
		return false;
	}
}
