
#include "battleship_creator.h"

#ifdef USE_BASE64_FORMAT

const byte Base64Decoder::BASE64DEC_LUT[256] = {
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 62, 64,
	52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 64, 64, 64, 64, 64, 64,
	64,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
	15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 64, 64, 64, 64, 63,
	64, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
	41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 64, 64, 64, 64, 64,
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
	64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
};

const byte Base64Encoder::BASE64ENC_LUT[65] = 
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789._";

#else

bool DataReader::ParseVal(char *val)
{
	int start = pos;
	while(data[pos] != '/'){
		if(data[pos] == '\0'){
			return false;
		}
		pos++;
	}
	
	data[pos++] = '\0';
	*val = (char)atoi(&data[start]);
	
	return true;
}

bool DataReader::ParseDigitList(char *digit, int size)
{
	for(int s = 0; s < size; s++){
		digit[s] = 0;
	}
	
	int s = 0;
	while(data[pos] != '/'){
		if(data[pos] >= '0' && data[pos] <= '9'){
			if(s < size){
				digit[s++] = data[pos] - '0';
			}
		}else if(data[pos] >= 'A' && data[pos] <= 'Z'){
			if(s < size){
				digit[s++] = data[pos] - 'A' + 10;
			}
		}else{
			return false;
		}
		pos++;
	}
	pos++;
	
	return true;
}

bool DataReader::ParseData(char *problem, int size)
{
	int s = 0;
	while(s < size){
		if(data[pos] == '.'){
			problem[s++] = UND;
		}else if(data[pos] == '0'){
			problem[s++] = SEA;
		}else if(data[pos] >= '1' && data[pos] <= '6'){
			problem[s++] = data[pos] - '1' + BSB;
		}else if(data[pos] != '/'){
			return false;
		}
		pos++;
	}
	return true;
}

#endif

char StageInfo::CountHintLine(char *answer, int pos, int diff, int size)
{
	char h = 0;
	
	for(int i = 0; i < size; i++){
		if(answer[pos] == BSA){
			h++;
		}
		pos += diff;
	}
	return h;
}

void StageInfo::CountHint(char *answer)
{
	for(int x = 0; x < width; x++){
		hint[x] = CountHintLine(answer, x, width, height);
	}
	for(int y = 0; y < height; y++){
		hint[width + y] = CountHintLine(answer, y * width, 1, width);
	}
}

bool StageInfo::CountBoatLine(char *answer, int pos, int p0, int p1, 
	int diff, int size)
{
	int count = 0;
	p0 -= diff;
	p1 -= diff;
	
	for(int i = 0; i < size; i++){
		if(answer[pos] == SEA){
			if(count == 1){
				if(answer[pos + p0] == SEA && answer[pos + p1] == SEA){
					hist[count - 1]++;
				}
			}else if(count > MAX_BOAT_SIZE){
				return false;
			}else if(count != 0){
				hist[count - 1]++;
			}
			count = 0;
		}else{
			count++;
		}
		pos += diff;
	}
	if(count == 1){
		if(answer[pos + p0] == SEA && answer[pos + p1] == SEA){
			hist[count - 1]++;
		}
	}else if(count > MAX_BOAT_SIZE){
		return false;
	}else if(count != 0){
		hist[count - 1]++;
	}
	
	return true;
}

bool StageInfo::CountBoat(char *answer)
{
	for(int i = 0; i < MAX_BOAT_SIZE; i++){
		hist[i] = 0;
	}
	
	if(!CountBoatLine(answer, 0, 1, 1, width, height)){
		return false;
	}
	for(int x = 1; x < width - 1; x++){
		if(!CountBoatLine(answer, x, -1, 1, width, height)){
			return false;
		}
	}
	if(!CountBoatLine(answer, width - 1, -1, -1, width, height)){
		return false;
	}
	
	for(int y = 0; y < height; y++){
		if(!CountBoatLine(answer, y * width, 0, 0, 1, width)){
			return false;
		}
	}
	
	return true;
}

bool StageInfo::PrepareProblem(char *answer)
{
	static const char LUT[16] = {
		PER, PER, PER, BSB, 
		PER, PER, PER, BSD, 
		
		PER, PER, PER, BSU, 
		BSB, BSR, BSL, BSC, 
	};
	
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int pos = y * width + x;
			
			if(problem[pos] == BSA){
				byte nb = 0;
				
				if(x == 0 || answer[pos - 1] == SEA){
					nb |= 1;
				}
				if(x == width - 1 || answer[pos + 1] == SEA){
					nb |= 2;
				}
				if(y == 0 || answer[pos - width] == SEA){
					nb |= 4;
				}
				if(y == height - 1 || answer[pos + width] == SEA){
					nb |= 8;
				}
				
				if(LUT[nb] == PER){
					return false;
				}
				problem[pos] = LUT[nb];
			}
		}
	}
	
	return true;
}

#ifdef USE_BASE64_FORMAT

bool StageInfo::ParseData(char *data)
{
	if(data[0] >= '0' && data[0] <= '9' && data[1] == '/'){
		level = data[0] - '0';
		data += 2;
	}else{
		level = 3;
	}
	
	Base64Decoder decoder((byte *)data);
	
	decoder.ReadBits(2);
	int size = (int)decoder.ReadBits(4);
	
	width = (char)size;
	height = (char)size;
	
	char *answer = new char[size * size];
	problem = new char[size * size];
	
	int b = 0;
	
	for(int y = 0; y < size; y++){
		for(int x = 0; x < size; x++){
			if(decoder.ReadBits(1) == 0){
				answer[b] = SEA;
				problem[b] = UND;
			}else if(decoder.ReadBits(1) == 0){
				answer[b] = BSA;
				problem[b] = UND;
			}else if(decoder.ReadBits(1) == 0){
				answer[b] = SEA;
				problem[b] = SEA;
			}else{
				answer[b] = BSA;
				problem[b] = BSA;
			}
			b++;
		}
	}
	
	CountHint(answer);
	if(!CountBoat(answer)){
		delete[] answer;
		return false;
	}
	if(!PrepareProblem(answer)){
		delete[] answer;
		return false;
	}
	delete[] answer;
	
	return true;
}

#else

bool StageInfo::ParseData(char *data)
{
	DataReader reader(data);
	
	if(!reader.ParseVal(&level)){
		return false;
	}
	if(!reader.ParseVal(&width)){
		return false;
	}
	if(!reader.ParseVal(&height)){
		return false;
	}
	if(width + height > MAX_BOARD_SIZE * 2){
		return false;
	}
	
	if(!reader.ParseDigitList(hint, width + height)){
		return false;
	}
	
	if(!reader.ParseDigitList(hist, MAX_BOAT_SIZE)){
		return false;
	}
	
	problem = new char[width * height];
	if(!reader.ParseData(problem, width * height)){
		return false;
	}
	
	return true;
}

#endif

Stage::Stage(StageInfo &info)
{
	width = info.width + 2;
	height = info.height + 2;
	hist = info.hist;
	hint = info.hint;
	status = new char[width * height];
	
	memset(status, SEA, width * height);
	
	for(int y = 0; y < info.height; y++){
		for(int x = 0; x < info.width; x++){
			char p = info.problem[y * info.width + x];
			int pos = (y + 1) * width + (x + 1);
			
			status[pos] = p;
		}
	}
}

Stage::Stage(const Stage &stage)
{
	width = stage.width;
	height = stage.height;
	hist = stage.hist;
	hint = stage.hint;
	status = new char[width * height];
	
	memcpy(status, stage.status, width * height);
}

Stage::Stage(char width_, char height_, char *hist_, char *hint_)
{
	width = width_ + 2;
	height = height_ + 2;
	hist = hist_;
	hint = hint_;
	status = new char[width * height];
	
	memset(status, SEA, width * height);
	
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			status[pos] = UND;
		}
	}
}

void Stage::Reset(const Stage &stage)
{
	memcpy(status, stage.status, width * height);
}

void Stage::Prepare(bool *error)
{
	bool updated = false;
	
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			
			if(status[pos] == BSC){
				SetCircle(pos, &updated, error);
			}else if(status[pos] == BSD){
				SetTriangle(pos, width, 1, &updated, error);
			}else if(status[pos] == BSU){
				SetTriangle(pos, -width, 1, &updated, error);
			}else if(status[pos] == BSL){
				SetTriangle(pos, -1, width, &updated, error);
			}else if(status[pos] == BSR){
				SetTriangle(pos, 1, width, &updated, error);
			}
		}
	}
}

void Stage::UpdateHintLine(int pos, int d1, int d2, int size, int hint_, 
	bool *updated, bool *error)
{
	int und = 0;
	int boat = 0;
	int p = pos;
	bool bsb = false;
	
	for(int i = 0; i < size; i++){
		if(status[p] == UND){
			und++;
		}else if(status[p] == BSB){
			bsb = true;
			boat++;
		}else if(status[p] == BSA){
			boat++;
		}
		p += d1;
	}
	
	if(und == 0){
	}else if(hint_ == und + boat){
		p = pos;
		for(int i = 0; i < size; i++){
			if(status[p] == UND){
				SetBoat(p, updated, error);
			}
			p += d1;
		}
	}else if(hint_ == boat){
		p = pos;
		for(int i = 0; i < size; i++){
			if(status[p] == UND){
				SetSea(p, updated, error);
			}
			p += d1;
		}
	}
	if(bsb && hint_ - boat < 2 && 
		status[pos - d1] == UND && status[pos + d1] == UND){
		p = pos;
		for(int i = 0; i < size; i++){
			if(status[p] == BSB){
				SetBSB(p, d2, d1, updated, error);
			}
			p += d1;
		}
	}
}

void Stage::UpdateHint(bool *updated, bool *error)
{
	for(int x = 1; x < width - 1; x++){
		UpdateHintLine(x + width, width, 1, height - 2, hint[x - 1], updated, 
			error);
	}
	for(int y = 1; y < height - 1; y++){
		UpdateHintLine(y * width + 1, 1, width, width - 2, 
			hint[width + y - 3], updated, error);
	}
}

void Stage::UpdateCell(bool *updated, bool *error)
{
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			
			if(status[pos] == BSB){
				if(status[pos - 1] == SEA || status[pos + 1] == SEA){
					SetBSB(pos, width, 1, updated, error);
				}else if(status[pos - width] == SEA || 
					status[pos + width] == SEA){
					SetBSB(pos, 1, width, updated, error);
				}else if(status[pos - 1] == BSA || status[pos - 1] == BSB || 
					status[pos + 1] == BSA || status[pos + 1] == BSB){
					SetBSB(pos, 1, width, updated, error);
				}else if(status[pos - width] == BSA || 
					status[pos - width] == BSB || 
					status[pos + width] == BSA || 
					status[pos + width] == BSB){
					SetBSB(pos, width, 1, updated, error);
				}
			}else if(status[pos] == BSA){
				SetSea(pos - 1 - width, updated, error);
				SetSea(pos - 1 + width, updated, error);
				SetSea(pos + 1 - width, updated, error);
				SetSea(pos + 1 + width, updated, error);
			}
		}
	}
}

void Stage::CountNumChunk(int pos, int diff, int size, int *und, char *comb, 
	int c, int *histLine)
{
	for(int i = 0; i < c; i++){
		status[und[comb[i]]] = BSA;
	}
	
	int histChunk[MAX_BOAT_SIZE - 1] = {};
	int count = 0;
	bool error = false;
	
	for(int i = 0; i < size; i++){
		if(status[pos] == SEA || status[pos] == UND){
			if(count > MAX_BOAT_SIZE){
				error = true;
				break;
			}else if(count >= 2){
				histChunk[count - 2]++;
			}
			count = 0;
		}else{
			count++;
		}
		pos += diff;
	}
	if(count > MAX_BOAT_SIZE){
		error = true;
	}else if(count >= 2){
		histChunk[count - 2]++;
	}
	
	if(!error){
		for(int i = 0; i < MAX_BOAT_SIZE - 1; i++){
			if(histChunk[i] > histLine[i]){
				histLine[i] = histChunk[i];
			}
		}
	}
	
	for(int i = 0; i < c; i++){
		status[und[comb[i]]] = UND;
	}
}

static void CountNumChunkCombination(int n, int r, CombinationData &data)
{
	if(r == 0){
		data.stage->CountNumChunk(data.pos, data.diff, data.size, data.und, 
			data.data, data.r, data.histLine);
	}else if(n >= 0){
		CountNumChunkCombination(n - 1, r, data);
		data.data[r - 1] = (char)n;
		CountNumChunkCombination(n - 1, r - 1, data);
	}
}

void Stage::CountNumLine(int pos, int diff, int size, int hint_, 
	int *histLine, int *sum)
{
	int u = 0;
	int b = 0;
	int p = pos;
	int und[MAX_BOARD_SIZE];
	
	for(int i = 0; i < size; i++){
		if(status[p] == UND){
			und[u++] = p;
		}else if(status[p] == BSB || status[p] == BSA){
			b++;
		}
		p += diff;
	}
	
	if(u == 0 || b >= hint_){
		CountNumChunk(pos, diff, size, NULL, NULL, 0, histLine);
		
		for(int i = 0; i < MAX_BOAT_SIZE - 1; i++){
			sum[i] += histLine[i];
		}
		return;
	}
	
	CombinationData comb;
	comb.r = hint_ - b;
	comb.stage = this;
	comb.pos = pos;
	comb.diff = diff;
	comb.size = size;
	comb.und = und;
	comb.histLine = histLine;
	
	CountNumChunkCombination(u - 1, comb.r, comb);
	
	for(int i = 0; i < MAX_BOAT_SIZE - 1; i++){
		sum[i] += histLine[i];
	}
}

void Stage::SetNumChunk(int pos, int diff, int size, int *und, char *comb, 
	int c, int boatSize, int histLine, bool *result)
{
	for(int i = 0; i < c; i++){
		status[und[comb[i]]] = BSA;
	}
	
	int histChunk = 0;
	int count = 0;
	int p = pos;
	
	for(int i = 0; i < size; i++){
		if(status[p] == SEA || status[p] == UND){
			if(count == boatSize){
				histChunk++;
			}
			count = 0;
		}else{
			count++;
		}
		p += diff;
	}
	if(count == boatSize){
		histChunk++;
	}
	
	if(histChunk == histLine){
		p = pos;
		for(int i = 0; i < size; i++){
			if(status[p] == SEA || status[p] == UND){
				result[i] = true;
			}
			p += diff;
		}
	}
	
	for(int i = 0; i < c; i++){
		status[und[comb[i]]] = UND;
	}
}

static void SetNumChunkCombination(int n, int r, CombinationData &data)
{
	if(r == 0){
		data.stage->SetNumChunk(data.pos, data.diff, data.size, data.und, 
			data.data, data.r, data.boatSize, data.hist, data.result);
	}else if(n >= 0){
		SetNumChunkCombination(n - 1, r, data);
		data.data[r - 1] = (char)n;
		SetNumChunkCombination(n - 1, r - 1, data);
	}
}

void Stage::SetNumLine(int pos, int diff, int size, int hint_, int boatSize, 
	int histLine, bool *updated, bool *error)
{
	int u = 0;
	int b = 0;
	int p = pos;
	int und[MAX_BOARD_SIZE];
	
	for(int i = 0; i < size; i++){
		if(status[p] == UND){
			und[u++] = p;
		}else if(status[p] == BSB || status[p] == BSA){
			b++;
		}
		p += diff;
	}
	
	if(u == 0 || b >= hint_){
		return;
	}
	
	bool result[MAX_BOARD_SIZE] = {};
	
	CombinationData comb;
	comb.r = hint_ - b;
	comb.stage = this;
	comb.pos = pos;
	comb.diff = diff;
	comb.size = size;
	comb.und = und;
	comb.boatSize = boatSize;
	comb.hist = histLine;
	comb.result = result;
	
	SetNumChunkCombination(u - 1, comb.r, comb);
	
	p = pos;
	for(int i = 0; i < size; i++){
		if(!result[i] && status[p] == UND){
			SetBoat(p, updated, error);
		}
		p += diff;
	}
}

void Stage::UpdateNum(bool *updated, bool *error)
{
	int histLine[MAX_BOARD_SIZE * 2 * (MAX_BOAT_SIZE - 1)] = {};
	int sum[MAX_BOAT_SIZE - 1] = {};
	
	for(int x = 1; x < width - 1; x++){
		CountNumLine(x + width, width, height - 2, hint[x - 1],
			&histLine[(MAX_BOAT_SIZE - 1) * (x - 1)], sum);
	}
	for(int y = 1; y < height - 1; y++){
		CountNumLine(y * width + 1, 1, width - 2, hint[width + y - 3],
			&histLine[(MAX_BOAT_SIZE - 1) * (width + y - 3)], sum);
	}
	
	for(int i = MAX_BOAT_SIZE - 2; i >= 0; i--){
		if(sum[i] == hist[i + 1]){
			for(int x = 1; x < width - 1; x++){
				if(histLine[(MAX_BOAT_SIZE - 1) * (x - 1) + i] != 0){
					SetNumLine(x + width, width, height - 2, hint[x - 1], 
						i + 2, histLine[(MAX_BOAT_SIZE - 1) * (x - 1) + i], 
						updated, error);
				}
			}
			for(int y = 1; y < height - 1; y++){
				if(histLine[(MAX_BOAT_SIZE - 1) * (width + y - 3) + i] != 0){
					SetNumLine(y * width + 1, 1, width - 2, 
						hint[width + y - 3], i + 2, 
						histLine[(MAX_BOAT_SIZE - 1) * (width + y - 3) + i], 
						updated, error);
				}
			}
		}else if(sum[i] < hist[i + 1]){
			*error = true;
		}
	}
}

bool Stage::IsSolved()
{
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			
			if(status[pos] == UND){
				return false;
			}
		}
	}
	return true;
}

bool Stage::CheckHintLine(int pos, int diff, int size, int hint_, bool *solved)
{
	int boat = 0;
	int und = 0;
	
	for(int i = 0; i < size; i++){
		if(status[pos] == UND){
			und++;
			*solved = false;
		}else if(status[pos] != SEA){
			boat++;
		}
		pos += diff;
	}
	
	return (hint_ >= boat && hint_ <= boat + und);
}

bool Stage::CheckHint(bool *solved)
{
	for(int x = 1; x < width - 1; x++){
		if(!CheckHintLine(x + width, width, height - 2, hint[x - 1], solved)){
			return false;
		}
	}
	for(int y = 1; y < height - 1; y++){
		if(!CheckHintLine(y * width + 1, 1, width - 2, hint[width + y - 3], 
			solved)){
			return false;
		}
	}
	return true;
}

bool Stage::CheckBoatLine(int pos, int p0, int p1, int diff, int size, 
	int *histLine)
{
	int count = 0;
	p0 -= diff;
	p1 -= diff;
	
	for(int i = 0; i < size; i++){
		if(status[pos] == SEA){
			if(count == 1){
				if(status[pos + p0] == SEA && status[pos + p1] == SEA){
					histLine[count - 1]++;
				}
			}else if(count > MAX_BOAT_SIZE){
				return false;
			}else if(count != 0){
				histLine[count - 1]++;
			}
			count = 0;
		}else{
			count++;
		}
		pos += diff;
	}
	if(count == 1){
		if(status[pos + p0] == SEA && status[pos + p1] == SEA){
			histLine[count - 1]++;
		}
	}else if(count > MAX_BOAT_SIZE){
		return false;
	}else if(count != 0){
		histLine[count - 1]++;
	}
	
	return true;
}

bool Stage::CheckBoat()
{
	int histLine[MAX_BOAT_SIZE] = {};
	
	for(int i = 0; i < MAX_BOAT_SIZE; i++){
		histLine[i] = 0;
	}
	
	if(!CheckBoatLine(0, 1, 1, width, height, histLine)){
		return false;
	}
	for(int x = 1; x < width - 1; x++){
		if(!CheckBoatLine(x, -1, 1, width, height, histLine)){
			return false;
		}
	}
	if(!CheckBoatLine(width - 1, -1, -1, width, height, histLine)){
		return false;
	}
	
	for(int y = 0; y < height; y++){
		if(!CheckBoatLine(y * width, 0, 0, 1, width, histLine)){
			return false;
		}
	}
	
	for(int i = 0; i < MAX_BOAT_SIZE; i++){
		if(histLine[i] != hist[i]){
			return false;
		}
	}
	
	return true;
}

bool Stage::CheckCell()
{
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			
			if(status[pos] == BSB || status[pos] == BSA){
				char s0 = status[pos - 1 - width];
				char s1 = status[pos - 1 + width];
				char s2 = status[pos + 1 - width];
				char s3 = status[pos + 1 + width];
				
				if(s0 == BSB || s0 == BSA || s1 == BSB || s1 == BSA || 
					s2 == BSB || s2 == BSA || s3 == BSB || s3 == BSA){
					return false;
				}
			}
		}
	}
	return true;
}

bool Stage::HasError()
{
	bool solved = true;
	if(!CheckHint(&solved) || !CheckCell()){
		return true;
	}
	if(solved && !CheckBoat()){
		return true;
	}
	
	return false;
}

int Stage::GetNextTryPos(int pos)
{
	int size = width * height;
	
	for(int i = pos; i < size; i++){
		if(status[i] == UND){
			return i;
		}
	}
	for(int i = 0; i < pos; i++){
		if(status[i] == UND){
			return i;
		}
	}
	
	return 0;
}

