
#ifdef USE_BASE64_FORMAT

class Base64Decoder
{
public:
	Base64Decoder(byte *src_)
	{
		src = src_;
		bit = 0;
		pos = 0;
	}
	
	byte ReadBits(byte len)
	{
		byte ret = 0;
		
		for(byte i = 0; i < len; i++){
			if(bit == 0){
				while(src[pos] != '\0'){
					current = BASE64DEC_LUT[src[pos++]];
					if(current != 64){
						break;
					}
				}
				bit = 6;
			}
			ret = (ret << 1) | ((current >> --bit) & 1);
		}
		
		return ret;
	}

private:
	byte *src;
	byte bit;
	byte pos;
	byte current;
	
	static const byte BASE64DEC_LUT[256];
};

class Base64Encoder
{
public:
	Base64Encoder(byte *dest_)
	{
		dest = dest_;
		bit = 0;
		pos = 0;
	}
	
	void WriteBits(byte data, byte len)
	{
		for(byte i = 0; i < len; i++){
			current = (current << 1) | ((data >> (len - i - 1)) & 1);
			bit++;
			
			if(bit == 6){
				dest[pos++] = BASE64ENC_LUT[current];
				current = 0;
				bit = 0;
			}
		}
	}
	
	void Flush()
	{
		while(bit != 0){
			WriteBits(0, 1);
		}
		dest[pos] = '\0';
	}

private:
	byte *dest;
	byte bit;
	byte pos;
	byte current;
	
	static const byte BASE64ENC_LUT[65];
};

#else

class DataReader
{
public:
	DataReader(char *data)
	{
		this->data = data;
		pos = 0;
	}
	
	bool ParseVal(char *val);
	bool ParseDigitList(char *digit, int size);
	bool ParseData(char *problem, int size);

private:
	char *data;
	int pos;
};

#endif

class StageInfo
{
public:
	char width;
	char height;
	char level;
	char hist[MAX_BOAT_SIZE];
	char hint[MAX_BOARD_SIZE * 2];  
	char *problem;
	
	StageInfo()
	{
		problem = NULL;
	}
	
	~StageInfo()
	{
		delete[] problem;
	}
	
	bool ParseData(char *data);

private:
	char CountHintLine(char *answer, int pos, int diff, int size);
	void CountHint(char *answer);
	bool CountBoatLine(char *answer, int pos, int p0, int p1, int diff, 
		int size);
	bool CountBoat(char *answer);
	bool PrepareProblem(char *answer);
};

class Stage
{
public:
	char width;
	char height;
	char *hist;
	char *hint;
	char *status;
	
	Stage(StageInfo &info);
	Stage(const Stage &stage);
	Stage(char width_, char height_, char *hist_, char *hint_);
	
	~Stage()
	{
		delete[] status;
	}
	
	void SetSea(int pos, bool *updated, bool *error)
	{
		if(status[pos] == UND){
			status[pos] = SEA;
			*updated = true;
		}else if(status[pos] != SEA){
			*error = true;
		}
	}
	
	void SetBoat(int pos, bool *updated, bool *error)
	{
		if(status[pos] == UND){
			status[pos] = BSA;
			*updated = true;
		}else if(status[pos] == SEA){
			*error = true;
		}
	}
	
	void SetCircle(int pos, bool *updated, bool *error)
	{
		SetSea(pos - 1, updated, error);
		SetSea(pos + 1, updated, error);
		SetSea(pos - width, updated, error);
		SetSea(pos + width, updated, error);
		SetSea(pos - 1 - width, updated, error);
		SetSea(pos - 1 + width, updated, error);
		SetSea(pos + 1 - width, updated, error);
		SetSea(pos + 1 + width, updated, error);
		status[pos] = BSA;
	}
	
	void SetTriangle(int pos, int d1, int d2, bool *updated, bool *error)
	{
		SetBoat(pos + d1, updated, error);
		if(!*error){
			SetSea(pos - d1, updated, error);
			SetSea(pos - d2, updated, error);
			SetSea(pos + d2, updated, error);
			SetSea(pos - d1 - d2, updated, error);
			SetSea(pos - d1 + d2, updated, error);
			SetSea(pos + d1 - d2, updated, error);
			SetSea(pos + d1 + d2, updated, error);
			SetSea(pos + d1 * 2 - d2, updated, error);
			SetSea(pos + d1 * 2 + d2, updated, error);
		}
		status[pos] = BSA;
	}
	
	void SetBSB(int pos, int d1, int d2, bool *updated, bool *error)
	{
		SetBoat(pos - d1, updated, error);
		SetBoat(pos + d1, updated, error);
		if(!*error){
			int p = pos - d2;
			SetSea(p, updated, error);
			SetSea(p - d1, updated, error);
			SetSea(p + d1, updated, error);
			SetSea(p - d1 * 2, updated, error);
			SetSea(p + d1 * 2, updated, error);
			
			p = pos + d2;
			SetSea(p, updated, error);
			SetSea(p - d1, updated, error);
			SetSea(p + d1, updated, error);
			SetSea(p - d1 * 2, updated, error);
			SetSea(p + d1 * 2, updated, error);
		}
		
		status[pos] = BSA;
		*updated = true;
	}
	
	void Reset(const Stage &stage);
	void Prepare(bool *error);
	void UpdateHint(bool *updated, bool *error);
	void UpdateCell(bool *updated, bool *error);
	void CountNumChunk(int pos, int diff, int size, int *und, char *comb, 
		int c, int *histLine);
	void SetNumChunk(int pos, int diff, int size, int *und, char *comb, 
		int c, int boatSize, int histLine, bool *result);
	void UpdateNum(bool *updated, bool *error);
	bool IsSolved();
	bool HasError();
	int GetNextTryPos(int pos);

private:
	void UpdateHintLine(int pos, int d1, int d2, int size, int hint, 
		bool *updated, bool *error);
	void CountNumLine(int pos, int diff, int size, int hint, 
		int *histLine, int *sum);
	void SetNumLine(int pos, int diff, int size, int hint, int boatSize, 
		int histLine, bool *updated, bool *error);
	bool CheckHintLine(int pos, int diff, int size, int hint, bool *solved);
	bool CheckHint(bool *solved);
	bool CheckBoatLine(int pos, int p0, int p1, int diff, int size, 
		int *histLine);
	bool CheckBoat();
	bool CheckCell();
};

struct CombinationData
{
	char data[MAX_BOARD_SIZE];
	int r;
	
	Stage *stage;
	int pos;
	int diff;
	int size;
	int *und;
	int *histLine;
	int boatSize;
	int hist;
	bool *result;
};

