﻿
#include "bridge_creator.h"

static void Help()
{
	printf("\nbridge puzzle solver/creator\n");
	printf("  To slove problems: bridge_creator.exe s (file name)\n");
	printf("  To create problem: bridge_creator.exe c[1-3]\n");
	printf("    where 1-3 specifies board size.\n");
}

static void Solve(const char *filename, bool dumpMode)
{
	FILE *file = fopen(filename, "rb");
	if(!file){
		printf("failed to open: %s\n", filename);
		return;
	}
	
	char line[2048];
	while(fgets(line, sizeof(line), file)){
		StageInfo info;
		if(!info.ParseData(line)){
			break;
		}
		SolveProblem(info, dumpMode);
	}
	fclose(file);
}

static FILE * GetFileName(const char *ext1, int ext2)
{
	SYSTEMTIME tm;
	GetLocalTime(&tm);
	
	int count = 0;
	char fileName[256];
	for( ; ; ){
		sprintf(fileName, "%04d%02d%02d_%02d(%s%d).dat", tm.wYear, 
			tm.wMonth, tm.wDay, ++count, ext1, ext2);
		
		if(GetFileAttributes(fileName) == INVALID_FILE_ATTRIBUTES){
			FILE *ret = fopen(fileName, "wb");
			if(!ret){
				printf("failed to open: %s\n", fileName);
			}
			return ret;
		}
	}
}

static void Save(FILE *file, StageInfo &info)
{
	fprintf(file, "%d/%d/%d/", info.level, info.height, info.width);
	
	for(int y = 0; y < info.height; y++){
		for(int x = 0; x < info.width; x++){
			char h = info.problem[y * info.width + x];
			if(h == BLANK){
				fprintf(file, ".");
			}else{
				fprintf(file, "%c", '1' + h - HINT_1);
			}
		}
		fprintf(file, "/");
	}
	fprintf(file, "\n");
}

static void Create(int size, int count)
{
	static const char SIZE[] = {
		10, 15, 20, 
	};
	
	FILE *file = GetFileName("size", size);
	if(!file){
		return;
	}
	
	Rand rand;
	StageInfo *info = new StageInfo[count];
	
	printf("creating problems...\n");
	printf("+----------------------------+\n");
	int diff = (count + 29) / 30;
	
#pragma omp parallel for
	for(int i = 0; i < count; i++){
		CreateProblem(rand, SIZE[size - 1], SIZE[size - 1], info[i]);
		if((i % diff) == diff - 1){
			printf("o");
		}
	}
	
	printf("\n");
	for(int i = 0; i < count; i++){
		Save(file, info[i]);
	}
	
	delete[] info;
	fclose(file);
	
	printf("finished.\n");
}

int main(int argc, char *argv[])
{
#ifdef _DEBUG
	_CrtSetDbgFlag(_CRTDBG_CHECK_ALWAYS_DF | _CRTDBG_LEAK_CHECK_DF | 
		_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG));
//	_CrtSetBreakAlloc(80);
#endif

	if(argc == 3 && (argv[1][0] == 's' || argv[1][0] == 'S')){
		Solve(argv[2], false);
	}else if(argc == 3 && (argv[1][0] == 'd' || argv[1][0] == 'D')){
		Solve(argv[2], true);
	}else if(argc == 2 && (argv[1][0] == 'c' || argv[1][0] == 'C')){
		int size = argv[1][1] - '0';
		if(size < 1 || size > 3){
			Help();
		}else{
			Create(size, 30);
		}
	}else{
		Help();
	}
	
	return 0;
}
