
#include "bridge_creator.h"

const char Problem::STATUS_LUT[4] = {
	XLINE1, XLINE1, YLINE1, YLINE1, 
};

void Problem::Initialize(Rand &rand_, char size_)
{
	rand = &rand_;
	size = size_ + 2;
	
	DIFF_LUT[0] = -1;
	DIFF_LUT[1] = 1;
	DIFF_LUT[2] = -size;
	DIFF_LUT[3] = size;
	
	answer = new byte[size * size];
	stage = new byte[size * size];
	link = new ProblemLink[size * size];
	innerLink = new InnerLinkInfo(size);
	solveStage = new Stage(size, size);
}

void Problem::Reset()
{
	memset(&answer[0], XLINE1, size);
	memset(&answer[size * (size - 1)], XLINE1, size);
	for(int y = 1; y < size - 1; y++){
		answer[y * size] = YLINE1;
		answer[(y + 1) * size - 1] = YLINE1;
		
		for(int x = 1; x < size - 1; x++){
			answer[y * size + x] = BLANK;
		}
	}
}

void Problem::SetHint(int pos)
{
	answer[pos] = HINT_1;
	
	if(answer[pos - 1] == BLANK) answer[pos - 1] = HINT_0;
	if(answer[pos + 1] == BLANK) answer[pos + 1] = HINT_0;
	if(answer[pos - size] == BLANK) answer[pos - size] = HINT_0;
	if(answer[pos + size] == BLANK) answer[pos + size] = HINT_0;
}

int Problem::SetLine(int pos, int direction)
{
	int diff = DIFF_LUT[direction];
	
	int p1 = pos + diff;
	if(answer[p1] != BLANK && answer[p1] != HINT_0){
		return -1;
	}
	
	bool flag[3] = {};
	int count = 0;
	
	for(int i = 0; i < 3; i++){
		p1 += diff;
		char s = answer[p1];
		if(s == BLANK){
			flag[i] = true;
			count++;
		}else if(s == HINT_1){
			if(rand->Next(10) == 0){
				flag[i] = true;
				count++;
			}
			break;
		}else if(s != HINT_0){
			break;
		}
	}
	if(count == 0){
		return -1;
	}
	
	int f = 0;
	int r = rand->Next(count);
	for( ; ; ){
		if(flag[f]){
			if(r-- == 0){
				break;
			}
		}
		f++;
	}
	
	char status = STATUS_LUT[direction];
	
	p1 = pos + diff;
	answer[p1] = status;
	for(int i = 0; i < f; i++){
		p1 += diff;
		answer[p1] = status;
	}
	
	p1 += diff;
	SetHint(p1);
	
	return p1;
}

void Problem::SetEdgeLink(int pos, int direction)
{
	SetHint(pos);
	while(pos != -1){
		pos = SetLine(pos, direction);
	}
}

void Problem::SetInnerLink()
{
	for( ; ; ){
		int pos = innerLink->Pop();
		if(pos == -1){
			break;
		}
		
		bool nextFlag[4] = {};
		for(int i = 0; i < 6; i++){
			int nextDirection = rand->Next(4);
			if(!nextFlag[nextDirection]){
				nextFlag[nextDirection] = true;
				int nextPos = SetLine(pos, nextDirection);
				if(nextPos != -1){
					innerLink->Push(nextPos);
				}
			}
		}
	}
}

void Problem::SetLink()
{
	if(rand->Next(2) == 0){
		int direction = (rand->Next(2) == 0) ? 1 : 3;
		
		SetEdgeLink(size + 1, direction);
		SetEdgeLink(size * (size - 1) - 2, direction - 1);
	}else{
		int direction = (rand->Next(2) == 0) ? 0 : 2;
		
		SetEdgeLink(size * 2 - 2, direction);
		SetEdgeLink(size * (size - 2) + 1, direction + 1);
	}
	
	innerLink->Reset();
	
	int pos = size + 1;
	for(int i = 0; i < size - 2; i++){
		if(answer[pos] == HINT_1){
			innerLink->Push(pos);
		}
		pos++;
	}
	
	pos = size * (size - 1) - 2;
	for(int i = 0; i < size - 2; i++){
		if(answer[pos] == HINT_1){
			innerLink->Push(pos);
		}
		pos--;
	}
	
	SetInnerLink();
	
	DumpStatus(answer, size, size);
}

int Problem::SearchLink(int pos, int direction)
{
	int diff = DIFF_LUT[direction];
	char status = STATUS_LUT[direction];
	
	int p1;
	for(p1 = pos + diff ; answer[p1] == status; p1 += diff){
	}
	
	return (p1 == pos + diff) ? -1 : p1;
}

int Problem::SearchRect(int pos, int *data)
{
	data[0] = pos;
	int d = 1;
	
	for(int x0 = 0; x0 < 3; x0++){
		data[d] = SearchLink(data[d - 1], 1);
		if(data[d] == -1){
			break;
		}
		
		int prev0 = ++d;
		for(int y0 = 0; y0 < 3; y0++){
			data[d] = SearchLink(data[d - 1], 3);
			if(data[d] == -1){
				break;
			}
			
			int prev1 = ++d;
			for(int x1 = 0; x1 < 3; x1++){
				data[d] = SearchLink(data[d - 1], 0);
				if(data[d] == -1){
					break;
				}
				
				int prev2 = ++d;
				for(int y1 = 0; y1 < 3; y1++){
					data[d] = SearchLink(data[d - 1], 2);
					if(data[d] == -1){
						break;
					}
					
					if(data[d] == pos){
						return d;
					}
					d++;
				}
				
				d = prev2;
			}
			
			d = prev1;
		}
		
		d = prev0;
	}
	
	return 0;
}

void Problem::RemoveRect(int *data, int count)
{
	int r = rand->Next(count);
	
	int p0 = data[r];
	int p1 = data[r + 1];
	
	int diff;
	if(p0 / size == p1 / size){
		diff = (p0 > p1) ? DIFF_LUT[0] : DIFF_LUT[1];
	}else{
		diff = (p0 > p1) ? DIFF_LUT[2] : DIFF_LUT[3];
	}
	
	for(int pos = p0 + diff; pos != p1; pos += diff){
		answer[pos] = BLANK;
	}
	SetHint(p0);
	SetHint(p1);
}

bool Problem::CheckRectLink(int pos, bool firstCheck)
{
	int data[13];
	int count = SearchRect(pos, data);
	if(count == 0){
		return false;
	}
	
	if(firstCheck){
		if(rand->Next(10) > 0){
			RemoveRect(data, count);
		}
		if(rand->Next(5) == 0){
			RemoveRect(data, count);
		}
	}
	
	return true;
}

void Problem::CheckOrphan(int pos)
{
	answer[pos] = HINT_2;
	
	int p1;
	for(p1 = pos - 1; answer[p1] == XLINE1; p1--){
	}
	if(answer[p1] == HINT_1){
		CheckOrphan(p1);
	}
	
	for(p1 = pos + 1; answer[p1] == XLINE1; p1++){
	}
	if(answer[p1] == HINT_1){
		CheckOrphan(p1);
	}
	
	for(p1 = pos - size; answer[p1] == YLINE1; p1 -= size){
	}
	if(answer[p1] == HINT_1){
		CheckOrphan(p1);
	}
	
	for(p1 = pos + size; answer[p1] == YLINE1; p1 += size){
	}
	if(answer[p1] == HINT_1){
		CheckOrphan(p1);
	}
}

int Problem::HitCount(int pos)
{
	int count = 0;
	
	if(answer[pos - 1] == XLINE1) count++;
	if(answer[pos + 1] == XLINE1) count++;
	if(answer[pos - size] == YLINE1) count++;
	if(answer[pos + size] == YLINE1) count++;
	
	return count;
}

bool Problem::CheckLink()
{
	for(int y = 1; y < size - 1; y++){
		for(int x = 1; x < size - 1; x++){
			int pos = y * size + x;
			if(answer[pos] == HINT_1){
				bool nextFlag[4] = {};
				
				for(int i = 0; i < 4; i++){
					int nextDirection = rand->Next(4);
					if(!nextFlag[nextDirection]){
						nextFlag[nextDirection] = true;
						SetLine(pos, nextDirection);
					}
				}
			}
		}
	}
	
	for(int y = 1; y < size - 1; y++){
		for(int x = 1; x < size - 1; x++){
			int pos = y * size + x;
			if(answer[pos] == HINT_1){
				CheckRectLink(pos, true);
			}
		}
	}
	
	int rectCount = 0;
	for(int y = 1; y < size - 1; y++){
		for(int x = 1; x < size - 1; x++){
			int pos = y * size + x;
			if(answer[pos] == HINT_1){
				if(CheckRectLink(pos, false)){
					rectCount++;
				}
			}
		}
	}
	if(rectCount >= 3){
		return false;
	}else if(rectCount == 2){
		if(rand->Next(10) < 2){
			return false;
		}
	}else if(rectCount == 1){
		if(rand->Next(10) < 5){
			return false;
		}
	}
	
	for(int pos = 0; pos < size * size; pos++){
		if(answer[pos] == HINT_1){
			CheckOrphan(pos);
			break;
		}
	}
	
	for(int y = 1; y < size - 1; y++){
		for(int x = 1; x < size - 1; x++){
			int pos = y * size + x;
			if(answer[pos] == HINT_1){
				DumpStatus(answer, size, size);
				return false;
			}else if(answer[pos] == HINT_0){
				answer[pos] = BLANK;
			}else if(answer[pos] == HINT_2){
				answer[pos] = (byte)(HINT_0 + HitCount(pos));
			}
		}
	}
	
	DumpStatus(answer, size, size);
	
	return true;
}

bool Problem::SetDoubleLine(StageInfo &info)
{
	int count = 0;
	for(int y = 1; y < size - 1; y++){
		bool line = false;
		for(int x = 1; x < size - 1; x++){
			int pos = y * size + x;
			if(line && answer[pos] >= HINT_0){
				link[count++].p1 = (word)pos;
				line = false;
			}else if(!line && answer[pos] == XLINE1){
				link[count].p0 = (word)(pos - 1);
				line = true;
			}
		}
	}
	
	for(int x = 1; x < size - 1; x++){
		bool line = false;
		for(int y = 1; y < size - 1; y++){
			int pos = y * size + x;
			if(line && answer[pos] >= HINT_0){
				link[count++].p1 = (word)pos;
				line = false;
			}else if(!line && answer[pos] == YLINE1){
				link[count].p0 = (word)(pos - size);
				line = true;
			}
		}
	}
	
	for(int y = 0; y < size; y++){
		for(int x = 0; x < size; x++){
			int pos = y * size + x;
			if(answer[pos] < HINT_0){
				answer[pos] = BLANK;
			}
		}
	}
	
	int ms = (count * 3) / 10;
	int me = (count * 4) / 10;
	
	Solve solve;
	
	for(int loop = 0; loop < 10; loop++){
		memcpy(stage, answer, size * size);
		
		for(int i = count - 1; i > 0; i--){
			int pos = rand->Next(i + 1);
			ProblemLink temp = link[i];
			link[i] = link[pos];
			link[pos] = temp;
		}
		
		for(int i = 0; i < ms; i++){
			stage[link[i].p0]++;
			stage[link[i].p1]++;
		}
		
		for(int i = ms; i < me; i++){
			solveStage->Reset(stage);
			
			if(solve.Process(*solveStage, false)){
				DumpStage(*solveStage);
				
				if(solve.GetCount()[2] != 0 || solve.GetCount()[3] != 0 || 
					rand->Next(20) == 0){
					info.level = solve.GetDifficulty();
					
					GetProblem(info);
					return true;
				}else{
					break;
				}
			}
			
			stage[link[i].p0]++;
			stage[link[i].p1]++;
		}
	}
	
	return false;
}

void Problem::GetProblem(StageInfo &info)
{
	for(int y = 1; y < size - 1; y++){
		for(int x = 1; x < size - 1; x++){
			int p0 = y * size + x;
			int p1 = (y - 1) * info.width + (x - 1);
			
			info.problem[p1] = stage[p0];
		}
	}
}

void Problem::Create(StageInfo &info)
{
	for( ; ; ){
		Reset();
		SetLink();
		
		if(!CheckLink()){
			continue;
		}
		if(SetDoubleLine(info)){
			break;
		}
	}
}

void CreateProblem(Rand &rand, char width, char height, StageInfo &info)
{
	info.width = width;
	info.height = height;
	info.problem = new char[width * height];
	
	Problem problem;
	problem.Initialize(rand, width);
	problem.Create(info);
}

