
struct ProblemLink
{
	word p0;
	word p1;
};

struct InnerLinkInfo
{
	word *pos;
	word current;
	word total;
	
	InnerLinkInfo(int size)
	{
		current = 0;
		total = 0;
		pos = new word[size * size];
	}
	
	~InnerLinkInfo()
	{
		delete[] pos;
	}
	
	void Reset()
	{
		total = 0;
		current = 0;
	}
	
	void Push(int p)
	{
		for(word i = 0; i < total; i++){
			if(pos[i] == (word)p){
				return;
			}
		}
		pos[total++] = (word)p;
	}
	
	int Pop()
	{
		if(total == current){
			return -1;
		}
		return (int)pos[current++];
	}
};

class Problem
{
public:
	Problem()
	{
		answer = NULL;
		stage = NULL;
		link = NULL;
		innerLink = NULL;
		solveStage = NULL;
	}
	
	~Problem()
	{
		delete answer;
		delete stage;
		delete link;
		delete innerLink;
		delete solveStage;
	}
	
	void Initialize(Rand &rand_, char size_);
	void Create(StageInfo &info);

private:
	char size;
	byte *answer;
	byte *stage;
	ProblemLink *link;
	InnerLinkInfo *innerLink;
	Stage *solveStage;
	Rand *rand;
	
	int DIFF_LUT[4];
	static const char STATUS_LUT[4];
	
	void Reset();
	void SetHint(int pos);
	int SetLine(int pos, int direction);
	void SetEdgeLink(int pos, int direction);
	void SetInnerLink();
	void SetLink();
	int SearchLink(int pos, int direction);
	int SearchRect(int pos, int *data);
	void RemoveRect(int *data, int count);
	bool CheckRectLink(int pos, bool firstCheck);
	void CheckRectBlank(int pos);
	void CheckOrphan(int pos);
	int HitCount(int pos);
	bool CheckLink();
	bool SetDoubleLine(StageInfo &info);
	void GetProblem(StageInfo &info);
};

void CreateProblem(Rand &rand, char width, char height, StageInfo &info);

