
#include "bridge_creator.h"

bool Solve::ProcessLoop1(Stage &stage)
{
	for( ; ; ){
		int updated = stage.UpdateHint();
		bool error = stage.error;
		
		DumpStage(stage);
		
		if(error){
			return false;
		}else if(updated == 0){
			return true;
		}
		tryCount[0] += updated;
	}
}

bool Solve::ProcessLoop2(Stage &stage)
{
	for( ; ; ){
		if(!ProcessLoop1(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		int updated = stage.UpdateLink1();
		bool error = stage.error;
		
		DumpStage(stage);
		
		if(error){
			return false;
		}else if(updated == 0){
			return true;
		}
		tryCount[1] += updated;
	}
}

bool Solve::ProcessLoop3(Stage &stage)
{
	for( ; ; ){
		if(!ProcessLoop2(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		int updated = stage.UpdateLink2();
		bool error = stage.error;
		
		DumpStage(stage);
		
		if(error){
			return false;
		}else if(updated == 0){
			return true;
		}
		tryCount[2] += updated;
	}
}

bool Solve::ProcessLoop4(Stage &stage)
{
	int pos = 0;
	
	for( ; ; ){
		if(!ProcessLoop3(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		if(!TryOnce(stage, &pos)){
			return false;
		}
		
		DumpStage(stage);
	}
}

bool Solve::Process(Stage &stage, bool dump)
{
	memset(tryCount, 0, sizeof(tryCount));
	tryCount[5] = (stage.width - 2) * (stage.height - 2);
	
	DumpStage(stage);
	
	stage.Prepare();
	
	bool result = ProcessLoop4(stage);
	
	if(dump){
		DumpResult(stage);
	}
	
	if(!result || !stage.IsSolved()){
		return false;
	}
	
	return true;
}

bool Solve::TryOnce(Stage &stage, int *pos)
{
	Stage substage(stage);
	
	int direction[2];
	int start = substage.GetNextTryPos(*pos + 1, direction);
	
	int p = start;
	if(p == 0){
		*pos = 0;
		return false;
	}
	
	for( ; ; ){
		substage.SetTry(p, direction[0]);
		
		if(!ProcessLoop3(substage) || substage.HasError()){
			tryCount[3]++;
			
			DumpStage(substage);
			stage.SetTry(p, direction[1]);
			*pos = p;
			return true;
		}
		substage.Reset(stage);
		
		substage.SetTry(p, direction[1]);
		
		if(!ProcessLoop3(substage) || substage.HasError()){
			tryCount[4]++;
			
			DumpStage(substage);
			stage.SetTry(p, direction[0]);
			*pos = p;
			return true;
		}
		substage.Reset(stage);
		
		p = substage.GetNextTryPos(p + 1, direction);
		if(p == start){
			break;
		}
	}
	
	*pos = p;
	return false;
}

char Solve::GetDifficulty()
{
	static const double coef[] = {
		80.4931, 1.7312, 2.9706, 18.0117, 18.6845, 18.6845, 0, 
	};
	static const double thre[] = {
		150, 300, 500, 800, 
	};
	
	int c = 0;
	double d = coef[c++];
	for(int i = 0; i < MAX_COUNT; i++){
		d += (double)tryCount[i] * coef[c++];
	}
	
	for(char i = 0; i < 4; i++){
		if(d < thre[i]){
			return i + 1;
		}
	}
	return 5;
}

bool SolveProblem(StageInfo &info, bool dumpMode)
{
	Stage stage(info);
	Solve solve;
	
	bool result = solve.Process(stage, !dumpMode);
	
	if(result){
		if(dumpMode){
			int *d = solve.GetCount();
			printf("%d,%d,%d,%d,%d,%d\n", d[0], d[1], d[2], d[3], d[4], d[5]);
		}else{
			printf("finished.\n");
		}
		return true;
	}else{
		printf("failed.\n");
		return false;
	}
}
