
class DataReader
{
public:
	DataReader(char *data)
	{
		this->data = data;
		pos = 0;
	}
	
	bool ParseVal(char *val);
	bool ParseData(char *problem, int size);

private:
	char *data;
	int pos;
};

class StageInfo
{
public:
	char width;
	char height;
	char level;
	char *problem;
	
	StageInfo()
	{
		problem = NULL;
	}
	
	~StageInfo()
	{
		delete[] problem;
	}
	
	bool ParseData(char *data);
};

struct Color
{
	short left;
	short openNode;
};

class Stage
{
public:
	char width;
	char height;
	bool error;
	byte *status;
	byte *lineInfo;
	word *colorPos;
	Color *colorData;
	word colorTotal;
	
	Stage(StageInfo &info);
	Stage(const Stage &stage);
	Stage(char width_, char height_);
	
	~Stage()
	{
		delete[] status;
		delete[] lineInfo;
		delete[] colorPos;
		delete[] colorData;
	}
	
	void Prepare();
	
	void Reset(const Stage &stage);
	void Reset(const byte *status_);
	int UpdateHint();
	int UpdateLink1();
	int UpdateLink2();
	int GetNextTryPos(int pos, int *direction);
	void SetTry(int pos, int direction);
	bool IsSolved();
	bool HasError();

private:
	static const dword LINE_LUT[8][256];
	
	void MergeColor(int p0, int p1, dword num);
	void SetXLine(int pos, dword line, dword num);
	void SetYLine(int pos, dword line, dword num);
	int UpdateHint(int pos);
	bool LinkCheckX1(int p0, int diff, short left0);
	bool LinkCheckY1(int p0, int diff, short left0);
	int UpdateLink1(int pos);
	short SearchLeft(int p0, int diff);
	bool UpdateLinkX2(int pos, int diff1, int diff2, int diff3, byte line1, 
		byte line2, byte line3, short left0);
	bool UpdateLinkY2(int pos, int diff1, int diff2, int diff3, byte line1, 
		byte line2, byte line3, short left0);
	int GetFixCount(int pos, byte *line);
	int UpdateLink2(int pos);
	bool GetTryCount(int pos, int *direction);
};

