
#include "crosssum_creator.h"


void BlockInfo::SetCell(int pos, bool cellSet)
{
	if(pos < width * height){
		status[pos] = cellSet ? 1 : 0;
	}
}

void BlockInfo::SetCellX(int x, int y, int len, bool cellSet)
{
	if(x + len > width){
		len = width - x;
	}
	
	int pos = y * width + x;
	for(int l = 0; l < len; l++){
		status[pos + l] = cellSet ? 1 : 0;
	}
}

void BlockInfo::CheckOrphan(int x, int y)
{
	int pos = y * width + x;
	status[pos] = 2;
	
	if(x > 0 && status[pos - 1] == 1){
		CheckOrphan(x - 1, y);
	}
	if(x < width - 1 && status[pos + 1] == 1){
		CheckOrphan(x + 1, y);
	}
	if(y > 0 && status[pos - width] == 1){
		CheckOrphan(x, y - 1);
	}
	if(y < height - 1 && status[pos + width] == 1){
		CheckOrphan(x, y + 1);
	}
}

bool BlockInfo::CheckStage()
{
	for(int y = 1; y < height; y++){
		for(int x = 0; x < width; ){
			int pos = y * width + x;
			if(status[pos] == 0){
				int len = 0;
				while(x + len + 1 < width && 
					status[pos + len + 1] != 0){
					len++;
				}
				if(len == 1 || len > 9){
					return false;
				}
				x += len + 1;
			}else{
				x++;
			}
		}
	}
	
	for(int x = 1; x < width; x++){
		for(int y = 0; y < height; ){
			int pos = y * width + x;
			if(status[pos] == 0){
				int len = 0;
				while(y + len + 1 < height && 
					status[pos + (len + 1) * width] != 0){
					len++;
				}
				if(len == 1 || len > 9){
					return false;
				}
				y += len + 1;
			}else{
				y++;
			}
		}
	}
	
	bool flag = false;
	for(int y = 1; y < height; y++){
		for(int x = 1; x < width; x++){
			int pos = y * width + x;
			if(status[pos] == 1){
				CheckOrphan(x, y);
				flag = true;
				break;
			}
		}
		if(flag){
			break;
		}
	}
	for(int y = 1; y < height; y++){
		for(int x = 1; x < width; x++){
			int pos = y * width + x;
			if(status[pos] == 1){
				return false;
			}
		}
	}
	
	return true;
}

int BlockInfo::CheckLen(int sx, int sy)
{
	int result = 0;
	
	for(int y = 1; y < sy; y++){
		for(int x = 0; x < sx; ){
			int pos = y * width + x;
			if(status[pos] == 0){
				int len = 0;
				while(x + len + 1 < sx && status[pos + len + 1] != 0){
					len++;
				}
				if(len == 1 || len > 9 || (len > 4 && x + len + 1 == sx)){
					result++;
				}
				x += len + 1;
			}else{
				x++;
			}
		}
	}
	
	for(int x = 1; x < sx; x++){
		for(int y = 0; y < sy; ){
			int pos = y * width + x;
			if(status[pos] == 0){
				int len = 0;
				while(y + len + 1 < sy && 
					status[pos + (len + 1) * width] != 0){
					len++;
				}
				if(len == 1 || len > 9 || (len > 4 && y + len + 1 == sy)){
					result++;
				}
				y += len + 1;
			}else{
				y++;
			}
		}
	}
	
	return result;
}

bool BlockInfo::AdjustLenX(int sx, int sy)
{
	bool updated = false;
	
	memset(prev_len, 0, sx);
	for(int y = 1; y < sy; y++){
		for(int x = 0; x < sx; ){
			int pos = y * width + x;
			if(status[pos] == 0){
				int len = 0;
				while(x + len + 1 < sx && status[pos + len + 1] != 0){
					len++;
				}
				if(len == 1){
					updated = true;
					if(x + len + 1 < sx){
						SetCell(pos + 2, true);
						x += 2;
					}else{
						SetCell(pos, true);
						x++;
					}
				}else if(len > 9 || (len > 4 && x + len + 1 == sx) || 
					(len > 2 && prev_len[x] == len)){
					updated = true;
					int mc = 10000;
					int mcp = 0;
					for(int i = 1; i <= len; i++){
						SetCell(pos + i, false);
						int c = CheckLen(sx, sy);
						SetCell(pos + i, true);
						if(c < mc){
							mcp = i;
							mc = c;
						}
					}
					SetCell(pos + mcp, false);
					x += mcp + 1;
				}else{
					x += len + 1;
				}
			}else{
				x++;
			}
		}
		
		memset(prev_len, 0, sy);
		for(int x = 0; x < sx; ){
			int pos = y * width + x;
			if(status[pos] == 0){
				int len = 0;
				while(x + len + 1 < sx && status[pos + len + 1] != 0){
					len++;
				}
				prev_len[x] = (byte)len;
				x += len + 1;
			}else{
				x++;
			}
		}
	}
	
	return updated;
}

bool BlockInfo::AdjustLenY(int sx, int sy)
{
	bool updated = false;
	
	memset(prev_len, 0, sy);
	for(int x = 1; x < sx; x++){
		for(int y = 0; y < sy; ){
			int pos = y * width + x;
			if(status[pos] == 0){
				int len = 0;
				while(y + len + 1 < sy && 
					status[pos + (len + 1) * width] != 0){
					len++;
				}
				if(len == 1){
					updated = true;
					if(y + len + 1 < sy){
						SetCell(pos + 2 * width, true);
						y += 2;
					}else{
						SetCell(pos, true);
						y++;
					}
				}else if(len > 9 || (len > 4 && y + len + 1 == sy) || 
					(x == 1 && len > 4) || (len > 2 && prev_len[y] == len)){
					updated = true;
					int mc = 10000;
					int mcp = 0;
					for(int i = 1; i <= len; i++){
						SetCell(pos + i * width, false);
						int c = CheckLen(sx, sy);
						SetCell(pos + i * width, true);
						if(c < mc){
							mcp = i;
							mc = c;
						}
					}
					SetCell(pos + mcp * width, false);
					y += mcp + 1;
				}else{
					y += len + 1;
				}
			}else{
				y++;
			}
		}
		
		memset(prev_len, 0, sy);
		for(int y = 0; y < sy; ){
			int pos = y * width + x;
			if(status[pos] == 0){
				int len = 0;
				while(y + len + 1 < sy && 
					status[pos + (len + 1) * width] != 0){
					len++;
				}
				prev_len[y] = (byte)len;
				y += len + 1;
			}else{
				y++;
			}
		}
	}
	
	return updated;
}

void BlockInfo::GetProblem(Stage &stage, HintInfo *hint, short *hintSize)
{
	short hs = 0;
	stage.hintSize = 0;
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int pos = y * width + x;
			word p = status[pos];
			
			if(p == 0){
				char xlen = 0;
				for(int i = pos + 1; 
					i < (int)width * height && status[i] != 0; i++){
					xlen++;
				}
				
				char ylen = 0;
				for(int i = pos + width; 
					i < (int)width * height && status[i] != 0; i += width){
					ylen++;
				}
				
				if(xlen > 0 || ylen > 0){
					hint[hs].len[0] = xlen;
					hint[hs].len[1] = ylen;
					hint[hs].pos = (short)pos;
					
					if(xlen == 9){
						hint[hs].hintPos[0] = 
							(short)stage.AddHintX(pos, xlen, 45);
					}else if(xlen > 1){
						hint[hs].hintPos[0] = 
							(short)stage.AddHintX(pos, xlen, 0);
					}else{
						hint[hs].hintPos[0] = -1;
					}
					if(ylen == 9){
						hint[hs].hintPos[1] = 
							(short)stage.AddHintY(pos, ylen, 45);
					}else if(ylen > 1){
						hint[hs].hintPos[1] = 
							(short)stage.AddHintY(pos, ylen, 0);
					}else{
						hint[hs].hintPos[1] = -1;
					}
					hs++;
				}
				
				stage.status[pos] = STG_BLK;
			}else{
				stage.status[pos] = 0x1ff;
			}
		}
	}
	
	*hintSize = hs;
}

bool BlockInfo::Create()
{
	int size = width * height;
	
	memset(status, 0, sizeof(byte) * size);
	
	int sx = (width == height) ? width / 2 + 1 : width;
	int sy = height / 2 + 1;
	
	for(int y = 1; y < sy; y++){
		for(int x = 1; x < sx; ){
			if(rand->Next(3) == 0){
				SetCellX(x, y, 1, false);
				x++;
			}
			
			int len = RandLen();
			SetCellX(x, y, len, true);
			x += len;
			SetCellX(x, y, 1, false);
			x++;
			
			if(rand->Next(6) == 0){
				SetCellX(x, y, 1, false);
				x++;
			}
		}
	}
	
	if(rand->Next(2) == 0){
		for(int x = 1; x < sx; x++){
			for(int y = 1; y < sy; ){
				int pos = y * width + x;
				if(status[pos] == 0){
					int len = 1;
					while(y + len < sy && status[pos + len * width] == 0){
						len++;
					}
					if(len == 3 && (y + len == sy || rand->Next(3) == 0)){
						SetCell(pos + 2 * width, true);
						SetCell(pos + 2 * width + 1, false);
					}else if(len >= 4 && rand->Next(2) == 0){
						SetCell(pos + 3 * width, true);
						SetCell(pos + 3 * width + 1, false);
					}else if(len >= 8){
						SetCell(pos + 3 * width, true);
						SetCell(pos + 3 * width + 1, false);
					}
					y += len;
				}else{
					y++;
				}
			}
		}
	}else{
		for(int x = 1; x < sx; x++){
			for(int y = 1; y < sy; ){
				int pos = y * width + x;
				if(status[pos] == 0){
					int len = 1;
					while(y + len < sy && status[pos + len * width] == 0){
						len++;
					}
					if(len == 3 && (y + len == sy || rand->Next(6) == 0)){
						SetCell(pos + 2 * width, true);
						SetCell(pos + 2 * width + 1, false);
					}else if(len >= 4 && rand->Next(4) == 0){
						SetCell(pos + 3 * width, true);
						SetCell(pos + 3 * width + 1, false);
					}else if(len >= 8 && rand->Next(2) == 0){
						SetCell(pos + 3 * width, true);
						SetCell(pos + 3 * width + 1, false);
					}
					y += len;
				}else{
					y++;
				}
			}
		}
	}
	
	if(AdjustLenY(sx, sy)){
		bool flag = false;
		for(int i = 0; i < 6; i++){
			if(!AdjustLenX(sx, sy) && !AdjustLenY(sx, sy)){
				flag = true;
				break;
			}
		}
		if(!flag){
			return false;
		}
	}
	
	if(width == height){
		for(int y = 1; y < sy; y++){
			for(int x = 1; x < sx; x++){
				int p0 = y * width + x;
				int p1 = (height - y) * width + (width - x);
				int p2 = (height - x) * width + y;
				int p3 = x * width + (width - y);
				
				status[p1] = status[p0];
				status[p2] = status[p0];
				status[p3] = status[p0];
			}
		}
	}else{
		for(int y = 1; y < sy; y++){
			for(int x = 1; x < sx; x++){
				int p0 = y * width + x;
				int p1 = (height - y) * width + (width - x);
				
				status[p1] = status[p0];
			}
		}
	}
	
	if(!CheckStage()){
		return false;
	}
	
	return true;
}

