﻿
#include "crosssum_creator.h"

static void Help()
{
	printf("\ncross sum solver/creator\n");
	printf("  To slove problems: crossssum_creator.exe s (file name)\n");
	printf("  To create problem: crossssum_creator.exe c[1-4]\n");
	printf("    where 1-4 specifies board size.\n");
}

static void Solve(const char *filename, bool dumpMode)
{
	FILE *file = fopen(filename, "rb");
	if(!file){
		printf("failed to open: %s\n", filename);
		return;
	}
	
	char line[2048];
	while(fgets(line, sizeof(line), file)){
		StageInfo info;
		if(!info.ParseData(line)){
			break;
		}
		SolveProblem(info, dumpMode);
	}
	fclose(file);
}

static FILE * GetFileName(const char *ext1, int ext2)
{
	SYSTEMTIME tm;
	GetLocalTime(&tm);
	
	int count = 0;
	char fileName[256];
	for( ; ; ){
		sprintf(fileName, "%04d%02d%02d_%02d(%s%d).dat", tm.wYear, 
			tm.wMonth, tm.wDay, ++count, ext1, ext2);
		
		if(GetFileAttributes(fileName) == INVALID_FILE_ATTRIBUTES){
			FILE *ret = fopen(fileName, "wb");
			if(!ret){
				printf("failed to open: %s\n", fileName);
			}
			return ret;
		}
	}
}

static void Save(FILE *file, StageInfo &info)
{
	static const char LUT[] = {
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 
		'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 
		'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 
		'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 
		'U', 'V', 'W', 'X', 'Y', 'Z', 
	};
	
	fprintf(file, "%d/%d/%d/", info.level, info.width - 1, info.height - 1);
	
	int skip = 0;
	for(int y = 1; y < info.height; y++){
		for(int x = 1; x < info.width; x++){
			int p = y * info.width + x;
			
			if(info.problem[p] == DIGIT_0){
				++skip;
			}else{
				if(skip != 0){
					if(skip > 16){
						fprintf(file, "z%c", 'k' + skip - 17);
					}else{
						fprintf(file, "%c", 'k' + skip - 1);
					}
					skip = 0;
				}
				
				if(info.problem[p] == BLOCK){
					fprintf(file, ".");
				}else{
					fprintf(file, "%c%c", LUT[info.hint[p * 2 + 1]], 
						LUT[info.hint[p * 2]]);
				}
			}
		}
	}
	
	if(skip != 0){
		if(skip > 16){
			fprintf(file, "z%c", 'k' + skip - 17);
		}else{
			fprintf(file, "%c", 'k' + skip - 1);
		}
	}
	
	for(int x = 1; x < info.width; x++){
		if(info.problem[x] == HINT){
			fprintf(file, "%c", LUT[info.hint[x * 2 + 1]]);
		}
	}
	for(int y = 1; y < info.height; y++){
		int p = y * info.width;
		
		if(info.problem[p] == HINT){
			fprintf(file, "%c", LUT[info.hint[p * 2]]);
		}
	}
	
	fprintf(file, "\n");
}

static void Create(int size, int count)
{
	static const char SIZE[] = {
		11, 11, 
		15, 15, 
		19, 15, 
		25, 17, 
	};
	
	FILE *file = GetFileName("size", size);
	if(!file){
		return;
	}
	
	Rand rand;
	StageInfo *info = new StageInfo[count];
	
	printf("creating problems...\n");
	printf("+----------------------------+\n");
	int diff = (count + 29) / 30;
	size = (size - 1) * 2;
	
#pragma omp parallel for
	for(int i = 0; i < count; i++){
		CreateProblem(rand, SIZE[size] + 1, SIZE[size + 1] + 1, info[i]);
		if((i % diff) == diff - 1){
			printf("o");
		}
	}
	
	printf("\n");
	for(int i = 0; i < count; i++){
		Save(file, info[i]);
	}
	
	delete[] info;
	fclose(file);
	
	printf("finished.\n");
}

int main(int argc, char *argv[])
{
#ifdef _DEBUG
	_CrtSetDbgFlag(_CRTDBG_CHECK_ALWAYS_DF | _CRTDBG_LEAK_CHECK_DF | 
		_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG));
//	_CrtSetBreakAlloc(86);
#endif

	DWORD start = GetTickCount();
	
	Stage::PrepareCombination3();
	
	if(argc == 3 && (argv[1][0] == 's' || argv[1][0] == 'S')){
		Solve(argv[2], false);
	}else if(argc == 3 && (argv[1][0] == 'd' || argv[1][0] == 'D')){
		Solve(argv[2], true);
	}else if(argc == 2 && (argv[1][0] == 'c' || argv[1][0] == 'C')){
		int size = argv[1][1] - '0';
		if(size < 1 || size > 4){
			Help();
		}else{
			Create(size, 30);
		}
	}else{
		Help();
	}
	
	printf("time=%d\n", GetTickCount() - start);
	return 0;
}
