
#include "crosssum_creator.h"

bool Solve::ProcessLoop1(Stage &stage)
{
	for( ; ; ){
		int updated = stage.UpdateAll();
		bool error = stage.error;
		
		DumpStage(stage);
		
		if(error){
			return false;
		}else if(updated == 0){
			return true;
		}
		tryCount[0] += updated;
	}
}

bool Solve::ProcessLoop2(Stage &stage)
{
	for( ; ; ){
		if(!ProcessLoop1(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		int updated = stage.UpdateCombination();
		bool error = stage.error;
		
		DumpStage(stage);
		
		if(error){
			return false;
		}else if(updated == 0){
			return true;
		}
		tryCount[2] += updated;
	}
}

bool Solve::ProcessLoop3(Stage &stage)
{
	for( ; ; ){
		if(!ProcessLoop2(stage)){
			return false;
		}
		if(stage.IsSolved()){
			return true;
		}
		
		if(!TryOnce(stage)){
			return false;
		}
		
		DumpStage(stage);
	}
}

bool Solve::Process(Stage &stage, bool dump)
{
	memset(tryCount, 0, sizeof(tryCount));
	tryCount[5] = stage.width * stage.height;
	
	DumpStage(stage);
	
	bool result = ProcessLoop3(stage);
	
	if(dump){
		DumpResult(stage);
	}
	
	if(!result || !stage.IsSolved()){
		return false;
	}
	
	return true;
}

bool Solve::ProcessOnce(Stage &stage, int *count)
{
	bool result = ProcessLoop1(stage);
	
	if(!result){
		return false;
	}
	*count = stage.LeftCount();
	
	return true;
}

bool Solve::ProcessSecond(Stage &stage, int *count)
{
	bool result = ProcessLoop2(stage);
	
	if(!result){
		return false;
	}
	*count = stage.LeftCount();
	
	return true;
}

bool Solve::ProcessFirst(Stage &stage, bool dump)
{
	memset(tryCount, 0, sizeof(tryCount));
	tryCount[5] = stage.width * stage.height;
	
	DumpStage(stage);
	
	bool result = ProcessLoop2(stage);
	
	if(dump){
		DumpResult(stage);
	}
	
	if(!result || !stage.IsSolved()){
		return false;
	}
	
	return true;
}

bool Solve::TryOnce(Stage &stage)
{
	Stage substage(stage);
	
	word num[2];
	int pos = substage.GetNextTryPos(0, num);
	
	if(pos == 0){
		return false;
	}
	
	for( ; ; ){
		substage.SetTry(pos, num[0]);
		
		if(!ProcessLoop3(substage)){
			tryCount[3]++;
			
			DumpStage(substage);
			stage.SetTry(pos, num[1]);
			return true;
		}
		substage.Reset(stage);
		
		substage.SetTry(pos, num[1]);
		
		if(!ProcessLoop3(substage)){
			tryCount[4]++;
			
			DumpStage(substage);
			stage.SetTry(pos, num[0]);
			return true;
		}
		substage.Reset(stage);
		
		pos = substage.GetNextTryPos(pos + 1, num);
		if(pos == 0){
			break;
		}
	}
	
	return false;
}

char Solve::GetDifficulty()
{
	static const double coef[] = {
		-1885.598, 0.546, 11.604, 15.720, 1.528, 957.54, 
	};
	static const double thre[] = {
		300, 600, 1000, 1800, 
	};
	
	double d = coef[0] + 
		coef[1] * (double)tryCount[0] + 
		coef[2] * (double)tryCount[1] + 
		coef[3] * (double)tryCount[2] + 
		coef[4] * (double)tryCount[5] + 
		coef[5] * (double)tryCount[0] / (double)tryCount[5];
	
	for(char i = 0; i < 4; i++){
		if(d < thre[i]){
			return i + 1;
		}
	}
	return 5;
}

bool SolveProblem(StageInfo &info, bool dumpMode)
{
	Stage stage(info);
	Solve solve;
	
	bool result = solve.Process(stage, !dumpMode);
	
	if(result){
		if(dumpMode){
			int *d = solve.GetCount();
			printf("%d,%d,%d,%d,%d,%d\n", d[0], d[1], d[2], d[3], d[4], d[5]);
		}else{
			printf("finished.\n");
		}
		return true;
	}else{
		printf("failed.\n");
		return false;
	}
}
