
#include "crosssum_creator.h"

bool DataReader::ParseVal(char *val)
{
	int start = pos;
	while(data[pos] != '/'){
		if(data[pos] == '\0'){
			return false;
		}
		pos++;
	}
	
	data[pos++] = '\0';
	*val = (char)atoi(&data[start]);
	
	return true;
}

bool DataReader::ParseData(char *problem, char *hint, int width, int height)
{
	static const char LUT[] = {
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
		 0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0, 
		 0, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 
		35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45,  0,  0,  0,  0,  0, 
		 0, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,  0,  0,  0,  0,  0, 
		 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	};
	
	memset(hint, 0, width * height * 2);
	
	int skip = 0;
	for(int y = 1; y < height; y++){
		for(int x = 1; x < width; x++){
			int p = y * width + x;
			
			if(skip > 0){
				problem[p] = DIGIT_0;
				--skip;
			}else{
				char d = data[pos++];
				
				if(d == '\0' || d > 'z'){
					return false;
				}else if(d >= 'k'){
					problem[p] = DIGIT_0;
					skip = d - 'k';
				}else if(d == '.'){
					problem[p] = BLOCK;
				}else{
					problem[p] = HINT;
					hint[p * 2 + 1] = LUT[d];
					
					d = data[pos++];
					if(d == '\0' || d > 'z'){
						return false;
					}
					hint[p * 2] = LUT[d];
				}
			}
		}
	}
	
	problem[0] = BLOCK;
	for(int x = 1; x < width; x++){
		if(problem[width + x] == DIGIT_0){
			problem[x] = HINT;
			
			char d = data[pos++];
			if(d == '\0' || d > 'z'){
				return false;
			}
			hint[x * 2 + 1] = LUT[d];
		}else{
			problem[x] = BLOCK;
		}
	}
	for(int y = 1; y < height; y++){
		int p = y * width;
		
		if(problem[p + 1] == DIGIT_0){
			problem[p] = HINT;
			
			char d = data[pos++];
			if(d == '\0' || d > 'z'){
				return false;
			}
			hint[p * 2] = LUT[d];
		}else{
			problem[p] = BLOCK;
		}
	}
	
	return true;
}


bool StageInfo::ParseData(char *data)
{
	DataReader reader(data);
	
	if(!reader.ParseVal(&level)){
		return false;
	}
	if(!reader.ParseVal(&width)){
		return false;
	}
	if(!reader.ParseVal(&height)){
		return false;
	}
	
	++width;
	++height;
	
	delete[] problem;
	delete[] hint;
	problem = new char[(int)width * height];
	hint = new char[(int)width * height * 2];
	
	if(!reader.ParseData(problem, hint, width, height)){
		return false;
	}
	
	return true;
}

Stage::Stage(StageInfo &info)
{
	width = info.width;
	height = info.height;
	error = false;
	
	status = new word[(int)width * height];
	hint = new HintData[(int)width * height];
	
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int pos = y * info.width + x;
			status[pos] = 0x1ff;
		}
	}
	
	hintSize = 0;
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int pos = y * info.width + x;
			char p = info.problem[pos];
			
			if(p == BLOCK){
				status[pos] = STG_BLK;
			}else if(p == HINT){
				status[pos] = STG_BLK;
				
				char yokoLen = 0;
				char tateLen = 0;
				for(int i = pos + 1; 
					i < (int)width * height && info.problem[i] == DIGIT_0; 
					i++){
					yokoLen++;
				}
				for(int i = pos + width; 
					i < (int)width * height && info.problem[i] == DIGIT_0; 
					i += width){
					tateLen++;
				}
				
				char yokoSum = info.hint[pos * 2];
				char tateSum = info.hint[pos * 2 + 1];
				
				if(yokoLen == 0){
				}else if(yokoLen == 1){
					status[pos + 1] = (1 << (yokoSum - 1));
				}else{
					AddHintX(pos, yokoLen, yokoSum);
				}
				
				if(tateLen == 0){
				}else if(tateLen == 1){
					status[pos + width] = (1 << (tateSum - 1));
				}else{
					AddHintY(pos, tateLen, tateSum);
				}
			}
		}
	}
}

Stage::Stage(const Stage &stage)
{
	width = stage.width;
	height = stage.height;
	error = stage.error;
	hintSize = stage.hintSize;
	
	status = new word[(int)width * height];
	memcpy(status, stage.status, (int)width * height * sizeof(word));
	
	hint = new HintData[(int)width * height];
	memcpy(hint, stage.hint, (int)hintSize * sizeof(HintData));
}

Stage::Stage(char width_, char height_)
{
	width = width_;
	height = height_;
	error = false;
	hintSize = 0;
	
	status = new word[(int)width * height];
	hint = new HintData[(int)width * height];
}

void Stage::Reset(const Stage &stage)
{
	error = stage.error;
	hintSize = stage.hintSize;
	
	memcpy(status, stage.status, (int)width * height * sizeof(word));
	memcpy(hint, stage.hint, hintSize * sizeof(HintData));
}

void Stage::Reset()
{
	error = false;
	hintSize = 0;
}

int Stage::AddHintX(int pos, char len, char sum)
{
	hint[hintSize].len = len;
	hint[hintSize].sum = sum;
	
	for(int i = 0; i < len; i++){
		hint[hintSize].pos[i] = (word)(pos + i + 1);
	}
	
	int ret = hintSize;
	hintSize++;
	
	return ret;
}

int Stage::AddHintY(int pos, char len, char sum)
{
	hint[hintSize].len = len;
	hint[hintSize].sum = sum;
	
	for(int i = 0; i < len; i++){
		hint[hintSize].pos[i] = (word)(pos + (i + 1) * width);
	}
	
	int ret = hintSize;
	hintSize++;
	
	return ret;
}

void Stage::SetSum(int pos, char sum)
{
	hint[pos].sum = sum;
}

bool Stage::UpdateHintBlock(HintData &h, int *updated)
{
	for(int i = 0; i < h.len; i++){
		word p1 = h.pos[i];
		word s1 = status[p1];
		int count = GetCandidateCount(s1);
		
		if(count == 1){
			for(int j = 0; j < h.len; j++){
				if(j != i){
					word prev = status[h.pos[j]];
					word next = prev & ~s1;
					
					if(prev != next){
						if(GetCandidateCount(next) == 1){
							(*updated)++;
						}
						if(next == 0){
							return false;
						}
					}
				}
			}
		}else if(h.len > 2 && count == 2){
			for(int j = i + 1; j < h.len; j++){
				word p2 = h.pos[j];
				if(s1 == status[p2]){
					for(int k = 0; k < h.len; k++){
						if(k != i && k != j){
							word prev = status[h.pos[k]];
							word next = prev & ~s1;
							
							if(prev != next){
								if(GetCandidateCount(next) == 1){
									(*updated)++;
								}
								if(next == 0){
									return false;
								}
							}
						}
					}
				}
			}
		}
	}
	
	return true;
}

bool Stage::UpdateHint0(HintData &h)
{
	bool ret = false;
	int fixed = 0;
	
	for(int i = 0; i < h.len; i++){
		word p1 = h.pos[i];
		word s1 = status[p1];
		int count = GetCandidateCount(s1);
		
		if(count == 1){
			for(int j = 0; j < h.len; j++){
				if(j != i){
					word prev = status[h.pos[j]];
					status[h.pos[j]] &= ~s1;
					if(status[h.pos[j]] == 0){
						error = true;
					}
					if(prev != status[h.pos[j]]){
						ret = true;
					}
				}
			}
			fixed++;
		}else if(h.len > 2 && count == 2){
			for(int j = i + 1; j < h.len; j++){
				word p2 = h.pos[j];
				if(s1 == status[p2]){
					for(int k = 0; k < h.len; k++){
						if(k != i && k != j){
							word prev = status[h.pos[k]];
							status[h.pos[k]] &= ~s1;
							if(status[h.pos[k]] == 0){
								error = true;
							}
							if(prev != status[h.pos[k]]){
								ret = true;
							}
						}
					}
				}
			}
		}
	}
	
	if(fixed == h.len){
		word num = 0;
		for(int i = 0; i < h.len; i++){
			num |= status[h.pos[i]];
		}
		h.sum = GetSum(num);
		h.len = 0;
	}
	
	return ret;
}

bool Stage::UpdateHint(HintData &h)
{
	if(h.sum == 0){
		return UpdateHint0(h);
	}
	
	for(int i = 0; i < h.len; i++){
		word p1 = h.pos[i];
		word s1 = status[p1];
		int count = GetCandidateCount(s1);
		
		if(count == 1){
			if(h.len == 2){
				if(h.sum < 3 || h.sum > 17){
					error = true;
					return false;
				}
				
				word p2 = (i == 0) ? h.pos[1] : h.pos[0];
				status[p2] &= UPDATE_UNFIXED2_LUT[h.sum - 3][s1];
				h.len = 0;
				if(status[p2] == 0){
					error = true;
					return false;
				}
			}else{
				char sum = GetSum(s1);
				if(h.sum <= sum){
					error = true;
					return false;
				}
				
				for(int j = i + 1; j < h.len; j++){
					h.pos[j - 1] = h.pos[j];
				}
				h.len--;
				h.sum -= sum;
				
				for(int j = 0; j < h.len; j++){
					status[h.pos[j]] &= ~s1;
					if(status[h.pos[j]] == 0){
						error = true;
						return false;
					}
				}
			}
			
			return true;
		}else if(h.len > 2 && count == 2){
			for(int j = i + 1; j < h.len; j++){
				word p2 = h.pos[j];
				if(s1 == status[p2]){
					char sum = GetSum(s1);
					if(h.sum <= sum){
						error = true;
						return false;
					}
					
					if(h.len == 3){
						word p3;
						if(i == 0 && j == 1){
							p3 = h.pos[2];
						}else if(i == 0 && j == 2){
							p3 = h.pos[1];
						}else{
							p3 = h.pos[0];
						}
						
						status[p3] = (1 << (h.sum - sum - 1));
						status[p3] &= ~s1;
						if(status[p3] == 0 || status[p3] >= 0x200){
							error = true;
							return false;
						}
						
						h.len = 2;
						h.sum = sum;
						h.pos[0] = p1;
						h.pos[1] = p2;
					}else{
						hint[hintSize].len = 2;
						hint[hintSize].sum = sum;
						hint[hintSize].pos[0] = p1;
						hint[hintSize].pos[1] = p2;
						hintSize++;
						
						for(int k = i + 1; k < j; k++){
							h.pos[k - 1] = h.pos[k];
						}
						for(int k = j + 1; k < h.len; k++){
							h.pos[k - 2] = h.pos[k];
						}
						h.len -= 2;
						h.sum -= sum;
						
						for(int k = 0; k < h.len; k++){
							status[h.pos[k]] &= ~s1;
							if(status[h.pos[k]] == 0){
								error = true;
								return false;
							}
						}
					}
					
					return true;
				}
			}
		}
	}
	
	return false;
}

void Stage::PrepareCombination3()
{
	for(char sum = 6; sum <= 24; sum++){
		for(word s0 = 0; s0 < 0x200; s0++){
			for(word s1 = 0; s1 < 0x200; s1++){
				word r = 0;
				word f0 = s0;
				for(word b0 = f0 & (-f0); b0 != 0; b0 = f0 & (-f0)){
					f0 -= b0;
					
					char sum2 = sum - GetSum(b0);
					if(sum2 >= 3 && sum2 <= 17){
						r |= UPDATE_UNFIXED2_LUT[sum2 - 3][s1 & ~b0] & ~b0;
					}
				}
				
				UPDATE_UNFIXED3_LUT[sum - 6][(s0 << 9) | s1] = r;
			}
		}
	}
}

static bool Combination3(word *status, word *pos, word *result, word f0, 
	char sum)
{
	sum -= Stage::GetSum(f0);
	if(sum < 6 || sum > 24){
		return false;
	}
	
	word s0 = status[pos[0]] & ~f0;
	word s1 = status[pos[1]] & ~f0;
	word s2 = status[pos[2]] & ~f0;
	
	s0 &= Stage::UPDATE_UNFIXED3_LUT[sum - 6][(s1 << 9) | s2];
	s1 &= Stage::UPDATE_UNFIXED3_LUT[sum - 6][(s2 << 9) | s0];
	s2 &= Stage::UPDATE_UNFIXED3_LUT[sum - 6][(s0 << 9) | s1];
	
	if(s2 == 0){
		return false;
	}
	
	result[0] |= s0;
	result[1] |= s1;
	result[2] |= s2;
	
	return true;
}

static bool Combination4(word *status, word *pos, word *result, word f0, 
	char sum)
{
	bool ret = false;
	
	word s0 = status[pos[0]] & ~f0;
	for(word b0 = s0 & (-s0); b0 != 0; b0 = s0 & (-s0)){
		s0 -= b0;
		
		word f1 = f0 | b0;
		if(Combination3(status, &pos[1], &result[1], f1, sum)){
			result[0] |= b0;
			ret = true;
		}
	}
	
	return ret;
}

static bool Combination5(word *status, word *pos, word *result, word f0, 
	char sum)
{
	bool ret = false;
	
	word s0 = status[pos[0]] & ~f0;
	for(word b0 = s0 & (-s0); b0 != 0; b0 = s0 & (-s0)){
		s0 -= b0;
		
		word f1 = f0 | b0;
		word s1 = status[pos[1]] & ~f1;
		for(word b1 = s1 & (-s1); b1 != 0; b1 = s1 & (-s1)){
			s1 -= b1;
			
			word f2 = f1 | b1;
			if(Combination3(status, &pos[2], &result[2], f2, sum)){
				result[0] |= b0;
				result[1] |= b1;
				ret = true;
			}
		}
	}
	
	return ret;
}

static bool Combination6(word *status, word *pos, word *result, word f0, 
	char sum)
{
	bool ret = false;
	
	word s0 = status[pos[0]] & ~f0;
	for(word b0 = s0 & (-s0); b0 != 0; b0 = s0 & (-s0)){
		s0 -= b0;
		
		word f1 = f0 | b0;
		word s1 = status[pos[1]] & ~f1;
		for(word b1 = s1 & (-s1); b1 != 0; b1 = s1 & (-s1)){
			s1 -= b1;
			
			word f2 = f1 | b1;
			word s2 = status[pos[2]] & ~f2;
			for(word b2 = s2 & (-s2); b2 != 0; b2 = s2 & (-s2)){
				s2 -= b2;
				
				word f3 = f2 | b2;
				if(Combination3(status, &pos[3], &result[3], f3, sum)){
					result[0] |= b0;
					result[1] |= b1;
					result[2] |= b2;
					ret = true;
				}
			}
		}
	}
	
	return ret;
}

static bool Combination7(word *status, word *pos, word *result, word f0, 
	char sum)
{
	bool ret = false;
	
	word s0 = status[pos[0]] & ~f0;
	for(word b0 = s0 & (-s0); b0 != 0; b0 = s0 & (-s0)){
		s0 -= b0;
		
		word f1 = f0 | b0;
		if(Combination6(status, &pos[1], &result[1], f1, sum)){
			result[0] |= b0;
			ret = true;
		}
	}
	
	return ret;
}

static bool Combination8(word *status, word *pos, word *result, word f0, 
	char sum)
{
	bool ret = false;
	
	word s0 = status[pos[0]] & ~f0;
	for(word b0 = s0 & (-s0); b0 != 0; b0 = s0 & (-s0)){
		s0 -= b0;
		
		word f1 = f0 | b0;
		word s1 = status[pos[1]] & ~f1;
		for(word b1 = s1 & (-s1); b1 != 0; b1 = s1 & (-s1)){
			s1 -= b1;
			
			word f2 = f1 | b1;
			if(Combination6(status, &pos[2], &result[2], f2, sum)){
				result[0] |= b0;
				result[1] |= b1;
				ret = true;
			}
		}
	}
	
	return ret;
}

static bool Combination9(word *status, word *pos, word *result, word f0, 
	char sum)
{
	bool ret = false;
	
	word s0 = status[pos[0]] & ~f0;
	for(word b0 = s0 & (-s0); b0 != 0; b0 = s0 & (-s0)){
		s0 -= b0;
		
		word f1 = f0 | b0;
		word s1 = status[pos[1]] & ~f1;
		for(word b1 = s1 & (-s1); b1 != 0; b1 = s1 & (-s1)){
			s1 -= b1;
			
			word f2 = f1 | b1;
			word s2 = status[pos[2]] & ~f2;
			for(word b2 = s2 & (-s2); b2 != 0; b2 = s2 & (-s2)){
				s2 -= b2;
				
				word f3 = f2 | b2;
				if(Combination6(status, &pos[3], &result[3], f3, sum)){
					result[0] |= b0;
					result[1] |= b1;
					result[2] |= b2;
					ret = true;
				}
			}
		}
	}
	
	return ret;
}

typedef bool (*CombinationFunc)(word* status, word* pos, word* result, 
	word f0, char sum);
static const CombinationFunc COMB_FUNC[6] = {
	Combination4, Combination5, Combination6, Combination7, Combination8, 
	Combination9, 
};

int Stage::UpdateCombination()
{
	int updated = 0;
	
	for(int t = 0; t < hintSize; t++){
		HintData &h = hint[t];
		
		if(h.len < 4 || h.sum == 0){
			continue;
		}
		
		word result[9] = {};
		if(!COMB_FUNC[h.len - 4](status, h.pos, result, 0, h.sum)){
			error = true;
			return 0;
		}
		
		for(int i = 0; i < h.len; i++){
			if(status[h.pos[i]] != result[i]){
				if(result[i] == 0){
					error = true;
					return 0;
				}
				
				status[h.pos[i]] = result[i];
				updated++;
			}
		}
	}
	
	return updated;
}

int Stage::UpdateAll()
{
	int updated = 0;
	
	for(int t = 0; t < hintSize; t++){
		HintData &h = hint[t];
		
		if(h.len >= 4){
			if(h.sum != 0){
				word s = GetPos(h.len, h.sum);
				if(s == 0){
					error = true;
					return 0;
				}
				
				word used = 0;
				word unfixedNum = 0x1ff;
				for(int i = 0; i < h.len; i++){
					int p1 = h.pos[i];
					used |= status[p1];
					unfixedNum &= UPDATE_UNFIXED_LUT[s - 1][status[p1]];
				}
				
				word usedNum = UPDATE_SUM_LUT[s - 1][used];
				
				for(int i = 0; i < h.len; i++){
					int p1 = h.pos[i];
					word prev = status[p1];
					
					status[p1] &= usedNum;
					status[p1] &= unfixedNum;
					
					if(prev != status[p1]){
						updated++;
						if(status[p1] == 0){
							error = true;
							return 0;
						}
					}
				}
			}
			
			while(UpdateHint(h)){
				updated++;
			}
			if(error){
				return 0;
			}
		}else if(h.len == 3){
			if(h.sum >= 25 || h.sum <= 5){
				if(h.sum != 0){
					error = true;
					return 0;
				}
			}else{
				int p1 = h.pos[0];
				int p2 = h.pos[1];
				int p3 = h.pos[2];
				
				word num = UPDATE_UNFIXED3_LUT
					[h.sum - 6][(status[p1] << 9) | status[p2]];
				word prev = status[p3];
				status[p3] = (prev & num);
				if(prev != status[p3]){
					updated++;
					if(status[p3] == 0){
						error = true;
						return 0;
					}
				}
				
				num = UPDATE_UNFIXED3_LUT
					[h.sum - 6][(status[p2] << 9) | status[p3]];
				prev = status[p1];
				status[p1] = (prev & num);
				if(prev != status[p1]){
					updated++;
					if(status[p1] == 0){
						error = true;
						return 0;
					}
				}
				
				num = UPDATE_UNFIXED3_LUT
					[h.sum - 6][(status[p3] << 9) | status[p1]];
				prev = status[p2];
				status[p2] = (prev & num);
				if(prev != status[p2]){
					updated++;
					if(status[p2] == 0){
						error = true;
						return 0;
					}
				}
			}
			
			while(UpdateHint(h)){
				updated++;
			}
			if(error){
				return 0;
			}
		}else if(h.len == 2){
			if(h.sum >= 18 || h.sum <= 2){
				if(h.sum != 0){
					error = true;
					return 0;
				}
			}else{
				int p1 = h.pos[0];
				int p2 = h.pos[1];
				
				word num = UPDATE_UNFIXED2_LUT[h.sum - 3][status[p1]];
				word prev = status[p2];
				status[p2] = (prev & num);
				if(prev != status[p2]){
					updated++;
					if(status[p2] == 0){
						error = true;
						return 0;
					}
				}
				
				num = UPDATE_UNFIXED2_LUT[h.sum - 3][status[p2]];
				prev = status[p1];
				status[p1] = (prev & num);
				if(prev != status[p1]){
					updated++;
					if(status[p1] == 0){
						error = true;
						return 0;
					}
				}
			}
			
			while(UpdateHint(h)){
				updated++;
			}
			if(error){
				return 0;
			}
		}
	}
	
	return updated;
}

bool Stage::IsFixedSum(int hintPos)
{
	return 
		(hintPos == -1 || hint[hintPos].len == 0 || hint[hintPos].sum != 0);
}

bool Stage::UpdateAll2Block(int hintPos, int sum, int count, int *updated)
{
	HintData &h = hint[hintPos];
	if(h.len == 0){
		return false;
	}
	
	word s = GetPos(h.len, (char)sum);
	if(s == 0){
		return false;
	}
	
	if(h.len >= 4){
		word used = 0;
		word unfixedNum = 0x1ff;
		for(int i = 0; i < h.len; i++){
			int p1 = h.pos[i];
			used |= status[p1];
			unfixedNum &= UPDATE_UNFIXED_LUT[s - 1][status[p1]];
		}
		
		word usedNum = UPDATE_SUM_LUT[s - 1][used];
		
		for(int i = 0; i < h.len; i++){
			int p1 = h.pos[i];
			word prev = status[p1];
			
			word next = (status[p1] & usedNum);
			next &= unfixedNum;
			
			if(prev != next){
				if(GetCandidateCount(next) <= count){
					(*updated)++;
				}
				if(next == 0){
					return false;
				}
			}
		}
		
		if(!UpdateHintBlock(h, updated)){
			return false;
		}
	}else if(h.len == 3){
		if(h.sum >= 25){
			return false;
		}
		
		int p1 = h.pos[0];
		int p2 = h.pos[1];
		int p3 = h.pos[2];
		
		word num = 
			UPDATE_UNFIXED3_LUT[sum - 6][(status[p1] << 9) | status[p2]];
		word prev = status[p3];
		word next = (prev & num);
		if(prev != next){
			if(GetCandidateCount(next) <= count){
				(*updated)++;
			}
			if(next == 0){
				return false;
			}
		}
		
		num = UPDATE_UNFIXED3_LUT[sum - 6][(status[p2] << 9) | status[p3]];
		prev = status[p1];
		next = (prev & num);
		if(prev != next){
			if(GetCandidateCount(next) <= count){
				(*updated)++;
			}
			if(next == 0){
				return false;
			}
		}
		
		num = UPDATE_UNFIXED3_LUT[sum - 6][(status[p3] << 9) | status[p1]];
		prev = status[p2];
		next = (prev & num);
		if(prev != next){
			if(GetCandidateCount(next) <= count){
				(*updated)++;
			}
			if(next == 0){
				return false;
			}
		}
	}else if(h.len == 2){
		if(sum >= 18){
			return false;
		}
		
		int p1 = h.pos[0];
		int p2 = h.pos[1];
		
		word num = UPDATE_UNFIXED2_LUT[sum - 3][status[p1]];
		word prev = status[p2];
		word next = (prev & num);
		if(prev != next){
			if(GetCandidateCount(next) <= count){
				(*updated)++;
			}
			if(next == 0){
				return false;
			}
		}
		
		num = UPDATE_UNFIXED2_LUT[sum - 3][status[p2]];
		prev = status[p1];
		next = (prev & num);
		if(prev != next){
			if(GetCandidateCount(next) <= count){
				(*updated)++;
			}
			if(next == 0){
				return false;
			}
		}
	}
	
	return true;
}

bool Stage::IsSolved()
{
	if(error){
		return false;
	}
	
	for(int t = 0; t < hintSize; t++){
		HintData &h = hint[t];
		if(h.len != 0){
			return false;
		}
	}
	
	return true;
}

int Stage::LeftCount()
{
	int count = 0;
	
	for(int t = 0; t < hintSize; t++){
		HintData &h = hint[t];
		
		if(h.len > 0 && h.sum == 0){
			int c1 = 0;
			for(int i = 0; i < h.len; i++){
				int p1 = h.pos[i];
				
				c1 += GetCandidateCount(status[p1]);
			}
			count += c1 - h.len;
		}
	}
	
	return count;
}

int Stage::GetNextTryPos(int pos, word *num)
{
	for(int t = 0; t < hintSize; t++){
		HintData &h = hint[t];
		
		if(h.len > 0){
			if(h.pos[0] >= pos){
				for(int i = 0; i < h.len; i++){
					int p1 = h.pos[i];
					if(GetCandidateCount(status[p1]) == 2){
						num[0] = status[p1] & -status[p1];
						num[1] = status[p1] & ~num[0];
						
						return p1;
					}
				}
			}else if(h.pos[h.len - 1] >= pos){
				for(int i = 0; i < h.len; i++){
					int p1 = h.pos[i];
					if(p1 >= pos && GetCandidateCount(status[p1]) == 2){
						num[0] = status[p1] & -status[p1];
						num[1] = status[p1] & ~num[0];
						
						return p1;
					}
				}
			}
		}
	}
	
	return 0;
}

void Stage::SetTry(int pos, word num)
{
	status[pos] = num;
}

