
class DataReader
{
public:
	DataReader(char *data)
	{
		this->data = data;
		pos = 0;
	}
	
	bool ParseVal(char *val);
	bool ParseData(char *problem, char *hint, int width, int height);

private:
	char *data;
	int pos;
};

class StageInfo
{
public:
	char width;
	char height;
	char level;
	char *problem;
	char *hint;
	
	StageInfo()
	{
		problem = NULL;
		hint = NULL;
	}
	
	~StageInfo()
	{
		delete[] problem;
		delete[] hint;
	}
	
	bool ParseData(char *data);
};

struct HintData
{
	char sum;
	char len;
	word pos[9];
};

class Stage
{
public:
	// UpdateCombination()len=2LUT
	static const word UPDATE_UNFIXED2_LUT[15][0x200];
	// UpdateCombination()len=3LUT
	static word UPDATE_UNFIXED3_LUT[19][0x40000];
	
	char width;
	char height;
	bool error;
	word *status;
	HintData *hint;
	int hintSize;
	
	Stage(StageInfo &info);
	Stage(const Stage &stage);
	Stage(char width_, char height_);
	
	~Stage()
	{
		delete[] status;
		delete[] hint;
	}
	
	void Reset(const Stage &stage);
	void Reset();
	int AddHintX(int pos, char len, char sum);
	int AddHintY(int pos, char len, char sum);
	void SetSum(int pos, char sum);
	
	int UpdateAll();
	bool UpdateAll2Block(int hintPos, int sum, int count, int *updated);
	int UpdateCombination();
	int GetNextTryPos(int pos, word *num);
	void SetTry(int pos, word num);
	bool IsSolved();
	int LeftCount();
	bool IsFixedSum(int hintPos);
	bool UpdateHint0(HintData &h);
	bool UpdateHint(HintData &h);
	bool UpdateHintBlock(HintData &h, int *updated);
	
	static void PrepareCombination3();
	
	static char GetCandidateCount(word s)
	{
		return BIT_COUNT[s];
	}
	
	static char GetSum(word s)
	{
		return BIT_SUM[s];
	}

private:
	static const char BIT_COUNT[0x200];
	static const char BIT_SUM[0x200];
	static const byte POS[9 * 46];
	
	// ́FׂẴZORAFgp\ȔzORAND
	//   for each cell: or cell -> 1
	//   for each cell: cell or LUT[1] -> cell
	static const word UPDATE_SUM_LUT[139][0x200];
	
	// ́FFixĂZOR
	// FFixĂȂZɑ΂Agp\ȔzORAND
	//   for each fix cell: or -> 1
	//   for each not fix cell: cell and LUT[1] -> cell
	static const word UPDATE_FIXED_LUT[139][0x200];
	
	// ́FeZ̒l
	// F̑̃Zɑ΂Agp\ȔzORAND
	//   for each cell: and LUT[cell] -> 1
	//   for each cell: cell and 1 -> cell
	static const word UPDATE_UNFIXED_LUT[139][0x200];
	
	static word GetPos(char len, char sum)
	{
		return POS[(len - 1) * 46 + sum];
	}
};

