
#include "lightup_creator.h"

void Problem::Initialize(Rand &rand_, char width_, char height_)
{
	rand = &rand_;
	
	width = width_;
	height = height_;
	
	answer = new Stage(width, height);
}

void Problem::Reset()
{
	delete answer;
	answer = new Stage(width, height);
	
	delete stage;
	stage = NULL;
}

void Problem::SetBlock(int num)
{
	for(int i = 0; i < num; i++){
		int x = rand->Next(width) + 1;
		int y = rand->Next(height) + 1;
		int pos = y * answer->width + x;
		
		if(answer->status[pos] == AISLE){
			answer->status[pos] = BLOCK;
		}
	}
	
	DumpStage(*answer);
}

void Problem::SearchOrphan()
{
	for(int y = 1; y < answer->height - 1; y++){
		for(int x = 1; x < answer->width - 1; x++){
			int pos = y * answer->width + x;
			
			if(answer->status[pos] == AISLE){
				int block = answer->CountBlock(pos);
				if(block == 4){
					int r = rand->Next(4);
					if(r == 0){
						answer->status[(x == 1) ? pos + 1 : pos - 1] = AISLE;
					}else if(r == 1){
						answer->status[(x == answer->width - 2) ? pos - 1 : pos + 1] = AISLE;
					}else if(r == 2){
						answer->status[(y == 1) ? pos + answer->width : pos - answer->width] = AISLE;
					}else{
						answer->status[(y == answer->height - 2) ? pos - answer->width : pos + answer->width] = AISLE;
					}
				}
			}else if(answer->status[pos] == BLOCK && x != 1 && 
				x != answer->width - 2 && y != 1 && 
				y != answer->height - 2){
				int block = answer->CountBlock(pos);
				if(block >= 4){
					int r;
					if(block == 4){
						r = rand->Next(16);
					}else if(block == 5){
						r = rand->Next(12);
					}else{
						r = rand->Next(8);
					}
					if(r == 0){
						answer->status[pos - 1 - answer->width] = AISLE;
					}else if(r == 1){
						answer->status[pos - 1] = AISLE;
					}else if(r == 2){
						answer->status[pos - 1 + answer->width] = AISLE;
					}else if(r == 3){
						answer->status[pos - answer->width] = AISLE;
					}else if(r == 4){
						answer->status[pos + answer->width] = AISLE;
					}else if(r == 5){
						answer->status[pos + 1 - answer->width] = AISLE;
					}else if(r == 6){
						answer->status[pos + 1] = AISLE;
					}else if(r == 7){
						answer->status[pos + 1 + answer->width] = AISLE;
					}
				}
			}
		}
	}
	
	DumpStage(*answer);
	stage = new Stage(*answer);
}

void Problem::SetLight(int num)
{
	for(int i = 0; i < num; i++){
		int x = rand->Next(width) + 1;
		int y = rand->Next(height) + 1;
		int pos = y * answer->width + x;
		
		if(answer->status[pos] == AISLE){
			answer->SetLight(pos);
		}
	}
	
	DumpStage(*answer);
}

void Problem::SearchAisle()
{
	for(int pos = 0; pos < answer->width * answer->height; pos++){
		if(answer->status[pos] == AISLE){
			answer->SetLight(pos);
		}
	}
	
	DumpStage(*answer);
}

bool Problem::SetHint()
{
	for(int y = 1; y < answer->height - 1; y++){
		for(int x = 1; x < answer->width - 1; x++){
			int pos = y * answer->width + x;
			
			if(answer->status[pos] == BLOCK){
				stage->status[pos] = 
					(char)(BLOCK + answer->CountLight(pos) + 1);
			}
		}
	}
	
	DumpStage(*stage);
	
	Solve solve;
	Stage solveStage(*stage);
	if(!solve.ProcessFast(solveStage)){
		return false;
	}
	
	return true;
}

void Problem::RemoveHint(int *ratio)
{
	int blockCount = 0;
	
	for(int y = 1; y < answer->height - 1; y++){
		for(int x = 1; x < answer->width - 1; x++){
			int pos = y * answer->width + x;
			
			if(answer->status[pos] == BLOCK){
				blockCount++;
			}
		}
	}
	
	int removeCount = 0;
	
	for(int y = 1; y < answer->height - 1; y++){
		for(int x = 1; x < answer->width - 1; x++){
			int pos = y * answer->width + x;
			
			if(stage->status[pos] == HINT_0 || stage->status[pos] == HINT_4){
				char prev = stage->status[pos];
				stage->status[pos] = BLOCK;
				
				Solve solve;
				Stage solveStage(*stage);
				if(solve.ProcessSimple(solveStage)){
					removeCount++;
				}else{
					stage->status[pos] = prev;
				}
			}
		}
	}
	
	for(int y = 1; y < answer->height - 1; y++){
		for(int x = 1; x < answer->width - 1; x++){
			int pos = y * answer->width + x;
			
			if(stage->status[pos] >= HINT_1 && stage->status[pos] <= HINT_3){
				char prev = stage->status[pos];
				stage->status[pos] = BLOCK;
				
				Solve solve;
				Stage solveStage(*stage);
				if(solve.ProcessSimple(solveStage)){
					removeCount++;
				}else{
					stage->status[pos] = prev;
				}
			}
		}
	}
	
	*ratio = (blockCount - removeCount) * 100 / blockCount;
	DumpStage(*stage);
}

bool Problem::GetProblem(StageInfo &info)
{
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int pos = (y + 1) * stage->width + (x + 1);
			
			info.problem[y * width + x] = stage->status[pos];
		}
	}
	
	Solve solve;
	Stage solveStage(*stage);
	if(solve.Process(solveStage, false)){
		info.level = solve.GetDifficulty();
		return true;
	}else{
		return false;
	}
}

void Problem::Create(StageInfo &info)
{
	int blockRatio = 30;
	int lightRatio = 30;
	
	int blockNum = width * height * blockRatio / 100;
	int lightNum = width * height * lightRatio / 100;
	int hintRatio;
	
	for( ; ; ){
		Reset();
		SetBlock(blockNum);
		SearchOrphan();
		SetLight(lightNum);
		SearchAisle();
		
		if(SetHint()){
			RemoveHint(&hintRatio);
			if(GetProblem(info)){
				break;
			}
		}
	}
}

void CreateProblem(Rand &rand, char width, char height, StageInfo &info)
{
	info.width = width;
	info.height = height;
	info.problem = new char[width * height];
	
	Problem problem;
	problem.Initialize(rand, width, height);
	problem.Create(info);
}
