
#include "lightup_creator.h"

const char StageInfo::PARSE_STAGE_STATUS[80] = {
	HINT_0, HINT_1, HINT_2, HINT_3, HINT_4, PSERR, PSERR, PSERR, 
	PSERR, PSERR, PSERR, PSERR, PSERR, PSERR, PSERR, PSERR, 
	PSERR, PSERR, BLOCK, BLOCK, BLOCK, BLOCK, BLOCK, BLOCK, 
	BLOCK, BLOCK, BLOCK, BLOCK, BLOCK, BLOCK, BLOCK, BLOCK, 
	BLOCK, BLOCK, BLOCK, BLOCK, BLOCK, BLOCK, BLOCK, BLOCK, 
	BLOCK, BLOCK, BLOCK, PSERR, PSERR, PSERR, PSERR, PSERR, 
	PSERR, AISLE, AISLE, AISLE, AISLE, AISLE, AISLE, AISLE, 
	AISLE, AISLE, AISLE, AISLE, AISLE, AISLE, AISLE, AISLE, 
	AISLE, AISLE, AISLE, AISLE, AISLE, AISLE, AISLE, AISLE, 
	AISLE, AISLE, AISLE, PSERR, PSERR, PSERR, PSERR, PSERR, 
};

const int StageInfo::PARSE_STAGE_NUM[80] = {
	 1,  1,  1,  1,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 
	15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,  0,  0,  0,  0,  0, 
	 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 
	16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26,  0,  0,  0,  0,  0, 
};

bool DataReader::ParseVal(char *val)
{
	int start = pos;
	while(data[pos] != '/'){
		if(data[pos] == '\0'){
			return false;
		}
		pos++;
	}
	
	data[pos++] = '\0';
	*val = (char)atoi(&data[start]);
	
	return true;
}

bool DataReader::ParseData(char *problem, int size)
{
	int s = 0;
	while(s < size){
		if(data[pos] == '\0'){
			return false;
		}else if(data[pos] == '.'){
			problem[s++] = AISLE;
		}else if(data[pos] >= '0' && data[pos] <= '4'){
			problem[s++] = HINT_0 + data[pos] - '0';
		}else if(data[pos] == '5'){
			problem[s++] = BLOCK;
		}
		pos++;
	}
	return true;
}

bool StageInfo::ParseData(char *data)
{
	DataReader reader(data);
	
	if(!reader.ParseVal(&level)){
		return false;
	}
	if(!reader.ParseVal(&height)){
		return false;
	}
	if(!reader.ParseVal(&width)){
		return false;
	}
	
	problem = new char[width * height];
	if(!reader.ParseData(problem, width * height)){
		return false;
	}
	
	return true;
}

/*
bool StageInfo::ParseData(char *data)
{
	width = (data[0] - '0') * 10 + (data[1] - '0');
	height = (data[2] - '0') * 10 + (data[3] - '0');
	level = (data[4] - '0') * 10 + (data[5] - '0');
	
	if(width < 0 || width > 99 || height < 0 || height > 99 || 
		level < 1 || level > 5){
		return false;
	}
	
	problem = new char[width * height];
	memset(problem, AISLE, width * height);
	
	int d = 6;
	int pos = 0;
	while(data[d] >= '0' && data[d] <= 0x7f && pos < width * height){
		char p = data[d++] - '0';
		
		int e = pos + PARSE_STAGE_NUM[p];
		while(pos < e && pos < width * height){
			problem[pos++] = PARSE_STAGE_STATUS[p];
		}
	}
	
	return true;
}
*/

const char Stage::CROSS_COUNT[10] = {
	0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 
};

const char Stage::LIGHT_COUNT[10] = {
	0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 
};

const char Stage::AISLE_COUNT[10] = {
	1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
};

const char Stage::BLOCK_COUNT[10] = {
	0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 
};

Stage::Stage(StageInfo &info)
{
	width = info.width + 2;
	height = info.height + 2;
	status = new char[width * height];
	
	memset(status, BLOCK, width * height);
	
	for(int y = 0; y < info.height; y++){
		for(int x = 0; x < info.width; x++){
			char p = info.problem[y * info.width + x];
			int pos = (y + 1) * width + (x + 1);
			
			status[pos] = p;
		}
	}
}

Stage::Stage(const Stage &stage)
{
	width = stage.width;
	height = stage.height;
	status = new char[width * height];
	
	memcpy(status, stage.status, width * height);
}

Stage::Stage(char width_, char height_)
{
	width = width_ + 2;
	height = height_ + 2;
	status = new char[width * height];
	
	memset(status, BLOCK, width * height);
	
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			status[pos] = AISLE;
		}
	}
}

void Stage::Reset(const Stage &stage)
{
	memcpy(status, stage.status, width * height);
}

void Stage::UpdateDiagonalHint3(int pos, int dx, int dy, bool *updated)
{
	if(status[pos + dx + dy] == HINT_1){
		if(status[pos + dx * 2 + dy] == AISLE){
			*updated = true;
			SetCross(pos + dx * 2 + dy);
		}
		if(status[pos + dx + dy * 2] == AISLE){
			*updated = true;
			SetCross(pos + dx + dy * 2);
		}
		if(status[pos - dx] == AISLE){
			*updated = true;
			SetLight(pos - dx);
		}
		if(status[pos - dy] == AISLE){
			*updated = true;
			SetLight(pos - dy);
		}
	}else if(status[pos + dx + dy] == HINT_2){
		if(status[pos + dx * 2 + dy] == LIGHT){
			if(status[pos + dx + dy * 2] == AISLE){
				 *updated = true;
				SetCross(pos + dx + dy * 2);
			}
			if(status[pos - dx] == AISLE){
				 *updated = true;
				SetLight(pos - dx);
			}
			if(status[pos - dy] == AISLE){
				 *updated = true;
				SetLight(pos - dy);
			}
		}else if(status[pos + dx + dy * 2] == LIGHT){
			if(status[pos + dx * 2 + dy] == AISLE){
				*updated = true;
				SetCross(pos + dx * 2 + dy);
			}
			if(status[pos - dx] == AISLE){
				*updated = true;
				SetLight(pos - dx);
			}
			if(status[pos - dy] == AISLE){
				*updated = true;
				SetLight(pos - dy);
			}
		}
		if(status[pos - dx] == LIGHT && status[pos - dy] == LIGHT){
			if(CROSS_COUNT[status[pos + dx * 2 + dy]] == 1 && status[pos + dx + dy * 2] == AISLE){
				*updated = true;
				SetLight(pos + dx + dy * 2);
			}else if(CROSS_COUNT[status[pos + dx + dy * 2]] == 1 && status[pos + dx * 2 + dy] == AISLE){
				*updated = true;
				SetLight(pos + dx * 2 + dy);
			}
		}
	}else if(status[pos + dx + dy] == HINT_3){
		if(status[pos - dy] == LIGHT && status[pos - dx] == LIGHT){
			if(status[pos + dx * 2 + dy] == AISLE){
				*updated = true;
				SetLight(pos + dx * 2 + dy);
			}
			if(status[pos + dx + dy * 2] == AISLE){
				*updated = true;
				SetLight(pos + dx + dy * 2);
			}
		}
	}
}

void Stage::UpdateDiagonalHint2(int pos, int dx, int dy, bool *updated)
{
	if(status[pos + dx + dy] == HINT_1){
		if(CROSS_COUNT[status[pos - dx]] == 1){
			if(status[pos - dy] == AISLE){
				*updated = true;
				SetLight(pos - dy);
			}
			if(status[pos + dx * 2 + dy] == AISLE){
				*updated = true;
				SetCross(pos + dx * 2 + dy);
			}
			if(status[pos + dx + dy * 2] == AISLE){
				*updated = true;
				SetCross(pos + dx + dy * 2);
			}
		}else if(CROSS_COUNT[status[pos - dy]] == 1){
			if(status[pos - dx] == AISLE){
				*updated = true;
				SetLight(pos - dx);
			}
			if(status[pos + dx * 2 + dy] == AISLE){
				*updated = true;
				SetCross(pos + dx * 2 + dy);
			}
			if(status[pos + dx + dy * 2] == AISLE){
				*updated = true;
				SetCross(pos + dx + dy * 2);
			}
		}
		if(CROSS_COUNT[status[pos + dx * 2 + dy]] == 1 && CROSS_COUNT[status[pos + dx + dy * 2]] == 1){
			if(status[pos - dy] == LIGHT && status[pos - dx] == AISLE){
				*updated = true;
				SetCross(pos - dx);
			}else if(status[pos - dx] == LIGHT && status[pos - dy] == AISLE){
				*updated = true;
				SetCross(pos - dy);
			}else if(CROSS_COUNT[status[pos - dy]] == 1 && status[pos - dx] == AISLE){
				*updated = true;
				SetLight(pos - dx);
			}else if(CROSS_COUNT[status[pos - dx]] == 1 && status[pos - dy] == AISLE){
				*updated = true;
				SetLight(pos - dy);
			}
		}
	}else if(status[pos + dx + dy] == HINT_2){
		if((CROSS_COUNT[status[pos + dx * 2 + dy]] == 1 && status[pos + dx + dy * 2] == LIGHT) || 
			(CROSS_COUNT[status[pos + dx + dy * 2]] == 1 && status[pos + dx * 2 + dy] == LIGHT)){
			if(status[pos - dy] == LIGHT && status[pos - dx] == AISLE){
				*updated = true;
				SetCross(pos - dx);
			}else if(status[pos - dx] == LIGHT && status[pos - dy] == AISLE){
				*updated = true;
				SetCross(pos - dy);
			}else if(CROSS_COUNT[status[pos - dy]] == 1 && status[pos - dx] == AISLE){
				*updated = true;
				SetLight(pos - dx);
			}else if(CROSS_COUNT[status[pos - dx]] == 1 && status[pos - dy] == AISLE){
				*updated = true;
				SetLight(pos - dy);
			}
		}
	}
}

void Stage::UpdateDiagonalHint1(int pos, int dx, int dy, bool *updated)
{
	if(status[pos + dx + dy] == HINT_1){
		if(CROSS_COUNT[status[pos - dx]] == 1 && CROSS_COUNT[status[pos - dy]] == 1){
			if(status[pos + dx * 2 + dy] == AISLE){
				*updated = true;
				SetCross(pos + dx * 2 + dy);
			}
			if(status[pos + dx + dy * 2] == AISLE){
				*updated = true;
				SetCross(pos + dx + dy * 2);
			}
		}
	}
}

void Stage::UpdateHint(int pos, bool *updated, bool *error)
{
	int hint = status[pos] - HINT_0;
	
	int light = CountLight(pos);
	if(hint == light){
		*updated = true;
		
		if(status[pos - 1] == AISLE) SetCross(pos - 1);
		if(status[pos + 1] == AISLE) SetCross(pos + 1);
		if(status[pos - width] == AISLE) SetCross(pos - width);
		if(status[pos + width] == AISLE) SetCross(pos + width);
		ClearHint(pos);
		return;
	}else if(hint < light){
		*error = true;
		return;
	}
	
	int cross = CountCross(pos);
	if(hint + cross == 4){
		*updated = true;
		
		if(status[pos - 1] == AISLE) SetLight(pos - 1);
		if(status[pos + 1] == AISLE) SetLight(pos + 1);
		if(status[pos - width] == AISLE) SetLight(pos - width);
		if(status[pos + width] == AISLE) SetLight(pos + width);
		ClearHint(pos);
	}else if(hint + cross == 3){
		if(status[pos - 1] == AISLE){
			if(status[pos - width] == AISLE && status[pos - width - 1] == AISLE){
				*updated = true;
				SetCross(pos - width - 1);
			}
			if(status[pos + width] == AISLE && status[pos + width - 1] == AISLE){
				*updated = true;
				SetCross(pos + width - 1);
			}
		}
		if(status[pos + 1] == AISLE){
			if(status[pos - width] == AISLE && status[pos - width + 1] == AISLE){
				*updated = true;
				SetCross(pos - width + 1);
			}
			if(status[pos + width] == AISLE && status[pos + width + 1] == AISLE){
				*updated = true;
				SetCross(pos + width + 1);
			}
		}
	}else if(hint + cross > 4){
		*error = true;
	}
	
	if(status[pos] == HINT_3){
		UpdateDiagonalHint3(pos, 1, width, updated);
		UpdateDiagonalHint3(pos, -1, width, updated);
		UpdateDiagonalHint3(pos, 1, -width, updated);
		UpdateDiagonalHint3(pos, -1, -width, updated);
	}else if(status[pos] == HINT_2){
		UpdateDiagonalHint2(pos, 1, width, updated);
		UpdateDiagonalHint2(pos, -1, width, updated);
		UpdateDiagonalHint2(pos, 1, -width, updated);
		UpdateDiagonalHint2(pos, -1, -width, updated);
	}else if(status[pos] == HINT_1){
		UpdateDiagonalHint1(pos, 1, width, updated);
		UpdateDiagonalHint1(pos, -1, width, updated);
		UpdateDiagonalHint1(pos, 1, -width, updated);
		UpdateDiagonalHint1(pos, -1, -width, updated);
	}
}

int Stage::CountAisle(int start, int diff, int num, int *found_pos)
{
	int count = 0;
	
	for(int pos = start + diff; status[pos] < BLOCK; pos += diff){
		if(status[pos] == AISLE){
			if(++count > num){
				return count;
			}
			*found_pos++ = pos;
		}
	}
	return count;
}

int Stage::CountAisle(int pos, int num, int *found_pos)
{
	int count = CountAisle(pos, 1, num, &found_pos[0]);
	if(count > num) return count;
	
	count += CountAisle(pos, -1, num - count, &found_pos[count]);
	if(count > num) return count;
	
	count += CountAisle(pos, width, num - count, &found_pos[count]);
	if(count > num) return count;
	
	count += CountAisle(pos, -width, num - count, &found_pos[count]);
	return count;
}

void Stage::UpdateAisle(int pos, bool *updated)
{
	if(CountAisle(pos, 0, NULL) == 0){
		*updated = true;
		
		SetLight(pos);
	}
}

void Stage::UpdateCross(int pos, bool *updated, bool *error)
{
	int found_pos;
	int count = CountAisle(pos, 1, &found_pos);
	
	if(count == 1){
		*updated = true;
		
		SetLight(found_pos);
	}else if(count == 0){
		*error = true;
	}
}

void Stage::Update(bool *updated, bool *error)
{
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			
			if(status[pos] >= HINT_0){
				UpdateHint(pos, updated, error);
			}else if(status[pos] == AISLE){
				UpdateAisle(pos, updated);
			}else if(status[pos] == CROSS){
				UpdateCross(pos, updated, error);
			}
		}
	}
}

int Stage::GetTryPos2(int prev_pos, int *try_pos, int *try_type)
{
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			
			if(pos <= prev_pos){
				continue;
			}
			
			if(status[pos] >= HINT_1){
				int aisle = CountAisle(pos);
				if(aisle == 2){
					if(status[pos - 1] == AISLE) *try_pos++ = pos - 1;
					if(status[pos + 1] == AISLE) *try_pos++ = pos + 1;
					if(status[pos - width] == AISLE) *try_pos++ = pos - width;
					if(status[pos + width] == AISLE) *try_pos++ = pos + width;
					
					*try_type = 0;
					return pos;
				}
			}else if(status[pos] == AISLE){
				int found_pos;
				int count = CountAisle(pos, 1, &found_pos);
				
				if(count == 1){
					*try_pos++ = pos;
					*try_pos++ = found_pos;
					
					*try_type = 1;
					return pos;
				}
			}else if(status[pos] == CROSS){
				int found_pos[2];
				int count = CountAisle(pos, 2, found_pos);
				
				if(count == 2){
					*try_pos++ = found_pos[0];
					*try_pos++ = found_pos[1];
					
					*try_type = 1;
					return pos;
				}
			}
		}
	}
	
	return -1;
}

int Stage::GetTryPos3(int prev_pos, int *try_pos, bool *light)
{
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			
			if(pos <= prev_pos){
				continue;
			}
			
			if(status[pos] >= HINT_1){
				int aisle = CountAisle(pos);
				if(aisle == 3){
					int hint = status[pos] - HINT_0;
					
					if(hint == 1){
						*light = true;
					}else if(hint == 2){
						*light = (CountLight(pos) == 1);
					}else{
						*light = false;
					}
					
					if(status[pos - 1] == AISLE) *try_pos++ = pos - 1;
					if(status[pos + 1] == AISLE) *try_pos++ = pos + 1;
					if(status[pos - width] == AISLE) *try_pos++ = pos - width;
					if(status[pos + width] == AISLE) *try_pos++ = pos + width;
					
					return pos;
				}
			}
		}
	}
	
	return -1;
}

void Stage::FindCross(Stage &stage1, Stage &stage2, bool *updated)
{
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			
			if(status[pos] == AISLE && stage1.status[pos] == TURN_ON && 
				stage2.status[pos] == TURN_ON){
				*updated = true;
				SetCross(pos);
			}
		}
	}
}

bool Stage::IsSolved()
{
	for(int y = 1; y < height - 1; y++){
		for(int x = 1; x < width - 1; x++){
			int pos = y * width + x;
			
			if(status[pos] == AISLE || status[pos] == CROSS){
				return false;
			}
		}
	}
	return true;
}

