
class DataReader
{
public:
	DataReader(char *data)
	{
		this->data = data;
		pos = 0;
	}
	
	bool ParseVal(char *val);
	bool ParseData(char *problem, int size);

private:
	char *data;
	int pos;
};

class StageInfo
{
public:
	char width;
	char height;
	char level;
	char *problem;
	
	StageInfo()
	{
		problem = NULL;
	}
	
	~StageInfo()
	{
		delete[] problem;
	}
	
	bool ParseData(char *data);

private:
	static const char PARSE_STAGE_STATUS[80];
	static const int PARSE_STAGE_NUM[80];
};

class Stage
{
public:
	char width;
	char height;
	char *status;
	
	Stage(StageInfo &info);
	Stage(const Stage &stage);
	Stage(char width_, char height_);
	
	~Stage()
	{
		delete[] status;
	}
	
	void SetLight(int pos)
	{
		status[pos] = LIGHT;
		
		TurnOnLine(pos, 1);
		TurnOnLine(pos, -1);
		TurnOnLine(pos, width);
		TurnOnLine(pos, -width);
	}
	
	void SetCross(int pos)
	{
		status[pos] = CROSS;
	}
	
	int CountLight(int pos)
	{
		return LIGHT_COUNT[status[pos - 1]] + LIGHT_COUNT[status[pos + 1]] + 
			LIGHT_COUNT[status[pos - width]] + 
			LIGHT_COUNT[status[pos + width]];
	}
	
	int CountBlock(int pos)
	{
		return BLOCK_COUNT[status[pos - 1]] + BLOCK_COUNT[status[pos + 1]] + 
			BLOCK_COUNT[status[pos - width]] + 
			BLOCK_COUNT[status[pos + width]];
	}
	
	int CountBlockEx(int pos)
	{
		return BLOCK_COUNT[status[pos - 1]] + BLOCK_COUNT[status[pos + 1]] + 
			BLOCK_COUNT[status[pos - width]] + 
			BLOCK_COUNT[status[pos + width]] + 
			BLOCK_COUNT[status[pos - 1 - width]] + 
			BLOCK_COUNT[status[pos + 1 - width]] + 
			BLOCK_COUNT[status[pos - 1 + width]] + 
			BLOCK_COUNT[status[pos + 1 + width]];
	}
	
	void Reset(const Stage &stage);
	void Update(bool *updated, bool *error);
	int GetTryPos2(int prev_pos, int *try_pos, int *try_type);
	int GetTryPos3(int prev_pos, int *try_pos, bool *light);
	void FindCross(Stage &stage1, Stage &stage2, bool *updated);
	bool IsSolved();

private:
	static const char CROSS_COUNT[10];
	static const char LIGHT_COUNT[10];
	static const char AISLE_COUNT[10];
	static const char BLOCK_COUNT[10];
	
	void ClearHint(int pos)
	{
		status[pos] = BLOCK;
	}
	
	void TurnOnLine(int start, int diff)
	{
		for(int pos = start + diff; status[pos] < BLOCK; pos += diff){
			status[pos] = TURN_ON;
		}
	}
	
	int CountCross(int pos)
	{
		return CROSS_COUNT[status[pos - 1]] + CROSS_COUNT[status[pos + 1]] + 
			CROSS_COUNT[status[pos - width]] + 
			CROSS_COUNT[status[pos + width]];
	}
	
	int CountAisle(int pos)
	{
		return AISLE_COUNT[status[pos - 1]] + AISLE_COUNT[status[pos + 1]] + 
			AISLE_COUNT[status[pos - width]] + 
			AISLE_COUNT[status[pos + width]];
	}
	
	void UpdateHint(int pos, bool *updated, bool *error);
	bool FindAisle(int start, int diff);
	int CountAisle(int start, int diff, int num, int *found_pos);
	int CountAisle(int pos, int num, int *found_pos);
	void UpdateAisle(int pos, bool *updated);
	void UpdateCross(int pos, bool *updated, bool *error);
	void UpdateDiagonalHint3(int pos, int dx, int dy, bool *updated);
	void UpdateDiagonalHint2(int pos, int dx, int dy, bool *updated);
	void UpdateDiagonalHint1(int pos, int dx, int dy, bool *updated);
};

