
class BinaryBoard
{
public:
	void Initialize(int /* width */, int height)
	{
		height += 2;
		
		memset(board, 0, sizeof(unsigned long long) * height);
	}
	
	void InitializeHints(int width, int height)
	{
		height += 2;
		
		unsigned long long side = (4ull << width) - 1;
		unsigned long long center = ((2ull << width) | 1);
		
		board[0] = board[height - 1] = side;
		for(int i = 1; i < height - 1; i++){
			board[i] = center;
		}
	}
	
	void Set(int x, int y)
	{
		board[y + 1] |= 2ull << x;
	}
	
	bool Get(int x, int y)
	{
		return ((board[y + 1] & (2ull << x)) != 0);
	}
	
	char CountAnswer(int x, int y)
	{
		static const char LUT[32] = {
			0, 1, 4, 3, 1, 2, 3, 2, 1, 2, 3, 2, 2, 3, 2, 1, 
			1, 2, 3, 2, 2, 3, 2, 1, 2, 3, 2, 1, 3, 4, 1, 0, 
		};
		
		unsigned int y0 = (unsigned int)((board[y] >> x) & 2);
		unsigned int y1 = (unsigned int)((board[y + 1] >> x) & 7);
		unsigned int y2 = (unsigned int)((board[y + 2] >> x) & 2);
		return LUT[(y0 << 3) | (y2 << 2) | y1];
	}
	
	bool CheckZero(int x, int y)
	{
		return (((board[y] | board[y + 1] | board[y + 2]) & (7ull << x)) == 0);
	}
	
	void CountHints(int x, int y, int *grid, int *slant)
	{
		static const int EDGE[8] = {
			0, 1, 0, 1, 1, 2, 1, 2, 
		};
		static const int SIDE[8] = {
			0, 0, 1, 1, 0, 0, 1, 1, 
		};
		
		unsigned int y0 = (unsigned int)((board[y] >> x) & 7);
		unsigned int y1 = (unsigned int)((board[y + 1] >> x) & 7);
		unsigned int y2 = (unsigned int)((board[y + 2] >> x) & 7);
		
		*grid = SIDE[y0] + EDGE[y1] + SIDE[y2];
		*slant = EDGE[y0] + EDGE[y2];
	}

private:
#if MAX_WIDTH >= MAX_HEIGHT
	unsigned long long board[MAX_WIDTH + 2];
#else
	unsigned long long board[MAX_HEIGHT + 2];
#endif
};

