
#include "slither-base.h"

static const char * HorizontalSide(ColorInfo &color, int p0, int p1)
{
	if(color.IsSame(p0, p1)){
		return "";
	}else if(color.IsDifferent(p0, p1)){
		return "";
	}else{
		return "@";
	}
}

static const char * VerticalSide(ColorInfo &color, int p0, int p1)
{
	if(color.IsSame(p0, p1)){
		return "";
	}else if(color.IsDifferent(p0, p1)){
		return "";
	}else{
		return "@";
	}
}

static const char * Hint(StageInfo &stage, int pos)
{
	if(stage.GetHint(pos) == -1){
		return "@";
	}else if(stage.GetHint(pos) == 0){
		return "O";
	}else if(stage.GetHint(pos) == 1){
		return "P";
	}else if(stage.GetHint(pos) == 2){
		return "Q";
	}else{
		return "R";
	}
}

void DumpAnswer(ColorInfo &color, StageInfo &stage, int width, int height)
{
	for(int x = 1; x < width - 1; x++){
		printf("E%s", HorizontalSide(color, x, x + width));
	}
	printf("E\n");
	
	for(int y = 1; y < height - 1; y++){
		int pos = y * width + 1;
		for(int x = 1; x < width - 1; x++){
			printf("%s%s", VerticalSide(color, pos - 1, pos), 
				Hint(stage, pos));
			++pos;
		}
		printf("%s\n", VerticalSide(color, pos - 1, pos));
		
		pos = y * width + 1;
		for(int x = 1; x < width - 1; x++){
			printf("E%s", HorizontalSide(color, pos, pos + width));
			++pos;
		}
		printf("E\n");
	}
	printf("\n");
}

#ifdef DUMP_PROBLEM

void DumpProblem(Problem &problem, int width, int height)
{
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			printf("%s", problem.GetProblem(x, y) ? "" : "E");
		}
		printf("\n");
	}
	printf("\n");
}

void DumpHint(char *hint, int width, int height)
{
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			if(hint[y * width + x] == 4){
				printf("  ");
			}else{
				printf(" %d", hint[y * width + x]);
			}
		}
		printf("\n");
	}
	printf("\n");
}

#else

void DumpProblem(Problem & /* problem */, int /* width */, int /* height */)
{
}

void DumpHint(char * /* hint */, int /* width */, int /* height */)
{
}

#endif

