
#include "slither-base.h"

void LogicPre0(ColorInfo &color, StageInfo &stage, int pos, int dx, int dy)
{
	//        x
	// 0 -> x 0 x
	//        x
	color.Merge(pos, pos + dx, true);
	color.Merge(pos, pos - dx, true);
	color.Merge(pos, pos + dy, true);
	color.Merge(pos, pos - dy, true);
	
	stage.Solved(pos);
}

void LogicPre3(ColorInfo &color, StageInfo &stage, int pos, int dx, int dy)
{
	//             o x
	// 3   3  -> | 3 | 3 |
	//             @ x
	if(stage.GetHint(pos + dx) == 3){
		color.Merge(pos, pos - dx, false);
		color.Merge(pos, pos + dx, false);
		color.Merge(pos, pos + dx + dx, true);
		color.Merge(pos - dy, pos - dy + dx, true);
		color.Merge(pos + dy, pos + dy + dx, true);
		color.Merge(pos - dy, pos + dy, false);
	}
	
	//             -
	// 3      -> | 3 @
	//   2         o 2 #
	//     3         $ 3 |
	//                 -
	int n = 1;
	while(stage.GetHint(pos + (dy + dx) * n) == 2){
		++n;
	}
	if(stage.GetHint(pos + (dy + dx) * n) == 3){
		color.Merge(pos, pos - dx, false);
		color.Merge(pos, pos - dy, false);
		color.Merge(pos + dx, pos + dy, false);
		color.Merge(pos + (dy + dx) * n, pos + (dy + dx) * n + dx, false);
		color.Merge(pos + (dy + dx) * n, pos + (dy + dx) * n + dy, false);
		color.Merge(pos + (dy + dx) * n - dx, pos + (dy + dx) * n - dy, false);
	}
}

void LogicPre2(ColorInfo &color, StageInfo &stage, int pos, int dx, int dy)
{
	//             -
	// 3      -> | 3 @
	//   2         o 2 #
	//     3         $ 3 |
	//                 -
	int n1 = 1;
	while(stage.GetHint(pos + (dy + dx) * n1) == 2){
		++n1;
	}
	int n2 = 1;
	while(stage.GetHint(pos - (dy + dx) * n2) == 2){
		++n2;
	}
	
	if(stage.GetHint(pos + (dy + dx) * n1) == 3 && 
		stage.GetHint(pos - (dy + dx) * n2) == 3){
		color.Merge(pos + (dy + dx) * n1, pos + (dy + dx) * n1 + dx, false);
		color.Merge(pos + (dy + dx) * n1, pos + (dy + dx) * n1 + dy, false);
		color.Merge(pos + (dy + dx) * n1 - dx, pos + (dy + dx) * n1 - dy, false);
		color.Merge(pos - (dy + dx) * n2, pos - (dy + dx) * n2 - dx, false);
		color.Merge(pos - (dy + dx) * n2, pos - (dy + dx) * n2 - dy, false);
		color.Merge(pos - (dy + dx) * n2 + dx, pos - (dy + dx) * n2 + dy, false);
	}
}

void Logic3(ColorInfo &color, StageInfo & /* stage */, int pos, int dx, int dy)
{
	//          -
	// 3 x -> | 3 x
	//          -
	if(color.IsSame(pos, pos + dx)){
		color.Merge(pos, pos - dx, false);
		color.Merge(pos, pos + dy, false);
		color.Merge(pos, pos - dy, false);
		return;
	}
	
	//   o       -
	// o 3  -> | 3 o
	//           @
	if(color.IsSame(pos - dx, pos - dy)){
		color.Merge(pos, pos - dx, false);
		color.Merge(pos + dx, pos + dy, false);
	
	//   o      o
	// @ 3 -> @ 3 |
	//          -
	}else if(color.IsDifferent(pos - dx, pos - dy)){
		color.Merge(pos, pos + dx, false);
		color.Merge(pos, pos + dy, false);
	}
	
	//            o
	// o 3 o -> | 3 |
	//            @
	if(color.IsSame(pos - dx, pos + dx)){
		color.Merge(pos, pos - dx, false);
		color.Merge(pos - dy, pos + dy, false);
	
	//            -
	// o 3 @ -> o 3 @
	//            -
	}else if(color.IsDifferent(pos - dx, pos + dx)){
		color.Merge(pos, pos - dy, false);
		color.Merge(pos, pos + dy, false);
	}
	
	// o      o o
	// @ 3 -> @ 3 |
	//          -
	if(color.IsDifferent(pos - dx, pos - dx - dy)){
		color.Merge(pos, pos + dx, false);
		color.Merge(pos, pos + dy, false);
		color.Merge(pos - dx - dy, pos - dy, true);
	}
	
	//          -
	// @ 3 -> @ 3 |
	// o      o o
	if(color.IsDifferent(pos - dx, pos - dx + dy)){
		color.Merge(pos, pos + dx, false);
		color.Merge(pos, pos - dy, false);
		color.Merge(pos - dx + dy, pos + dy, true);
	}
	
	//     o        o
	// o 3   -> | 3
	if(color.IsSame(pos - dx, pos + dx - dy)){
		color.Merge(pos, pos - dx, false);
	
	//     @        @
	// o 3   -> o 3
	//            -
	}else if(color.IsDifferent(pos - dx, pos + dx - dy)){
		color.Merge(pos, pos + dy, false);
	}
	
	// o 3   -> | 3
	//     o        o
	if(color.IsSame(pos - dx, pos + dx + dy)){
		color.Merge(pos, pos - dx, false);
	
	//            -
	// o 3   -> o 3
	//     @        @
	}else if(color.IsDifferent(pos - dx, pos + dx + dy)){
		color.Merge(pos, pos - dy, false);
	}
}

void Logic2(ColorInfo &color, StageInfo &stage, int pos, int dx, int dy)
{
	//   o       o
	// o 2  -> o 2 @
	//           @
	if(color.IsSame(pos - dx, pos - dy)){
		color.Merge(pos - dx, pos + dx, false);
		color.Merge(pos - dx, pos + dy, false);
	
	//   o      o
	// @ 2 -> @ 2 #
	//          $
	}else if(color.IsDifferent(pos - dx, pos - dy)){
		color.Merge(pos + dx, pos + dy, false);
	}
	
	//            @
	// o 2 o -> o 2 o
	//            @
	if(color.IsSame(pos - dx, pos + dx)){
		color.Merge(pos - dx, pos - dy, false);
		color.Merge(pos - dx, pos + dy, false);
	
	//            #
	// o 2 @ -> o 2 @
	//            $
	}else if(color.IsDifferent(pos - dx, pos + dx)){
		color.Merge(pos - dy, pos + dy, false);
		
		// 3 2 -> | 3 2
		if(stage.GetHint(pos - dx) == 3){
			color.Merge(pos - dx, pos - dx - dx, false);
		}
	}
	
	if(color.IsSame(pos, pos - dx)){
		//     x      o x
		// x 2   -> x 2 @
		//            -
		if(color.IsSame(pos, pos + dx - dy)){
			color.Merge(pos, pos + dy, false);
			color.Merge(pos + dx, pos - dy, false);
		//            -
		// x 2   -> x 2 @
		//     x      o x
		}else if(color.IsSame(pos, pos + dx + dy)){
			color.Merge(pos, pos - dy, false);
			color.Merge(pos + dx, pos + dy, false);
		}
		
		//     3      o 3
		// x 2   -> x 2 @
		//            -
		if(stage.GetHint(pos - dy + dx) == 3){
			color.Merge(pos, pos + dy, false);
			color.Merge(pos + dx, pos - dy, false);
		//            -
		// x 2   -> x 2 @
		//     3      o 3
		}else if(stage.GetHint(pos + dy + dx) == 3){
			color.Merge(pos, pos - dy, false);
			color.Merge(pos + dx, pos + dy, false);
		}
	}
	
	if(stage.GetHint(pos - dx) == 3){
		//  |x     |x
		// 3 2 -> 3 2 |
		//        o o
		if(color.IsSame(pos, pos - dy)){
			if(color.IsDifferent(pos - dy, pos - dy - dx)){
				color.Merge(pos, pos + dx, false);
				color.Merge(pos - dx + dy, pos + dy, true);
			}
		//        o o
		// 3 2 -> 3 2 |
		//  |x     |x
		}else if(color.IsSame(pos, pos + dy)){
			if(color.IsDifferent(pos + dy, pos + dy - dx)){
				color.Merge(pos, pos + dx, false);
				color.Merge(pos - dx - dy, pos - dy, true);
			}
		}
	}
}

void Logic1(ColorInfo &color, StageInfo &stage, int pos, int dx, int dy)
{
	//   o       x
	// o 1  -> x 1 o
	//           @
	if(color.IsSame(pos - dx, pos - dy)){
		color.Merge(pos, pos - dx, true);
		color.Merge(pos + dx, pos + dy, false);
	
	//   o      o
	// @ 1 -> @ 1 x
	//          x
	}else if(color.IsDifferent(pos - dx, pos - dy)){
		color.Merge(pos, pos + dx, true);
		color.Merge(pos, pos + dy, true);
	}
	
	//            o
	// o 1 o -> x 1 x
	//            @
	if(color.IsSame(pos - dx, pos + dx)){
		color.Merge(pos, pos - dx, true);
		color.Merge(pos - dy, pos + dy, false);
	
	//            x
	// o 1 @ -> o 1 @
	//            x
	}else if(color.IsDifferent(pos - dx, pos + dx)){
		color.Merge(pos, pos - dy, true);
		color.Merge(pos, pos + dy, true);
	}
	
	//          x
	// | 1 -> | 1 x
	//          x
	if(color.IsDifferent(pos, pos - dx)){
		color.Merge(pos, pos + dx, true);
		color.Merge(pos, pos - dy, true);
		color.Merge(pos, pos + dy, true);
	}
	
	if(stage.GetHint(pos - dx) == 1){
		// o   o      
		// 1   1 -> 1 x 1
		if(color.IsSame(pos - dx - dy, pos - dy)){
			color.Merge(pos, pos - dx, true);
		// 1   1 -> 1 x 1
		// o   o      
		}else if(color.IsSame(pos - dx + dy, pos + dy)){
			color.Merge(pos, pos - dx, true);
		}
	}else if(stage.GetHint(pos - dx) == 3){
		//          x
		// 3 1 -> 3 1 x
		// o o    -
		if(color.IsSame(pos - dx + dy, pos + dy)){
			color.Merge(pos - dx, pos - dx + dy, false);
			color.Merge(pos, pos - dy, true);
			color.Merge(pos, pos + dx, true);
		// o o    -
		// 3 1 -> 3 1 x
		//          x
		}else if(color.IsSame(pos - dx - dy, pos - dy)){
			color.Merge(pos - dx, pos - dx - dy, false);
			color.Merge(pos, pos + dy, true);
			color.Merge(pos, pos + dx, true);
		}
	}
}

void LogicX0(ColorInfo &color, StageInfo & /* stage */, int pos, int dx, int dy)
{
	//   o    o o
	// o @ -> o @
	if(color.IsSame(pos + dx, pos + dy)){
		if(color.IsDifferent(pos + dx, pos + dx + dy)){
			color.Merge(pos, pos + dx, true);
		}
	}
	
	// o @    o @
	// # $ -> o @
	if(color.IsDifferent(pos, pos + dx) && 
		color.IsDifferent(pos + dy, pos + dy + dx)){
		color.Merge(pos, pos + dy, true);
		color.Merge(pos + dx, pos + dx + dy, true);
	}
	
	//   -        -
	// |   | -> |   |
	//            x
	if(color.IsDifferent(pos, pos - dy) && 
		color.IsDifferent(pos, pos - dx) && 
		color.IsDifferent(pos, pos + dx)){
		color.Merge(pos, pos + dy, true);
	}
}

void LogicX1(ColorInfo &color, StageInfo & /* stage */, int pos, int dx, int dy)
{
	//   o        o
	// o   o -> o o o
	//   o        o
	if(color.IsSame(pos - dx, pos - dy) && 
		color.IsSame(pos - dx, pos + dy) && 
		color.IsSame(pos - dx, pos + dx)){
		color.Merge(pos, pos - dx, true);
	}
}

