
#include "slither-base.h"

void ProblemHint::Initialize(Rand &rand, int width, int height)
{
	this->rand = &rand;
	this->width = width;
	this->height = height;
	
	if(!hint){
		hint = new char[width * height];
		poses = new int[width * height];
	}
}

void ProblemHint::Create(char *answer)
{
	int size = width * height;
	for(int i = 0; i < size; i++){
		hint[i] = 4;
	}
	zero.Initialize(width, height);
	
	count = 0;
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int pos = y * width + x;
			if(answer[pos] == 3){
				if(rand->Next(4) > 0){
					hint[pos] = 3;
					++count;
				}
			}else if(answer[pos] == 0 && zero.CheckZero(x, y)){
				if(rand->Next(2) == 0){
					hint[pos] = 0;
					zero.Set(x, y);
					++count;
				}
			}
		}
	}
}

bool ProblemHint::Add(char *answer, Solve &solve, Solve &subsolve, 
	Loop &loop, int loopSize, int firstColor)
{
	solve.Initialize(hint, width, height, loop);
	
	subsolve.Initialize(solve);
	subsolve.ProcessColor1();
	if(subsolve.GetColor().Count() > firstColor){
		return false;
	}
	
	int minColor = subsolve.GetColor().Count();
	int posCount;
	
	while(minColor >= loopSize){
		posCount = 0;
		
		for(int y = 0; y < height; y++){
			for(int x = 0; x < width; x++){
				int pos = y * width + x;
				if(hint[pos] == 4 && !solve.IsSolvedPos(x, y) && 
					answer[pos] != 3 && 
					(answer[pos] != 0 || zero.CheckZero(x, y))){
					subsolve.Initialize(solve);
					subsolve.Update(x, y, answer[pos]);
					subsolve.ProcessColor1();
					subsolve.Update(x, y, -1);
					
					if(subsolve.GetColor().Count() < minColor){
						minColor = subsolve.GetColor().Count();
						poses[0] = pos;
						posCount = 1;
					}else if(subsolve.GetColor().Count() == minColor && 
						posCount != 0){
						poses[posCount++] = pos;
					}
				}
			}
		}
		if(posCount == 0){
			return false;
		}
		
		int p = poses[rand->Next(posCount)];
		hint[p] = answer[p];
		if(answer[p] == 0){
			zero.Set(p % width, p / width);
		}
		++count;
		
		if(minColor == 1){
			return true;
		}
		solve.Update(p % width, p / width, answer[p]);
	}
	
	subsolve.Initialize(solve);
	subsolve.ProcessLoop();
	minColor = subsolve.GetColor().Count();
	
	while(minColor != 1){
		posCount = 0;
		
		for(int y = 0; y < height; y++){
			for(int x = 0; x < width; x++){
				int pos = y * width + x;
				if(hint[pos] == 4 && (answer[pos] != 0 || 
					zero.CheckZero(x, y))){
					subsolve.Initialize(solve);
					subsolve.Update(x, y, answer[pos]);
					subsolve.ProcessLoop();
					subsolve.Update(x, y, -1);
					
					if(subsolve.GetColor().Count() < minColor){
						minColor = subsolve.GetColor().Count();
						poses[0] = pos;
						posCount = 1;
					}else if(subsolve.GetColor().Count() == minColor && 
						posCount != 0){
						poses[posCount++] = pos;
					}
				}
			}
		}
		if(posCount == 0){
			return false;
		}
		
		int p = poses[rand->Next(posCount)];
		hint[p] = answer[p];
		if(answer[p] == 0){
			zero.Set(p % width, p / width);
		}
		++count;
		
		solve.Update(p % width, p / width, answer[p]);
	}
	
	return true;
}

void ProblemHint::Remove(char *answer, Solve &solve, Loop &loop)
{
	solve.Initialize(hint, width, height, loop);
	solve.ProcessColor1();
	int minColor = solve.GetColor().Count();
	
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int pos = y * width + x;
			if(hint[pos] != 4){
				hint[pos] = 4;
				
				solve.Initialize(hint, width, height, loop);
				solve.ProcessColor1();
				if(solve.GetColor().Count() > minColor){
					hint[pos] = answer[pos];
				}else{
					--count;
				}
			}
		}
	}
	
	solve.Initialize(hint, width, height, loop);
	solve.ProcessLoop();
	minColor = solve.GetColor().Count();
	
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int pos = y * width + x;
			if(hint[pos] != 4){
				hint[pos] = 4;
				
				solve.Initialize(hint, width, height, loop);
				solve.ProcessLoop();
				if(solve.GetColor().Count() > minColor || rand->Next(5) > 0){
					hint[pos] = answer[pos];
				}else{
					--count;
				}
			}
		}
	}
	
	zero.Initialize(width, height);
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			if(hint[y * width + x] == 0){
				zero.Set(x, y);
			}
		}
	}
}

void ProblemHint::AdditionalAdd(char *answer, int minHint)
{
	int posCount;
	
	if(count >= minHint){
		return;
	}
	
	BinaryBoard hintCount;
	hintCount.InitializeHints(width, height);
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			if(hint[y * width + x] != 4){
				hintCount.Set(x, y);
			}
		}
	}
	
	while(count < minHint){
		posCount = 0;
		
		for(int y = 0; y < height; y++){
			for(int x = 0; x < width; x++){
				int pos = y * width + x;
				if(hint[pos] == 4 && (answer[pos] != 0 || zero.CheckZero(x, y))){
					int grid;
					int slant;
					hintCount.CountHints(x, y, &grid, &slant);
					if(grid <= 2 && slant <= 3 && grid + slant <= 3){
						poses[posCount++] = pos;
					}
				}
			}
		}
		
		if(posCount == 0){
			return;
		}
		
		int p = poses[rand->Next(posCount)];
		hint[p] = answer[p];
		if(answer[p] == 0){
			zero.Set(p % width, p / width);
		}
		hintCount.Set(p % width, p / width);
		++count;
	}
}
