
#include "slither-base.h"

void Problem::Initialize(Rand &rand, int width, int height)
{
	this->rand = &rand;
	this->width = width;
	this->height = height;
	
	answer = new char[width * height];
}

void Problem::Create()
{
	problem.Initialize(width, height);
	rproblem.Initialize(height, width);
	
	bool rotate = (rand->Next(2) == 0);
	int x1 = 0;
	
	for(int x0 = 0; x0 < width; x0 = x1){
		x1 = rotate ? rand->Next(3) * 4 + 8 + x0 : rand->Next(4) * 4 + 4 + x0;
		if(x1 > width){
			x1 = width;
		}
		if(!rotate && width - x1 == 4){
			x1 = width;
		}
		
		if(rotate){
			if(x0 == 0){
				CreateBlock(rproblem, 0, height, x0, x1);
				RotateBlock(rproblem, problem, 0, height, x0, x1);
			}else{
				CreateBlock(rproblem, 0, height, x0 + 1, x1);
				RotateBlock(rproblem, problem, 0, height, x0 + 1, x1);
			}
		}else{
			CreateBlock(problem, x0, x1, 0, height);
			RotateBlock(problem, rproblem, x0, x1, 0, height);
		}
		if(x0 != 0){
			CreateConnection(problem, x0, 0, height);
			
			for(int y = 0; y < height; y++){
				if(problem.Get(x0 + 1, y) && !problem.Get(x0, y - 1) && 
					!problem.Get(x0 - 1, y - 1) && !problem.Get(x0 - 1, y) && 
					!problem.Get(x0 - 1, y + 1)){
					if(rand->Next(3) > 0){
						problem.Set(x0, y);
					}
				}
			}
		}
		rotate = !rotate;
	}
	
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			answer[y * width + x] = problem.CountAnswer(x, y);
		}
	}
}

void Problem::CreateConnection(BinaryBoard &board, int x, int y0, int y1)
{
	bool rev = (rand->Next(2) == 0);
	bool con = false;
	for(int y = y0; y < y1; y++){
		int yy = rev ? y1 + y0 - y - 1 : y;
		if(board.Get(x - 1, yy) && board.Get(x + 1, yy)){
			if(rand->Next(2) == 0){
				board.Set(x, yy);
				con = true;
				break;
			}
		}
	}
	
	if(!con){
		int y = rand->Next(y1 - y0) + y0;
		board.Set(x, y);
		board.Set(x + 1, y);
		board.Set(x + 2, y);
		if(!board.Get(x - 1, y)){
			board.Set(x - 1, y);
		}
		if(!board.Get(x - 2, y)){
			board.Set(x - 2, y);
		}
	}
}

void Problem::CreateBlock(BinaryBoard &board, int x0, int x1, int y0, int y1)
{
	for(int x = x0; x < x1; x += 4){
		if(rand->Next(8) == 0){
			CreateLine(board, x, y0, y1 - 2);
		}else if(rand->Next(4) == 0){
			CreateLine(board, x, y0, y1 - 1);
		}else{
			CreateLine(board, x, y0, y1);
		}
	}
	for(int x = x0 + 4; x < x1; x += 4){
		CreateConnection(board, x, y0, y1);
	}
	if(x0 == 0){
		for(int y = y0; y < y1; y++){
			if(board.Get(1, y) && !board.Get(0, y - 1)){
				if(rand->Next(3) > 0){
					board.Set(0, y);
				}
			}
		}
	}
}

void Problem::RotateBlock(BinaryBoard &src, BinaryBoard &dest, int x0, 
	int x1, int y0, int y1)
{
	for(int y = y0; y < y1; y++){
		for(int x = x0; x < x1; x++){
			if(src.Get(x, y)){
				dest.Set(y, x);
			}
		}
	}
}

void Problem::CheckZero(BinaryBoard &board, int x, int y, int y0)
{
	if(x > 0 && board.CountAnswer(x - 1, y) == 0 && 
		board.CountAnswer(x - 1, y - 1) == 0){
		if(board.Get(x - 3, y0) || board.Get(x - 4, y0)){
			for(int i = 2; !board.Get(x - i, y0); i++){
				board.Set(x - i, y0);
			}
		}
	}else if(board.CountAnswer(x, y) == 0 && 
		board.CountAnswer(x, y - 1) == 0){
		for(int i = 1; !board.Get(x + i, y0); i++){
			board.Set(x + i, y0);
		}
	}else if(board.CountAnswer(x + 1, y) == 0 && 
		board.CountAnswer(x + 1, y - 1) == 0){
		for(int i = 2; !board.Get(x + i, y0); i++){
			board.Set(x + i, y0);
		}
	}
}

void Problem::CreateLine(BinaryBoard &board, int x, int y0, int y1)
{
	int p0 = rand->Next(3) + 1;
	int p1 = rand->Next(3) + 1;
	int y = y0;
	
	for( ; ; ){
		board.Set(x + p0, y);
		board.Set(x + (p0 + p1) / 2, y);
		board.Set(x + p1, y);
		if(y != y0){
			CheckZero(board, x, y, y);
		}
		
		if(++y == y1){
			break;
		}
		board.Set(x + p1, y);
		if(rand->Next(5) < 2){
			board.Set(x + p0, y);
			board.Set(x + (p0 + p1) / 2, y);
		}else{
			CheckZero(board, x, y, y - 1);
		}
		
		if(rand->Next(2) == 0){
			if(++y == y1){
				break;
			}
			board.Set(x + p1, y);
			if(rand->Next(5) < 2){
				board.Set(x + p0, y);
				board.Set(x + (p0 + p1) / 2, y);
			}else{
				CheckZero(board, x, y, y);
			}
		}
		
		if(++y == y1){
			break;
		}
		
		p0 = p1;
		p1 = rand->Next(3) + 1;
		if(p0 == p1){
			int r = rand->Next(2) + 1;
			if(x > 0 && board.Get(x - 1, y) && p0 > r){
				for(int i = 0; i < r; i++){
					board.Set(x + i, y);
				}
			}else if(r == 1){
				if(p0 == 2){
					board.Set(x + 1, y);
				}else{
					board.Set(x + 2, y);
				}
			}else{
				board.Set(x + 1, y);
				board.Set(x + 2, y);
				board.Set(x + 3, y);
			}
		}
	}
}

bool CreateProblem(Rand &rand, int width, int height, char *result, 
	int *difficulty)
{
	Solve solve;
	Solve subsolve;
	Loop loop;
	
	int size = width * height;
	int rh = rand.Next(36);
	int loopSize = size * (72 + rand.Next(17)) / 1000;
	int maxHint = size * (440 + rh) / 1000;
	int minHint = size * (360 + rh) / 1000;
	int firstColor = size * 80 / 100;
	
	Problem problem;
	problem.Initialize(rand, width, height);
	
	for( ; ; ){
		problem.Create();
		solve.Initialize(problem.GetAnswer(), width, height, loop);
		if(solve.ProcessLoop() && solve.GetColor().Count() == 1){
			DumpProblem(problem, width, height);
			DumpHint(problem.GetAnswer(), width, height);
			break;
		}
	}
	
	ProblemHint hint;
	hint.Initialize(rand, width, height);
	hint.Create(problem.GetAnswer());
	
	if(!hint.Add(problem.GetAnswer(), solve, subsolve, loop, loopSize, 
		firstColor)){
		DumpPrint("failed to adding hint\n");
		return false;
	}
	DumpPrint("solved width hint=%d\n", hint.GetCount());
	
	hint.Remove(problem.GetAnswer(), solve, loop);
	DumpPrint("removing hints, hint=%d\n", hint.GetCount());
	
	if(hint.GetCount() > maxHint){
		DumpPrint("failed to create problem, too many hints\n");
		return false;
	}
	
	if(!hint.Add(problem.GetAnswer(), solve, subsolve, loop, loopSize, 
		firstColor)){
		DumpPrint("failed to adding hint\n");
		return false;
	}
	DumpPrint("solved width hint=%d\n", hint.GetCount());
	
	if(hint.GetCount() > maxHint){
		DumpPrint("failed to create problem, too many hints\n");
		return false;
	}
	
	hint.AdditionalAdd(problem.GetAnswer(), minHint);
	DumpPrint("adding hint, hint=%d\n", hint.GetCount());
	
	solve.Initialize(hint.GetHint(), width, height, loop);
	solve.ProcessColor1();
	int count = solve.GetColor().Count();
	
	bool error = false;
	solve.Initialize(hint.GetHint(), width, height, loop);
	solve.ProcessRecursive(false, &error);
	if(error){
		DumpPrint("failed to create problem, internal error\n");
		return false;
	}
	
	int d = solve.GetTestCount() * 60 + solve.GetTryCount() * 30 + 
		solve.GetLoopCount() * 3 + count - 4;
	d = d * 400 / size;
	if(d < 0){
		d = 0;
	}else if(d > 49){
		d = 49;
	}
	*difficulty = d;

#ifdef DUMP_PROBLEM
	DumpPrint(
		"info: difficulty=%d, test=%d, try=%d, loop=%d, color=%d, color1=%d\n", 
		d, solve.GetTestCount(), solve.GetTryCount(), solve.GetLoopCount(), 
		solve.GetColorCount(), count);
	solve.Dump();
	DumpHint(hint.GetHint(), width, height);
#endif

	memcpy(result, hint.GetHint(), size);
	return true;
}
