
#include "slither-base.h"

static const int WIDTH[5] = {16, 20, 24, 32, MAX_WIDTH};
static const int HEIGHT[5] = {12, 12, 16, 16, MAX_HEIGHT};

static void Help()
{
	printf("\nloop maker solver/creator\n");
	printf("  To slove problems:                            loopmaker_solver.exe s file\n");
	printf("  To create problem grouped by difficulty:      loopmaker_solver.exe c\n");
	printf("  To create problems with specified board size: loopmaker_solver.exe c[1-5]\n");
	printf("    where 1-5 specifies board size.\n");
}

static void Solve(const char *filename)
{
	DWORD t = GetTickCount();
	
	FILE *file = fopen(filename, "rb");
	if(!file){
		printf("failed to open: %s\n", filename);
		return;
	}
	
	fseek(file, 0, SEEK_END);
	int size = ftell(file);
	fseek(file, 0, SEEK_SET);
	
	char *hint = new char[size];
	fread(hint, 1, size, file);
	fclose(file);
	
	int p0 = 0;
	
	if(hint[0] >= '1' && hint[0] <= '5' && hint[1] == '/'){
		for( ; ; ){
			int pp = p0;
			int p1 = 0;
			int temp = 0;
			int height = 0;
			int width = 0;
			int h = 0;
			
			for( ; ; ){
				if(hint[p0] == '/'){
					hint[p0] = '\0';
					if(temp == 0){
						temp = atoi(&hint[pp]);
						pp = p0 + 1;
					}else if(height == 0){
						height = atoi(&hint[pp]);
						pp = p0 + 1;
					}else if(width == 0){
						width = atoi(&hint[pp]);
						++p0;
						pp = 0;
						break;
					}
				}
				if(++p0 == size){
					break;
				}
			}
			if(temp == 0 || width <= 0 || height <= 0){
				break;
			}
			for( ; ; ){
				if(hint[p0] == '.'){
					hint[p1++] = 4;
				}else if(hint[p0] >= '0' && hint[p0] <= '3'){
					hint[p1++] = hint[p0] - '0';
				}else if(hint[p0] == '/'){
					if(width != p1 - pp){
						break;
					}
					
					++h;
					pp = p1;
				}else if(hint[p0] == '\n' || p0 == size - 1){
					++p0;
					break;
				}
				++p0;
			}
			if(h != height){
				break;
			}
			SolveProblem(hint, width, height);
			
			if(p0 == size){
				break;
			}
		}
	}else{
		int pp = 0;
		int p1 = 0;
		int height = 0;
		int width = 0;
		
		for( ; ; ){
			if(hint[p0] == '#'){
				while(p0 < size && hint[p0] != '\n'){
					++p0;
				}
			}else if(hint[p0] == '.'){
				hint[p1++] = 4;
			}else if(hint[p0] >= '0' && hint[p0] <= '3'){
				hint[p1++] = hint[p0] - '0';
			}else if((hint[p0] == '\n' || p0 == size - 1) && pp != p1){
				if(width == 0){
					width = p1 - pp;
				}else if(width != p1 - pp){
					printf("invalid problem data.\n");
					return;
				}
				
				++height;
				pp = p1;
			}
			if(++p0 == size){
				break;
			}
		}
		if(width <= 0 || height <= 0){
			printf("invalid problem data.\n");
			return;
		}
		SolveProblem(hint, width, height);
	}
	
	delete[] hint;
	
	t = GetTickCount() - t;
	printf("\nfinished. time = %d.%03d sec\n", t / 1000, t % 1000);
}

struct ProblemData
{
	int size;
	int difficulty;
	char *hint;
};

static void Create(ProblemData *data, int count, int diff)
{
	Rand rand;
	
#pragma omp parallel for
	for(int i = 0; i < count; i++){
		for( ; ; ){
			if(CreateProblem(rand, WIDTH[data[i].size], HEIGHT[data[i].size], 
				data[i].hint, &data[i].difficulty)){
				break;
			}
		}
		if((i % diff) == diff - 1){
			printf("o");
		}
	}
}

static void InsertionSort(ProblemData *data, int count)
{
	for(int i = 1; i < count; i++){
		ProblemData t = data[i];
		int j;
		for(j = i - 1; j >= 0 && 
			data[j].size * 0x10000 + data[j].difficulty > 
			t.size * 0x10000 + t.difficulty; j--){
			data[j + 1] = data[j];
		}
		data[j + 1] = t;
	}
}

static void Save(FILE *file, ProblemData &data)
{
	int width = WIDTH[data.size];
	int height = HEIGHT[data.size];
	
	fprintf(file, "%d/%d/%d/", data.difficulty / 10 + 1, height, width);
	
	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			int h = data.hint[y * width + x];
			if(h == 4){
				fprintf(file, ".");
			}else{
				fprintf(file, "%d", h);
			}
		}
		fprintf(file, "/");
	}
	fprintf(file, "\n");
}

static FILE * GetFileName(const char *ext1, int ext2)
{
	SYSTEMTIME tm;
	GetLocalTime(&tm);
	
	int count = 0;
	char fileName[256];
	for( ; ; ){
		if(ext2 == 0){
			sprintf(fileName, "%04d%02d%02d_%02d(%s).dat", tm.wYear, 
				tm.wMonth, tm.wDay, ++count, ext1);
		}else{
			sprintf(fileName, "%04d%02d%02d_%02d(%s%d).dat", tm.wYear, 
				tm.wMonth, tm.wDay, ++count, ext1, ext2);
		}
		if(GetFileAttributes(fileName) == INVALID_FILE_ATTRIBUTES){
			FILE *ret = fopen(fileName, "wb");
			if(!ret){
				printf("failed to open: %s\n", fileName);
			}
			return ret;
		}
	}
}

static void Create(int size)
{
	static int COUNT[3][3] = {
		{ 20,  8,  2 }, 
		{ 15, 10,  5 }, 
		{ 10, 10, 10 },
	};
	
	FILE *file[3] = {NULL, NULL, NULL};
	ProblemData *data = NULL;
	int count = 0;
	
	DWORD t = GetTickCount();
	
	if(size == 0){
		file[0] = GetFileName("easy", 0);
		file[1] = GetFileName("normal", 0);
		file[2] = GetFileName("hard", 0);
		if(!file[0] || !file[1] || !file[2]){
			goto EXIT;
		}
		count = 270;
	}else{
		file[0] = GetFileName("size", size);
		if(!file[0]){
			goto EXIT;
		}
		count = 30;
	}
	
	data = new ProblemData[count];
	for(int i = 0; i < count; i++){
		data[i].size = (size == 0) ? i % 3 : size - 1;
		data[i].hint = new char[WIDTH[data[i].size] * HEIGHT[data[i].size]];
	}
	
	printf("creating problems...\n");
	printf("+----------------------------+\n");
	
	Create(data, count, (count + 29) / 30);
	InsertionSort(data, count);
	
	if(size == 0){
		for(int f = 0; f < 3; f++){
			for(int s = 0; s < 3; s++){
				for(int i = s * 90 + f * 20; 
					i < s * 90 + f * 20 + COUNT[f][s]; i++){
					Save(file[f], data[i]);
				}
			}
		}
	}else{
		for(int i = 0; i < count; i++){
			Save(file[0], data[i]);
		}
	}

EXIT:
	if(data){
		for(int i = 0; i < count; i++){
			delete[] data[i].hint;
		}
		delete[] data;
	}
	for(int i = 0; i < 3; i++){
		if(file[i]){
			fclose(file[i]);
		}
	}
	
	t = GetTickCount() - t;
	printf("\nfinished. time = %d.%03d sec\n", t / 1000, t % 1000);
}

int main(int argc, char *argv[])
{
	if(argc == 3 && (argv[1][0] == 's' || argv[1][0] == 'S')){
		Solve(argv[2]);
	}else if(argc == 2 && (argv[1][0] == 'c' || argv[1][0] == 'C')){
		int size = argv[1][1] - '0';
		if(size < 0 || size > 5){
			Create(0);
		}else{
			Create(size);
		}
	}else{
		Help();
	}
	
	return 0;
}
