
#include "slither-base.h"

void Solve::Initialize(const char *hint, int width, int height, Loop &loop)
{
	width += 2;
	height += 2;
	
	this->loop = &loop;
	this->width = width;
	this->height = height;
	size = width * height;
	testCount = 0;
	tryCount = 0;
	colorCount = 0;
	loopCount = 0;
	answer = 0;
	
	stage.Initialize(hint, width - 2, height - 2, size);
	color.Initialize(size);
	loop.Initialize(size, 1, width);
	
	int ll = (height - 1) * width;
	for(int i = 0; i < width; i++){
		color.Merge(0, i, true);
		color.Merge(0, ll + i, true);
	}
	for(int i = 0; i < height; i++){
		color.Merge(0, i * width, true);
		color.Merge(0, i * width + width - 1, true);
	}
	
	for(int i = 0; i < size; i++){
		if(stage.GetHint(i) == 0){
			LogicPre0(color, stage, i,  1,  width);
		}else if(stage.GetHint(i) == 3){
			LogicPre3(color, stage, i,  1,  width);
			LogicPre3(color, stage, i, -width,  1);
			LogicPre3(color, stage, i, -1, -width);
			LogicPre3(color, stage, i,  width, -1);
		}
	}
}

void Solve::Initialize(Solve &base)
{
	loop = base.loop;
	width = base.width;
	height = base.height;
	size = base.size;
	testCount = base.testCount;
	tryCount = base.tryCount;
	colorCount = base.colorCount;
	loopCount = base.loopCount;
	answer = base.answer;
	
	stage.Initialize(base.stage, size);
	color.Initialize(base.color);
	loop->Initialize();
}

void Solve::ProcessColor1()
{
	for(int i = width; i < size - width; i++){
		if(!stage.IsSolved(i)){
			if(stage.GetHint(i) == 3){
				Logic3(color, stage, i,  1,  width);
				Logic3(color, stage, i, -width,  1);
				Logic3(color, stage, i, -1, -width);
				Logic3(color, stage, i,  width, -1);
			}else if(stage.GetHint(i) == 2){
				Logic2(color, stage, i,  1,  width);
				Logic2(color, stage, i, -width,  1);
				Logic2(color, stage, i, -1, -width);
				Logic2(color, stage, i,  width, -1);
			}else if(stage.GetHint(i) == 1){
				Logic1(color, stage, i,  1,  width);
				Logic1(color, stage, i, -width,  1);
				Logic1(color, stage, i, -1, -width);
				Logic1(color, stage, i,  width, -1);
			}else{
				LogicX1(color, stage, i,  1,  width);
			}
			if(color.IsSolved(i, i - 1) && color.IsSolved(i, i + 1) && 
				color.IsSolved(i, i - width) && color.IsSolved(i, i + width)){
				stage.Solved(i);
			}
		}
		LogicX0(color, stage, i,  1,  width);
		LogicX0(color, stage, i, -width,  1);
		LogicX0(color, stage, i, -1, -width);
		LogicX0(color, stage, i,  width, -1);
	}
}

bool Solve::ProcessLoop()
{
	for( ; ; ){
		for( ; ; ){
			int count = color.Count();
			ProcessColor1();
			++colorCount;
			
			if(color.HasError()){
				return false;
			}else if(color.Count() == 1){
				return loop->FinalCheck(color);
			}else if(color.Count() == count){
				break;
			}
		}
		
		++loopCount;
		
		if(!loop->ClosedCheck(color)){
			return !loop->HasError();
		}
	}
}

bool Solve::Process(bool dump, bool *error)
{
	Solve subSolve;
	
	for( ; ; ){
		if(!ProcessLoop()){
			*error = true;
			return false;
		}
		if(color.Count() == 1){
			if(dump){
				Dump();
			}
			
			++answer;
			*error = false;
			return true;
		}
		
		if(!TryOnce(subSolve)){
			*error = false;
			return false;
		}
	}
}

void Solve::ProcessRecursive(bool dump, bool *error)
{
	if(Process(dump, error) || *error){
		return;
	}
	TryTwice(dump, error);
}

bool Solve::TryOnce(Solve &subSolve)
{
	int pos0 = 0;
	int pos1 = 0;
	for( ; ; ){
		pos1 = color.GetTryPosition(pos1 + 1);
		if(pos1 == -1){
			pos0 = color.GetTryPosition(pos0 + 1);
			pos1 = color.GetTryPosition(pos0 + 1);
			if(pos0 == -1 || pos1 == -1){
				return false;
			}
		}
		
		subSolve.Initialize(*this);
		subSolve.color.Merge(pos0, pos1, true);
		if(!subSolve.ProcessLoop()){
			color.Merge(pos0, pos1, false);
			++tryCount;
			return true;
		}
		
		subSolve.Initialize(*this);
		subSolve.color.Merge(pos0, pos1, false);
		if(!subSolve.ProcessLoop()){
			color.Merge(pos0, pos1, true);
			++tryCount;
			return true;
		}
	}
}

void Solve::TryTwice(bool dump, bool *error)
{
	Solve subSolve;
	
	int pos0 = 0;
	int pos1 = 0;
	for( ; ; ){
		pos1 = color.GetTryPosition(pos1 + 1);
		if(pos1 == -1){
			pos0 = color.GetTryPosition(pos0 + 1);
			pos1 = color.GetTryPosition(pos0 + 1);
			if(pos0 == -1 || pos1 == -1){
				*error = true;
				return;
			}
		}
		
		subSolve.Initialize(*this);
		subSolve.color.Merge(pos0, pos1, true);
		subSolve.Process(false, error);
		if(*error){
			color.Merge(pos0, pos1, false);
			++tryCount;
			Process(dump, error);
			return;
		}
		
		subSolve.Initialize(*this);
		subSolve.color.Merge(pos0, pos1, false);
		subSolve.Process(false, error);
		if(*error){
			color.Merge(pos0, pos1, true);
			++tryCount;
			Process(dump, error);
			return;
		}
	}
}

void Solve::Update(int x, int y, char hint)
{
	int i = (y + 1) * width + x + 1;
	stage.SetHint(i, hint);
	
	if(hint == 0){
		LogicPre0(color, stage, i,  1,  width);
	}else if(hint == 3){
		LogicPre3(color, stage, i,  1,  width);
		LogicPre3(color, stage, i, -width,  1);
		LogicPre3(color, stage, i, -1, -width);
		LogicPre3(color, stage, i,  width, -1);
	}else if(hint == 2){
		LogicPre2(color, stage, i,  1,  width);
		LogicPre2(color, stage, i, -width,  1);
		LogicPre2(color, stage, i, -1, -width);
		LogicPre2(color, stage, i,  width, -1);
	}
}

int SolveProblem(const char *hint, int width, int height)
{
	Loop loop;
	Solve solve;
	
	solve.Initialize(hint, width, height, loop);
	
	bool error = false;
	solve.ProcessRecursive(false, &error);
	
	if(error){
		printf("failed to solve.\n");
		return 0;
	}else{
		printf("found %d answer(s).\n", solve.GetAnswer());
		return solve.GetAnswer();
	}
}
