
#include "sudoku_creator.h"

void SetValue(WORD *work, WORD *map, int pos, WORD val)
{
	work[pos] = val;
	map[pos] = 0x1ff;
	
	int y = pos / 9;
	int x = pos % 9;
	int a = (y / 3) * 3 + (x / 3);
	
	WORD *yoko = map + y * 9;
	WORD *tate = map + x;
	WORD *area = map + AREA_OFFSETS[a];
	
	yoko[0] |= val; yoko[1] |= val; yoko[2] |= val;
	yoko[3] |= val; yoko[4] |= val; yoko[5] |= val;
	yoko[6] |= val; yoko[7] |= val; yoko[8] |= val;
	
	tate[0] |= val; tate[9] |= val; tate[18] |= val;
	tate[27] |= val; tate[36] |= val; tate[45] |= val;
	tate[54] |= val; tate[63] |= val; tate[72] |= val;
	
	area[0] |= val; area[1] |= val; area[2] |= val;
	area[9] |= val; area[10] |= val; area[11] |= val;
	area[18] |= val; area[19] |= val; area[20] |= val;
}

static bool Logic1(WORD *work, WORD *map)
{
	bool ret = false;
	
	for(int a = 0; a < 9; ++a){
		WORD *workArea = work + AREA_OFFSETS[a];
		WORD *mapArea = map + AREA_OFFSETS[a];
		
		WORD areaVal = 
			workArea[0] | workArea[1] | workArea[2] | 
			workArea[9] | workArea[10] | workArea[11] | 
			workArea[18] | workArea[19] | workArea[20];
		
		for(int i = 0; i < 9; ++i){
			WORD val = (WORD)(1 << i);
			
			if((areaVal & val) == 0){
				int count = 0;
				int pos = 0;
				
				for(int j = 0; j < 9; ++j){
					if((mapArea[AREA_INDEXES[j]] & val) == 0){
						++count;
						pos = AREA_INDEXES[j];
					}
				}
				
				if(count == 1){
					ret = true;
					SetValue(work, map, AREA_OFFSETS[a] + pos, val);
				}
			}
		}
	}
	
	return ret;
}

static void Logic2Map(WORD *work, WORD *map, int area, int yoko1, int yoko2, int tate1, int tate2)
{
	WORD *workArea = work + AREA_OFFSETS[area];
	WORD *mapArea = map + AREA_OFFSETS[area];
	WORD *yokoArea1 = map + AREA_OFFSETS[yoko1];
	WORD *yokoArea2 = map + AREA_OFFSETS[yoko2];
	WORD *tateArea1 = map + AREA_OFFSETS[tate1];
	WORD *tateArea2 = map + AREA_OFFSETS[tate2];
	
	WORD areaVal = 
		workArea[0] | workArea[1] | workArea[2] | 
		workArea[9] | workArea[10] | workArea[11] | 
		workArea[18] | workArea[19] | workArea[20];
	
	WORD tempYoko1 = mapArea[0] & mapArea[1] & mapArea[2];
	WORD tempYoko2 = mapArea[9] & mapArea[10] & mapArea[11];
	WORD tempYoko3 = mapArea[18] & mapArea[19] & mapArea[20];
	WORD tempTate1 = mapArea[0] & mapArea[9] & mapArea[18];
	WORD tempTate2 = mapArea[1] & mapArea[10] & mapArea[19];
	WORD tempTate3 = mapArea[2] & mapArea[11] & mapArea[20];
	
	WORD maskYoko1 = (tempYoko2 & tempYoko3) & ~tempYoko1;
	WORD maskYoko2 = (tempYoko1 & tempYoko3) & ~tempYoko2;
	WORD maskYoko3 = (tempYoko1 & tempYoko2) & ~tempYoko3;
	WORD maskTate1 = (tempTate2 & tempTate3) & ~tempTate1;
	WORD maskTate2 = (tempTate1 & tempTate3) & ~tempTate2;
	WORD maskTate3 = (tempTate1 & tempTate2) & ~tempTate3;
	
	for(int i = 0; i < 9; ++i){
		WORD val = (WORD)(1 << i);
		
		if((areaVal & val) == 0){
			if(maskYoko1 & val){
				yokoArea1[0] |= val; yokoArea1[1] |= val; yokoArea1[2] |= val;
				yokoArea2[0] |= val; yokoArea2[1] |= val; yokoArea2[2] |= val;
			}
			
			if(maskYoko2 & val){
				yokoArea1[9] |= val; yokoArea1[10] |= val; yokoArea1[11] |= val;
				yokoArea2[9] |= val; yokoArea2[10] |= val; yokoArea2[11] |= val;
			}
			
			if(maskYoko3 & val){
				yokoArea1[18] |= val; yokoArea1[19] |= val; yokoArea1[20] |= val;
				yokoArea2[18] |= val; yokoArea2[19] |= val; yokoArea2[20] |= val;
			}
			
			if(maskTate1 & val){
				tateArea1[0] |= val; tateArea1[9] |= val; tateArea1[18] |= val;
				tateArea2[0] |= val; tateArea2[9] |= val; tateArea2[18] |= val;
			}
			
			if(maskTate2 & val){
				tateArea1[1] |= val; tateArea1[10] |= val; tateArea1[19] |= val;
				tateArea2[1] |= val; tateArea2[10] |= val; tateArea2[19] |= val;
			}
			
			if(maskTate3 & val){
				tateArea1[2] |= val; tateArea1[11] |= val; tateArea1[20] |= val;
				tateArea2[2] |= val; tateArea2[11] |= val; tateArea2[20] |= val;
			}
		}
	}
}

static bool Logic2(WORD *work, WORD *map)
{
	WORD prevMap[9 * 9];
	memcpy(prevMap, map, sizeof(WORD) * 9 * 9);
	
	Logic2Map(work, map, 0, 1, 2, 3, 6);
	Logic2Map(work, map, 1, 0, 2, 4, 7);
	Logic2Map(work, map, 2, 0, 1, 5, 8);
	Logic2Map(work, map, 3, 4, 5, 0, 6);
	Logic2Map(work, map, 4, 3, 5, 1, 7);
	Logic2Map(work, map, 5, 3, 4, 2, 8);
	Logic2Map(work, map, 6, 7, 8, 0, 3);
	Logic2Map(work, map, 7, 6, 8, 1, 4);
	Logic2Map(work, map, 8, 6, 7, 2, 5);
	
	return (memcmp(prevMap, map, sizeof(WORD) * 9 * 9) != 0);
}

static inline WORD WordCount(WORD val)
{
	val = ((val & 0xaaaa) >> 1) + (val & 0x5555);
	val = ((val & 0xcccc) >> 2) + (val & 0x3333);
	val = ((val & 0xf0f0) >> 4) + (val & 0x0f0f);
	val = ((val & 0xff00) >> 8) + (val & 0x00ff);
	
	return val;
}

static bool Logic3(WORD *work, WORD *map)
{
	bool ret = false;
	
	for(int i = 0; i < 9 * 9; ++i){
		if(work[i] == 0){
			WORD val = (~(map[i]) & 0x1ff);
			
			if(WordCount(val) == 1){
				ret = true;
				SetValue(work, map, i, val);
			}
		}
	}
	
	return ret;
}

static bool Logic4(WORD *work, WORD *map)
{
	bool ret = false;
	
	for(int a = 0; a < 9; ++a){
		WORD *workTate = work + a;
		WORD *mapTate = map + a;
		
		WORD tateVal = 
			workTate[0] | workTate[9] | workTate[18] | 
			workTate[27] | workTate[36] | workTate[45] | 
			workTate[54] | workTate[63] | workTate[72];
		
		WORD *workYoko = work + a * 9;
		WORD *mapYoko = map + a * 9;
		
		WORD yokoVal = 
			workYoko[0] | workYoko[1] | workYoko[2] | 
			workYoko[3] | workYoko[4] | workYoko[5] | 
			workYoko[6] | workYoko[7] | workYoko[8];
		
		for(int i = 0; i < 9; ++i){
			WORD val = (WORD)(1 << i);
			
			if((tateVal & val) == 0){
				int count = 0;
				int pos = 0;
				
				for(int j = 0; j < 9; ++j){
					if((mapTate[j * 9] & val) == 0){
						++count;
						pos = j * 9;
					}
				}
				
				if(count == 1){
					ret = true;
					SetValue(work, map, a + pos, val);
				}
			}
			
			if((yokoVal & val) == 0){
				int count = 0;
				int pos = 0;
				
				for(int j = 0; j < 9; ++j){
					if((mapYoko[j] & val) == 0){
						++count;
						pos = j;
					}
				}
				
				if(count == 1){
					ret = true;
					SetValue(work, map, a * 9 + pos, val);
				}
			}
		}
	}
	
	return ret;
}

void Search2(WORD *work, WORD *map, const WORD *indexes)
{
	WORD poses[9];
	WORD size = 0;
	
	for(WORD i = 0; i < 9; ++i){
		if(work[indexes[i]] == 0){
			if(WordCount(~(map[indexes[i]]) & 0x1ff) == 2){
				poses[size++] = i;
			}
		}
	}
	
	if(size >= 2){
		for(WORD i = 0; i < size - 1; ++i){
			for(WORD j = i + 1; j < size; ++j){
				WORD val = (WORD)(~(map[indexes[poses[i]]] & map[indexes[poses[j]]]) & 0x1ff);
				if(WordCount(val) == 2){
					for(WORD k = 0; k < 9; ++k){
						if(k != poses[i] && k != poses[j]){
							map[indexes[k]] |= val;
						}
					}
				}
			}
		}
	}
}

void Search3(WORD *work, WORD *map, const WORD *indexes)
{
	WORD poses[9];
	WORD size = 0;
	
	for(WORD i = 0; i < 9; ++i){
		if(work[indexes[i]] == 0){
			WORD count = WordCount(~(map[indexes[i]]) & 0x1ff);
			if(count == 2 || count == 3){
				poses[size++] = i;
			}
		}
	}
	
	if(size >= 3){
		for(WORD i = 0; i < size - 2; ++i){
			for(WORD j = i + 1; j < size - 1; ++j){
				for(WORD k = j + 1; k < size; ++k){
					WORD val = (WORD)(~(map[indexes[poses[i]]] & map[indexes[poses[j]]] & map[indexes[poses[k]]]) & 0x1ff);
					if(WordCount(val) == 3){
						for(WORD l = 0; l < 9; ++l){
							if(l != poses[i] && l != poses[j] && l != poses[k]){
								map[indexes[l]] |= val;
							}
						}
					}
				}
			}
		}
	}
}

static bool Logic5(WORD *work, WORD *map)
{
	WORD prevMap[9 * 9];
	memcpy(prevMap, map, sizeof(WORD) * 9 * 9);
	
	for(int i = 0; i < 9; ++i){
		Search2(work + i * 9, map + i * 9, YOKO_INDEXES);
		Search2(work + i, map + i, TATE_INDEXES);
		Search2(work + AREA_OFFSETS[i], map + AREA_OFFSETS[i], AREA_INDEXES);
	}
	
	return (memcmp(prevMap, map, sizeof(WORD) * 9 * 9) != 0);
}

static bool Logic6(WORD *work, WORD *map)
{
	WORD prevMap[9 * 9];
	memcpy(prevMap, map, sizeof(WORD) * 9 * 9);
	
	for(int i = 0; i < 9; ++i){
		Search3(work + i * 9, map + i * 9, YOKO_INDEXES);
		Search3(work + i, map + i, TATE_INDEXES);
		Search3(work + AREA_OFFSETS[i], map + AREA_OFFSETS[i], AREA_INDEXES);
	}
	
	return (memcmp(prevMap, map, sizeof(WORD) * 9 * 9) != 0);
}

static inline bool Solved(WORD *work)
{
	for(int i = 0; i < 9 * 9; ++i){
		if(work[i] == 0){
			return false;
		}
	}
	
	return true;
}

static void Initialize(char *problem, WORD *work, WORD *map, int *counter)
{
	for(int i = 0; i < 9 * 9; ++i){
		work[i] = SOLVE_VALS[problem[i]];
	}
	
	WORD yoko[9];
	WORD tate[9];
	WORD area[9];
	
	for(int i = 0; i < 9; ++i){
		WORD *py = &work[i * 9];
		WORD *pt = &work[i];
		WORD *pa = &work[AREA_OFFSETS[i]];
		
		yoko[i] = (py[0] | py[1] | py[2] | py[3] | py[4] | py[5] | py[6] | py[7] | py[8]);
		tate[i] = (pt[0] | pt[9] | pt[18] | pt[27] | pt[36] | pt[45] | pt[54] | pt[63] | pt[72]);
		area[i] = (pa[0] | pa[1] | pa[2] | pa[9] | pa[10] |  pa[11] | pa[18] | pa[19] | pa[20]);
	}
	
	for(int i = 0; i < 9 * 9; ++i){
		if(work[i]){
			map[i] = 0x1ff;
		}else{
			int y = i / 9;
			int x = i % 9;
			int a = (y / 3) * 3 + (x / 3);
			
			map[i] = (yoko[y] | tate[x] | area[a]);
		}
	}
	
	memset(counter, 0, sizeof(int) * 6);
}

static void Finalize(WORD *work, char *answer)
{
	if(!answer){
		return;
	}
	
	for(int i = 0; i < 9 * 9; ++i){
		if(work[i] > 0x0010){
			if(work[i] > 0x0040){
				answer[i] = (work[i] == 0x0080) ? 8 : 9;
			}else{
				answer[i] = (work[i] == 0x0020) ? 6 : 7;
			}
		}else{
			if(work[i] > 0x0004){
				answer[i] = (work[i] == 0x0008) ? 4 : 5;
			}else if(work[i] > 0x0001){
				answer[i] = (work[i] == 0x0002) ? 2 : 3;
			}else{
				answer[i] = (work[i] == 0x0000) ? 0 : 1;
			}
		}
	}
}

static inline void Logic1Loop(WORD *work, WORD *map, int *counter)
{
	while(Logic1(work, map)){
		++counter[0];
	}
}

static inline void Logic2Loop(WORD *work, WORD *map, int *counter)
{
	while(Logic2(work, map)){
		++counter[1];
		Logic1Loop(work, map, counter);
	}
}

static inline void Logic3Loop(WORD *work, WORD *map, int *counter)
{
	while(Logic3(work, map)){
		++counter[2];
		Logic1Loop(work, map, counter);
		Logic2Loop(work, map, counter);
	}
}

static inline void Logic4Loop(WORD *work, WORD *map, int *counter)
{
	while(Logic4(work, map)){
		++counter[3];
		Logic1Loop(work, map, counter);
		Logic2Loop(work, map, counter);
		Logic3Loop(work, map, counter);
	}
}

static inline void Logic5Loop(WORD *work, WORD *map, int *counter)
{
	while(Logic5(work, map)){
		++counter[4];
		Logic1Loop(work, map, counter);
		Logic2Loop(work, map, counter);
		Logic3Loop(work, map, counter);
		Logic4Loop(work, map, counter);
	}
}

static inline void Logic6Loop(WORD *work, WORD *map, int *counter)
{
	while(Logic6(work, map)){
		++counter[5];
		Logic1Loop(work, map, counter);
		Logic2Loop(work, map, counter);
		Logic3Loop(work, map, counter);
		Logic4Loop(work, map, counter);
		Logic5Loop(work, map, counter);
	}
}

void LogicLoop(WORD *work, WORD *map, int *counter)
{
	Logic6Loop(work, map, counter);
}

bool Solve(char *problem, char *answer, int *counter)
{
	WORD work[9 * 9];
	WORD map[9 * 9];
	
	Initialize(problem, work, map, counter);
	
	Logic1Loop(work, map, counter);
	if(Solved(work)){
		Finalize(work, answer);
		return true;
	}
	
	Logic2Loop(work, map, counter);
	if(Solved(work)){
		Finalize(work, answer);
		return true;
	}
	
	Logic3Loop(work, map, counter);
	if(Solved(work)){
		Finalize(work, answer);
		return true;
	}
	
	Logic4Loop(work, map, counter);
	if(Solved(work)){
		Finalize(work, answer);
		return true;
	}
	
	Logic5Loop(work, map, counter);
	if(Solved(work)){
		Finalize(work, answer);
		return true;
	}
	
	Logic6Loop(work, map, counter);
	if(Solved(work)){
		Finalize(work, answer);
		return true;
	}
	
	Finalize(work, answer);
	return false;
}

