
#include "sudoku_creator.h"

struct Stack
{
	int pos;
	int yoko[9 * 9];
	int tate[9 * 9];
	int area[9 * 9];
};


static inline void SetAnswer(WORD *work, WORD *yokoMap, WORD *tateMap, WORD *areaMap, int pos, int yoko, int tate, int area, WORD val)
{
	work[pos] = val;
	
	yokoMap[yoko] |= val;
	tateMap[tate] |= val;
	areaMap[area] |= val;
}

static inline void ResetAnswer(WORD *work, WORD *yokoMap, WORD *tateMap, WORD *areaMap, int pos, int yoko, int tate, int area)
{
	WORD val = ~work[pos];
	work[pos] = 0;
	
	yokoMap[yoko] &= val;
	tateMap[tate] &= val;
	areaMap[area] &= val;
}

static inline void PushStack(Stack *stack, WORD *work, WORD *yokoMap, WORD *tateMap, WORD *areaMap, int pos, int yoko, int tate, int area, WORD val)
{
	int stackPos = stack->pos++;
	stack->yoko[stackPos] = yoko;
	stack->tate[stackPos] = tate;
	stack->area[stackPos] = area;
	
	SetAnswer(work, yokoMap, tateMap, areaMap, pos, yoko, tate, area, val);
}

static inline void PopStack(Stack *stack, WORD *work, WORD *yokoMap, WORD *tateMap, WORD *areaMap)
{
	int stackPos = --(stack->pos);
	int pos = stack->yoko[stackPos] * 9 + stack->tate[stackPos];
	
	ResetAnswer(work, yokoMap, tateMap, areaMap, pos, stack->yoko[stackPos], stack->tate[stackPos], stack->area[stackPos]);
}

static inline bool LogicSolve(Stack *stack, WORD *work, WORD *yokoMap, WORD *tateMap, WORD *areaMap, int pos, bool *error)
{
	bool ret = false;
	
	for(int i = pos; i < 9 * 9; ++i){
		if(work[i] == 0){
			int yoko = i / 9;
			int tate = i % 9;
			int area = (yoko / 3) * 3 + (tate / 3);
			
			WORD map = yokoMap[yoko] | tateMap[tate] | areaMap[area];
			map = (~map & 0x1ff);
			
			if(map == 0){
				*error = true;
				return false;
			}else if((map & (map - 1)) == 0){
				PushStack(stack, work, yokoMap, tateMap, areaMap, i, yoko, tate, area, map);
				ret = true;
			}
		}
	}
	
	return ret;
}

static inline void LogicSolveLoop(Stack *stack, WORD *work, WORD *yokoMap, WORD *tateMap, WORD *areaMap, int pos, bool *error)
{
	while(LogicSolve(stack, work, yokoMap, tateMap, areaMap, pos, error)){
	}
}

static void BackTrack(Stack *stack, WORD *work, WORD *result, WORD *yokoMap, WORD *tateMap, WORD *areaMap, int pos, int *num)
{
	int stackPos = stack->pos;
	
	bool error = false;
	LogicSolveLoop(stack, work, yokoMap, tateMap, areaMap, pos, &error);
	
	if(error){
		while(stackPos != stack->pos){
			PopStack(stack, work, yokoMap, tateMap, areaMap);
		}
		
		return;
	}
	
	while(work[pos] != 0){
		if(++pos == 9 * 9){
			++(*num);
			memcpy(result, work, sizeof(WORD) * 9 * 9);
			
			while(stackPos != stack->pos){
				PopStack(stack, work, yokoMap, tateMap, areaMap);
			}
			
			return;
		}
	}
	
	int yoko = pos / 9;
	int tate = pos % 9;
	int area = (yoko / 3) * 3 + (tate / 3);
	
	WORD map = yokoMap[yoko] | tateMap[tate] | areaMap[area];
	
	if(map != 0x1ff){
		for(int i = 1; i < 10; i++){
			if((map & SOLVE_VALS[i]) == 0){
				SetAnswer(work, yokoMap, tateMap, areaMap, pos, yoko, tate, area, SOLVE_VALS[i]);
				BackTrack(stack, work, result, yokoMap, tateMap, areaMap, pos, num);
				ResetAnswer(work, yokoMap, tateMap, areaMap, pos, yoko, tate, area);
			}
			if(*num > 1){
				break;
			}
		}
	}
	
	while(stackPos != stack->pos){
		PopStack(stack, work, yokoMap, tateMap, areaMap);
	}
	
	return;
}

static inline int GetAnswer(WORD *work, WORD *result)
{
	WORD yokoMap[9];
	WORD tateMap[9];
	WORD areaMap[9];
	
	for(int i = 0; i < 9; ++i){
		WORD *py = &work[i * 9];
		WORD *pt = &work[i];
		WORD *pa = &work[AREA_OFFSETS[i]];
		
		yokoMap[i] = (py[0] | py[1] | py[2] | py[3] | py[4] | py[5] | py[6] | py[7] | py[8]);
		tateMap[i] = (pt[0] | pt[9] | pt[18] | pt[27] | pt[36] | pt[45] | pt[54] | pt[63] | pt[72]);
		areaMap[i] = (pa[0] | pa[1] | pa[2] | pa[9] | pa[10] |  pa[11] | pa[18] | pa[19] | pa[20]);
	}
	
	Stack stack;
	stack.pos = 0;
	
	int num = 0;
	BackTrack(&stack, work, result, yokoMap, tateMap, areaMap, 0, &num);
	
	return num;
}


static inline char & Value(char *values, int x, int y){ return values[y * 9 + x]; }

static void CornerPattern(Rand &rand, char *pattern, int corner, int side)
{
	for(int i = 0; i < corner; i++){
		int x, y;
		
		do{
			x = (int)(rand.rand() * 3.0);
			y = (int)(rand.rand() * 3.0);
		}while(Value(pattern, x, y) != 0);
		
		Value(pattern, x, y) = 1;
		Value(pattern, 8 - x, 8 - y) = 1;
		Value(pattern, 8 - y, x) = 1;
		Value(pattern, y, 8 - x) = 1;
	}
	
	for(int i = 0; i < side; i++){
		int x, y;
		
		do{
			x = (int)(rand.rand() * 3.0);
			y = (int)(rand.rand() * 3.0) + 3;
		}while(Value(pattern, x, y) != 0);
		
		Value(pattern, x, y) = 1;
		Value(pattern, 8 - x, 8 - y) = 1;
		Value(pattern, 8 - y, x) = 1;
		Value(pattern, y, 8 - x) = 1;
	}
}

static inline int MakePattern(Rand &rand, char *pat)
{
	char pattern[9 * 9] = {};
	
	int r = (int)(rand.rand() * 5.0);
	if(r == 0){
		Value(pattern, 4, 4) = 1;
		CornerPattern(rand, pattern, 3, 3);
	}else if(r == 1){
		int r2 = (int)(rand.rand() * 4.0);
		if(r2 == 0){
			Value(pattern, 3, 3) = 1;
			Value(pattern, 5, 5) = 1;
		}else if(r2 == 1){
			Value(pattern, 4, 3) = 1;
			Value(pattern, 4, 5) = 1;
		}else if(r2 == 2){
			Value(pattern, 3, 5) = 1;
			Value(pattern, 5, 3) = 1;
		}else{
			Value(pattern, 3, 4) = 1;
			Value(pattern, 5, 4) = 1;
		}
		
		CornerPattern(rand, pattern, 2, 3);
		
		int x, y;
		
		do{
			x = (int)(rand.rand() * 6.0);
			if(x >= 3){
				x += 3;
			}
			y = (int)(rand.rand() * 3.0);
		}while(Value(pattern, x, y) != 0);
		
		Value(pattern, x, y) = 1;
		Value(pattern, 8 - x, 8 - y) = 1;
	}else if(r == 2){
		int r2 = (int)(rand.rand() * 2.0);
		if(r2 == 0){
			Value(pattern, 3, 3) = 1;
			Value(pattern, 5, 5) = 1;
		}else{
			Value(pattern, 3, 5) = 1;
			Value(pattern, 5, 3) = 1;
		}
		
		CornerPattern(rand, pattern, 3, 2);
		
		int x, y;
		
		do{
			x = (int)(rand.rand() * 3.0);
			y = (int)(rand.rand() * 6.0);
			if(y < 3){
				x += 3;
			}
		}while(Value(pattern, x, y) != 0);
		
		Value(pattern, x, y) = 1;
		Value(pattern, 8 - x, 8 - y) = 1;
	}else{
		int r2 = (int)(rand.rand() * 2.0);
		if(r2 == 0){
			Value(pattern, 3, 3) = 1;
			Value(pattern, 5, 5) = 1;
			Value(pattern, 3, 5) = 1;
			Value(pattern, 5, 3) = 1;
		}else{
			Value(pattern, 3, 4) = 1;
			Value(pattern, 4, 3) = 1;
			Value(pattern, 4, 5) = 1;
			Value(pattern, 5, 4) = 1;
		}
		
		CornerPattern(rand, pattern, 2, 3);
	}
	
	int l = 0;
	for(int i = 0; i < 9 * 9; i++){
		if(pattern[i] != 0){
			pat[l++] = (char)i;
		}
	}
	
	return l;
}



static inline bool SearchProblem(Rand &rand, char *pattern, int patternLen, WORD *result)
{
	WORD work[9 * 9] = {};
	WORD map[9 * 9] = {};
	int counter[6] = {};
	
	for(int i = 0; i < 9; i++){
		int pos;
		do{
			pos = pattern[(int)(rand.rand() * patternLen)];
		}while(work[pos]);
		
		SetValue(work, map, pos, SOLVE_VALS[i]);
	}
	
	LogicLoop(work, map, counter);
	
	for(int i = 0; i < patternLen; i++){
		int pos = pattern[i];
		
		if(work[pos] != 0){
			continue;
		}
		
		int r = (int)(rand.rand() * 9.0) + 1;
		for(int j = 0; j < 9; j++){
			if((map[pos] & SOLVE_VALS[r]) == 0){
				SetValue(work, map, pos, SOLVE_VALS[r]);
				break;
			}
			if(++r == 10){
				r = 1;
			}
		}
		
		if(work[pos] == 0){
			return false;
		}
		
		LogicLoop(work, map, counter);
	}
	
	return (GetAnswer(work, result) == 1);
}

static inline char BitToNum(WORD work)
{
	if(work > 0x0010){
		if(work > 0x0040){
			return (work == 0x0080) ? 8 : 9;
		}else{
			return (work == 0x0020) ? 6 : 7;
		}
	}else{
		if(work > 0x0004){
			return (work == 0x0008) ? 4 : 5;
		}else if(work == 0x0004){
			return 3;
		}else if(work == 0x0002){
			return 2;
		}else{
			return 1;
		}
	}
}

bool MakeProblem(Rand &rand, char *problem, char *answer, int *counter)
{
	char pattern[9 * 9];
	int patternLen = MakePattern(rand, pattern);
	
	WORD result[9 * 9];
	int count = 0;
	while(!SearchProblem(rand, pattern, patternLen, result)){
		if(count++ == 200){
			return false;
		}
	}
	
	memset(problem, 0, 9 * 9);
	for(int i = 0; i < patternLen; i++){
		problem[pattern[i]] = BitToNum(result[pattern[i]]);
	}
	
	if(!Solve(problem, answer, counter)){
		return false;
	}
	
	bool flag[10] = {};
	for(int i = 0; i < 9 * 9; i++){
		flag[problem[i]] = true;
	}
	for(int i = 1; i < 10; i++){
		if(!flag[i]){
			return false;
		}
	}
	
	return true;
}

