
#include "sudoku_creator.h"

int calc_level(int *counter)
{
	int level = counter[0] + counter[1] * 100 + counter[2] * 10000 + 
		counter[3] * 1000000;
	
	if(level <= 7){
		return 1;
	}else if(level <= 9){
		return 2;
	}else if(level <= 11){
		return 3;
	}else if(level <= 99){
		return 4;
	}else if(level <= 199){
		return 5;
	}else if(level <= 9999){
		return 6;
	}else if(level <= 19999){
		return 7;
	}else if(level <= 999999){
		return 8;
	}else if(level <= 1999999){
		return 9;
	}else{
		return 10;
	}
}

static FILE * get_file_name(const char *ext1, int ext2)
{
	SYSTEMTIME tm;
	GetLocalTime(&tm);
	
	int count = 0;
	char fileName[256];
	for( ; ; ){
		sprintf(fileName, "%04d%02d%02d_%02d(%s%d).dat", tm.wYear, 
			tm.wMonth, tm.wDay, ++count, ext1, ext2);
		
		if(GetFileAttributes(fileName) == INVALID_FILE_ATTRIBUTES){
			FILE *ret = fopen(fileName, "wb");
			if(!ret){
				printf("failed to open: %s\n", fileName);
			}
			return ret;
		}
	}
}

static void save(FILE *file, ProblemData &problem)
{
	fprintf(file, "%d/", problem.level);
	
	for(int y = 0; y < 9; y++){
		for(int x = 0; x < 9; x++){
			char h = problem.data[y * 9 + x];
			if(h == 0){
				fprintf(file, ".");
			}else{
				fprintf(file, "%c", '0' + h);
			}
		}
		fprintf(file, "/");
	}
	fprintf(file, "\n");
}

void create(int diff, int count)
{
	static const int diff_level[] = {
		1, 1, 1, 1, 2, 2, 2, 3, 3, 3
	};
	
	FILE *file = get_file_name("diff", diff);
	if(!file){
		return;
	}
	
	Rand rand;
	ProblemData *problem = new ProblemData[count];
	
	printf("creating problems...\n");
	printf("+----------------------------+\n");
	
	int counter[6];
	for(int i = 0; i < count; i++){
		while(true){
			if(!MakeProblem(rand, problem[i].data, problem[i].answer, 
				counter)){
				continue;
			}
			if(counter[4] != 0 || counter[5] != 0){
				continue;
			}
			
			problem[i].level = calc_level(counter);
			if(diff_level[problem[i].level - 1] == diff){
				break;
			}
		}
		
		printf("o");
	}
	
	for(int i = 0; i < count; i++){
		save(file, problem[i]);
	}
	
	delete[] problem;
	fclose(file);
	
	printf("\nfinished.\n");
}

void help()
{
	printf("\number place creator\n");
	printf("  To create problem: numberplace_creator.exe c[1-3]\n");
	printf("    where 1-3 specifies difficulty.\n");
}

int main(int argc, char *argv[])
{
#ifdef _DEBUG
	_CrtSetDbgFlag(_CRTDBG_CHECK_ALWAYS_DF | _CRTDBG_LEAK_CHECK_DF | 
		_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG));
#endif

	if(argc == 2 && (argv[1][0] == 'c' || argv[1][0] == 'C')){
		int diff = argv[1][1] - '0';
		if(diff < 1 || diff > 3){
			help();
		}else{
			create(diff, 30);
		}
	}else{
		help();
	}
	
	return 0;
}

